/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/
#include <math.h>
#include <Inventor/misc/SoChildList.h>
#include "InternalVariable.h"
#include "BehaviorAction.h"
#include "BaseActivity.h"
#include "ReleasingMechanism.h"
#include "InternalVariable.h"
#include "ActivityGroup.h"

#define MAXVALUE 50.0
#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a

SO_NODE_SOURCE(ActivityGroup);

ActivityGroup::ActivityGroup() : SoGroup()
{
    SO_NODE_CONSTRUCTOR(ActivityGroup);
    SO_NODE_ADD_FIELD(errTolerance, (.1));

    // Default changed 3/15/95 hardts
    SO_NODE_ADD_FIELD(showState, (ActivityGroup::SHOWNEVER));
    
    SO_NODE_DEFINE_ENUM_VALUE(ShowType, SHOWALWAYS);
    SO_NODE_DEFINE_ENUM_VALUE(ShowType, SHOWNEVER);
    SO_NODE_DEFINE_ENUM_VALUE(ShowType, SHOWONSWITCH);
    SO_NODE_SET_SF_ENUM_TYPE(showState, ShowType);
    enableNotify(FALSE);
}

void
ActivityGroup::initClass()
{
    SO_NODE_INIT_CLASS(ActivityGroup, SoGroup, "Group");  
}


ActivityGroup::~ActivityGroup()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
ActivityGroup::readInstance(SoInput *in, unsigned short flags)
#else
ActivityGroup::readInstance(SoInput *in)
#endif
{
    int i;
    int lastChild;
    SoNode *child;
#ifdef SO_VERSION_REVISION
    if (SoGroup::readInstance(in, flags)) {
#else
    if (SoGroup::readInstance(in)) {
#endif
	lastChild = getNumChildren();
	for(i=0;i<lastChild;i++){
	    child = getChild(i);
	    if(!(child->isOfType(BaseActivity::getClassTypeId()))){
		printf("WARNING ACTIVITYGROUP CONTAINS A NON-ACTIVITY\n");
	    }
	}
	return (1);
    }
    else {
	return (0);
    }
}

void ActivityGroup::doAction(SoAction *action)
{
    int lastChildIndex = getNumChildren();
    int i;
    
    if(!action->isOfType(BehaviorUpdateAction::getClassTypeId())){
	SoGroup::doAction(action);
    }
    else {
	switchedThisTick = 0;
	BehaviorUpdateAction *bua = (BehaviorUpdateAction *)action;
	bua->currentActivityGroup = getChildren();
	
	bua->whichPass = BehaviorUpdateAction::INITIALPASS;
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i);
	
	/* 
	 * have them do a first pass at determining their value
	 * using BehaviorUpdateAction doAction
	 */ 
	bua->whichPass = BehaviorUpdateAction::ITERATIONPASS;
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i); 
	/*
	 * now check if ok or interate until ok
	 */
	bua->whichPass = BehaviorUpdateAction::ITERATIONPASS;
	iterateToEquilibrium(20, action);
	
	/* find activity with max value and make it active */
	findAndSetActiveChild();
	
	/* report state if requested */
	if(showState.getValue() == ActivityGroup::SHOWALWAYS)
	    printState();
	else if((showState.getValue() == ActivityGroup::SHOWONSWITCH)
	    && switchedThisTick){
	    printf("switched at tick %f \n", bua->tick);
	    printState();
	}
	
	/* allow activities to make recommendations */
	bua->whichPass = BehaviorUpdateAction::RECOMMENDERPASS;
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i);
	
	/* have active child dispatch */   
	bua->whichPass = BehaviorUpdateAction::DISPATCHPASS;
	children->traverse(action, activeActivityIndex);
		  
    }
}

void ActivityGroup::printState()
{
    int i, lastChild = getNumChildren();
    for(i=0;i<lastChild;i++){
	((BaseActivity *)getChild(i))->printState();
    }
}

int ActivityGroup::printState(char **bufs)
{
    int i, lastChild = getNumChildren();
    for(i=0;i<lastChild;i++){
	((BaseActivity *)getChild(i))->printState(bufs[i]);
    }
    return lastChild;
}

void ActivityGroup::turnOffChildren()
{
    int i, lastChild = getNumChildren();
    for(i=0;i<lastChild;i++){
	((BaseActivity *)getChild(i))->setActive(FALSE);
    }
    
}
float ActivityGroup::findAndSetActiveChild()
{
    int i, lastChild = getNumChildren();
    float delta = 0;
    float maxValue = -1e6, tv;
    int maxi = -1;
    
    /*
     * find child with maximum value (in fact,  only one should be
     * non-zero
     */
    for(i=0;i<lastChild;i++){
	tv = ((BaseActivity *)getChild(i))->value.getValue();
	if(tv>maxValue){
	    maxValue = tv;
	    maxi = i;
	}
    }
    if(activeActivityIndex != maxi){
	switchedThisTick = 1;
	activeActivityIndex = maxi;
    }
    activeActivity = (BaseActivity *)getChild(activeActivityIndex);
    
    /*
     * set active flag
     */
    for(i=0;i<lastChild;i++){
	((BaseActivity *)getChild(i))->setActive((i==maxi)?TRUE : FALSE);
    }
    return delta;		
}

float ActivityGroup::getDeltaAndResetValueForAll()
{
    int i, lastChild = getNumChildren();
    float delta = 0;
    
    for(i=0;i<lastChild;i++){
	delta += ((BaseActivity *)getChild(i))->getDeltaInValueAndReset(TRUE);
    }
    return delta;		
}

float ActivityGroup::jiggle()
{
    int i, lastChild = getNumChildren();
    float maxdf = -1e6,df;
    int maxi = -1;
    
    /*
     * find activity with highest value before inhibition
     */
    for(i=0;i<lastChild;i++){
	df = ((BaseActivity *)getChild(i))->getValueBeforeInhibition();
	if(df>maxdf){
	    maxdf = df;
	    maxi = i;
	}
    }
    /*
     * set its value as if it was active
     */
     for(i=0;i<lastChild;i++){
	((BaseActivity *)getChild(i))->value.setValue((i==maxi)?maxdf : 0.0);
	((BaseActivity *)getChild(i))->previousValue = ((i==maxi)?maxdf : 0.0);
    }
    return maxdf;		
}

void ActivityGroup::iterateToEquilibrium(int count, SoAction *action)
{
    int i, iters = 0;
    float err,lasterr=1e6;

    while(count--){
	/*
	 * Compute Delta from last iteration (value - previousValue)
	 * and check if within tolerance. If it is you are done.
	 */
	err = getDeltaAndResetValueForAll();
	if(err<errTolerance.getValue()){
	    return;
	}
	else{
	    /*
	     * check if you are making any progress. and if you aren't
	     * jiggle the system to see if that helps
	     */
	    if(lasterr==err){
		if(iters>4){
		    // printf("NOT CONVERGING 1\n");
		    //printState();
		}
		/*
		 * find the activity with the highest value before 
		 * inhibition and use it as the activities value for
		 * the next iteration
		 */
		jiggle();
	    }
	    lasterr = err;
	}
	/*
	 *  Ok,  try a new iteration by sending the children the
	 *  action again
	 */
	int lastChildIndex = getNumChildren();
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i); 
	iters++;
    }
    /*
     * if you get here,  system did not converge.
     */
    printf("WARNING DID NOT CONVERGE -> %f \n",err);
}












