/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#include <math.h>
#include <Inventor/misc/SoChildList.h>
#include "InternalVariable.h"
#include "BehaviorAction.h"
#include "BaseActivity.h"
#include "ReleasingMechanism.h"
#include "InternalVariable.h"
#include "ActivityGroup.h"

#define MAXVALUE 50.0
#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a


void testCallback (void *d, BehaviorUpdateAction *act)
{
    printf("i got called \n");
}

SO_NODE_SOURCE(BaseActivity);

BaseActivity::BaseActivity() : BaseBehaviorNode()
{
    SO_NODE_CONSTRUCTOR(BaseActivity);
    SO_NODE_ADD_FIELD(pname, ("BaseActivity"));
    SO_NODE_ADD_FIELD(value, (0));
    SO_NODE_ADD_FIELD(leaf, (TRUE));
    SO_NODE_ADD_FIELD(inhibitoryGains,(2.0));
    SO_NODE_ADD_FIELD(active, (FALSE));
    SO_NODE_ADD_FIELD(ignoreSuggestions, (TRUE));
    SO_NODE_ADD_FIELD(leafCBName, ("none"));
    SO_NODE_ADD_FIELD(leafNotifyCBName, ("none"));
    SO_NODE_ADD_FIELD(suggestionCBName, ("none"));
    activityGroupIndex = -1;
    indexIntoChildList = -1;
    reinforcementGroupIndex = -1;
    dovIndex = -1;
    leafCallback = NULL;
    leafNotifyCallback = NULL;
    suggestionCallback = NULL;
}

void
BaseActivity::initClass()
{
    SO_NODE_INIT_CLASS(BaseActivity, BaseBehaviorNode, "BaseBehaviorNode");
}


BaseActivity::~BaseActivity()
{
}

void BaseActivity::addADependsOnVariable(SoNode *aDov)
{
    if(dovIndex == -1){
	addChild(new DependsOnVariables());
	dovIndex = getNumChildren()-1;
    }
    ((SoGroup *)getChild(dovIndex))->addChild(aDov);
}

SbBool	
#ifdef SO_VERSION_REVISION
BaseActivity::readInstance(SoInput *in, unsigned short flags)
#else
BaseActivity::readInstance(SoInput *in)
#endif
{
    int i;
    int lastChild;
    SoNode *child;
    
#ifdef SO_VERSION_REVISION
    if (SoGroup::readInstance(in, flags)) {
#else
    if (SoGroup::readInstance(in)) {
#endif
	lastChild = getNumChildren();
	for(i=0;i<lastChild;i++){
	   child = getChild(i);
	    if(child->isOfType(DependsOnVariables::getClassTypeId())){
		dovIndex = i;
		break;
	    }
	}
	for(i=0;i<lastChild;i++){
	   child = getChild(i);
	    if(child->isOfType(ActivityGroup::getClassTypeId())){
		activityGroupIndex = i;
		break;
	    }
	}
	for(i=0;i<lastChild;i++){
	   child = getChild(i);
	    if(child->isOfType(ReinforcementGroup::getClassTypeId())){
		reinforcementGroupIndex = i;
		break;
	    }
	}
	leaf.setValue((activityGroupIndex> -1)?FALSE : TRUE);
	
	if(leafCBName.getValue() != SbString("none")){
//	    printf("installing leaf callback %s \n", 
//	    leafCBName.getValue().getString());
	    setLeafCallback(getActCBByName(leafCBName.getValue()));
	}
	if(leafNotifyCBName.getValue() != SbString("none")){
//	    printf("installing leaf notify callback %s \n", 
//	    leafNotifyCBName.getValue().getString());
	    leafNotifyCallback = getActCBByName(leafNotifyCBName.getValue());
	}
	if(suggestionCBName.getValue() != SbString("none")){
//	    printf("installing suggestion callback %s \n", 
//	    suggestionCBName.getValue().getString());
	    setSuggestionCallback(getActCBByName(suggestionCBName.getValue()));
	}
	return (1);
    }
    else {
	return (0);
    }
}

void BaseActivity::setActive(SbBool isOn)
{
    if((active.getValue() != isOn)
	&& (!isOn)
	&& (activityGroupIndex != -1)){
	
	((ActivityGroup *)getChild(activityGroupIndex))->turnOffChildren();
    }
    active.setValue(isOn);
}

void BaseActivity::installPronome(SoAction *action)
{
    BehaviorUpdateAction *bua = (BehaviorUpdateAction *)action;
    DependsOnVariables *dov;
    dov = (dovIndex != -1)?(DependsOnVariables *)getChild(dovIndex) : NULL;
    if(dov){
	ReleasingMechanism *rm = dov->getRMwithMaxValue();
	if(rm){
	    objectOfInterest = rm->objectOfInterest;
	    bua->pronomes.add(&rm->objectOfInterest);
	}
    }
}

InternalVariable *BaseActivity::getMaxUnderlyingMotivation()
{
    if((reinforcementGroupIndex > -1) && (dovIndex > -1)){
	ReinforcementGroup *rg = (ReinforcementGroup *)getChild(reinforcementGroupIndex);
	return rg->getMaxUnderlyingMotivation();
    }
    else return NULL;
}

void BaseActivity::doLearning(SoAction *action)
{
    if((reinforcementGroupIndex > -1) && (dovIndex > -1)){
	DependsOnVariables *dov = (DependsOnVariables *)getChild(dovIndex);
//	printf("calcing reinforcement \n");
	getChild(reinforcementGroupIndex)->doAction(action);
//	printf("getting reinforcement \n");
	float r = ((ReinforcementGroup *)getChild(reinforcementGroupIndex))->rsum;
	if(r> 1.0e-4){
	    ReleasingMechanism *rm = dov->getRMwithMaxValue();
	    rm->doLearning(r, (BehaviorUpdateAction *)action);
	}
    }
}

void BaseActivity::removePronome(SoAction *action)
{
   BehaviorUpdateAction *bua = (BehaviorUpdateAction *)action;
   bua->pronomes.remove(&objectOfInterest);  
}

void BaseActivity::doAction(SoAction *action)
{
    int numIndices;
    int lastChildIndex;
    const int *indices;
    int i;
    float v;
    
    if(action->getPathCode(numIndices, indices) == 
	SoAction::IN_PATH)
	lastChildIndex = indices[numIndices -1];
    else
	lastChildIndex = getNumChildren();
    if(action->isOfType(InternalVariableUpdateAction::getClassTypeId())){
	v = value.getValue();
	previousValue = v;
	for(i=0;i<lastChildIndex;i++){
	    InternalVariableUpdateAction *ba = (InternalVariableUpdateAction *)action;
	    ba->value = (active.getValue())?v :0;
//	    printf("doAction of BaseBehavior \n");
	    children->traverse(action, i);
	    ba->value = v;
	}
    }
    else if(action->isOfType(BehaviorUpdateAction::getClassTypeId())){
	/*
	 * first time through let dependsOnVariables get a chance to
	 * update
	 */
	BehaviorUpdateAction *bua = (BehaviorUpdateAction *)action;
	if((bua->whichPass == BehaviorUpdateAction::INITIALPASS)
	    && (dovIndex != -1)){
	    if(active.getValue()==TRUE)
		doLearning(action);
	    if(reinforcementGroupIndex > -1)
		bua->maxUnderlyingMotivation = getMaxUnderlyingMotivation();
	    else
		bua->maxUnderlyingMotivation = NULL;
	    /*
	     * adds this activities pronome to pronomelist
	     */
	  //   bua->pronomes.add(&objectOfInterest); 
	    
	    /*
	     * The next two lines have the effect of updating the
	     * DependsOnVariables child. This is when the releasing
	     * mechanisms would get evaluated.
	     */
	    previousValue = value.getValue();
	    getChild(dovIndex)->doAction(action);
	}
	else if(bua->whichPass == BehaviorUpdateAction::ITERATIONPASS){
	    updateValueAfterInhibition(*((BehaviorUpdateAction *)action)
	    ->currentActivityGroup);
	}
	else if(bua->whichPass == BehaviorUpdateAction::RECOMMENDERPASS){
	    if(active.getValue() != TRUE){
		installPronome(bua);
		doRecommendation(action);
		removePronome(bua);
	    }
	}
	else if(bua->whichPass == BehaviorUpdateAction::DISPATCHPASS){
	    installPronome(bua);
	    if(activityGroupIndex != -1)
		children->traverse(action,activityGroupIndex);
	    else {
		if(leafNotifyCallback != NULL)
		    leafNotifyCallback(this, (BehaviorUpdateAction *)action); 
		doLeafAction(action);
	    }
	    removePronome(bua); 
	}
	
    }
    else {
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i);    
    }
}
void BaseActivity::setSuggestionCallback(activityCB callback)
{
    suggestionCallback = callback;
}

void BaseActivity::setLeafCallback(activityCB callback)
{
    leafCallback = callback;
}

void BaseActivity::doLeafAction(SoAction *action)
{
    if((leafCallback  != NULL))
	leafCallback(this, (BehaviorUpdateAction *)action);
}

void BaseActivity::doRecommendation(SoAction *a)
{
    if((suggestionCallback  != NULL))
	suggestionCallback(this, (BehaviorUpdateAction *)a);
}

float BaseActivity::calcInhibitionOn(int i)
{
   int maxNumGains = inhibitoryGains.getNum();
   const float *gains = inhibitoryGains.getValues(0);
   float gainToUse = (i<maxNumGains)? gains[i] : gains[maxNumGains-1];
   return gainToUse * previousValue;
}

float BaseActivity::sumInhibition(SoNodeList &otherActivities)
{
    int i, numChildren;
    inhibition = 0.0;
    numChildren = otherActivities.getLength();
    SoNode *child;
    
    /* this is to cache the index of this guy in his parent's childlist */
    if(indexIntoChildList < 0){
	for(i=0;i<numChildren;i++){
	    child = otherActivities[i];
	    if(child == this)
		break;
	}
	indexIntoChildList = (indexIntoChildList < numChildren)?i :-1;
	if(indexIntoChildList == -1){
	    printf("indexNotValid \n");
	    return 0;
	}
    }
    
    /* this walks through */
    for(i=0;i<numChildren;i++){
	if((otherActivities[i]->isOfType(BaseActivity::getClassTypeId()))
	&& (otherActivities[i] != this)){
	    child = otherActivities[i];
	    inhibition += ((BaseActivity *)child)->calcInhibitionOn(indexIntoChildList);
	}
    }
	return inhibition;
}

void BaseActivity::updateValueAfterInhibition(SoNodeList &otherActivities)
{
    DependsOnVariables *dov;
    dov = (dovIndex != -1)?(DependsOnVariables *)getChild(dovIndex) : NULL;
    if(dov != NULL){
	float vai = dov->levelOfInterest * (dov->value - sumInhibition(otherActivities));
	vai = MAX(vai, 0.0);
	value.setValue(vai);    
    }
}

float BaseActivity::getValueBeforeInhibition()
{
    DependsOnVariables *dov;
    dov = (dovIndex != -1)?(DependsOnVariables *)getChild(dovIndex) : NULL;
    if(dov != NULL){
	float vai = dov->levelOfInterest * dov->value;
	return vai;    
    }
    return 0;
}
    
float BaseActivity::getDeltaInValueAndReset(SbBool shouldReset)
{
    float delta = fabs(value.getValue() - previousValue);
    if(shouldReset== TRUE)
	previousValue = value.getValue();
    return delta;
}
  
void BaseActivity::printState(char *buf)
{
    DependsOnVariables *dov;
    dov = (dovIndex != -1)?(DependsOnVariables *)getChild(dovIndex) : NULL;
    if(dov != NULL){
	if(!buf){
	    printf("%s int-> %4.3f  rm-> %4.1f iv -> %4.1f inh-> %4.1f value -> %f\n", 
		pname.getValue().getString(), 
		dov->levelOfInterest, 
		dov->releasingMechanismSum, 
		dov->internalVariablesSum,
		inhibition, 
		value.getValue());
		 
	}
	else {
	    sprintf(buf, 
		"%s interest -> %f  external -> %f internal -> %f inhibition -> %f value -> %f\n", 
		pname.getValue().getString(), 
		dov->levelOfInterest, 
		dov->releasingMechanismSum, 
		dov->internalVariablesSum,
		inhibition, 
		value.getValue());
	}
    }
}









