/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/SoDB.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFNode.h>
#include "ReleasingMechanism.h"
#include "BehaviorAction.h"
#include "Pronome.h"

#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a

SO_NODE_SOURCE(ReleasingMechanism);

ReleasingMechanism::ReleasingMechanism() : BaseBehaviorNode()
{
    SO_NODE_CONSTRUCTOR(ReleasingMechanism);
    SO_NODE_ADD_FIELD(typeTag, (0));
    SO_NODE_ADD_FIELD(minValue, (0));
    SO_NODE_ADD_FIELD(maxValue, (50));
    SO_NODE_ADD_FIELD(activeRange, ((0, 0 ,20)));
    SO_NODE_ADD_FIELD(period, (1));
    SO_NODE_ADD_FIELD(sensor, (0));
    SO_NODE_ADD_FIELD(findCBName, ("none"));
    SO_NODE_ADD_FIELD(filterCBName, ("none"));
    SO_NODE_ADD_FIELD(weightCBName, ("none"));
    SO_NODE_ADD_FIELD(type, (ReleasingMechanism::STDRELEASER));
    
    SO_NODE_DEFINE_ENUM_VALUE(ReleaserType, STDRELEASER);
    SO_NODE_DEFINE_ENUM_VALUE(ReleaserType, SUMMINGRELEASER);
    SO_NODE_DEFINE_ENUM_VALUE(ReleaserType, AVGINGRELEASER);
    SO_NODE_DEFINE_ENUM_VALUE(ReleaserType, TRIGGERRELEASER);
    SO_NODE_SET_SF_ENUM_TYPE(type, ReleaserType);
    
    SO_NODE_ADD_FIELD(reinforcementWindow, (0));
    SO_NODE_ADD_FIELD(reinforcedMaxValue, (50));
    SO_NODE_ADD_FIELD(experimentGain, (0));
    reinforcementReceivedAt = 0;
    initialMaxValue = 0;
    lastAdjustedAt = 0;
    maxValueAt = 0;
    rmFilterCallback = NULL;
    rmWeightCallback = NULL;
    rmFindCallback = NULL;
}

void
ReleasingMechanism::initClass()
{
    SO_NODE_INIT_CLASS(ReleasingMechanism, BaseBehaviorNode, "BaseBehaviorNode");
}


ReleasingMechanism::~ReleasingMechanism()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
ReleasingMechanism::readInstance(SoInput *in, unsigned short flags) 
#else
ReleasingMechanism::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (BaseBehaviorNode::readInstance(in, flags)) {
#else
    if (BaseBehaviorNode::readInstance(in)) {
#endif
	if(type.getValue() == ReleasingMechanism::SUMMINGRELEASER){
	    hist = new float[(int)period.getValue()+1];
	    next = 0;
	    int i;
	    for(i=0;i<(int)period.getValue();i++)
		hist[i] = 0;
	}
	else{
	    hist = new float;
	    *hist = 0.0;
	}
//	printf("%s %s %s \n", 
//	    findCBName.getValue().getString(), 
//	    filterCBName.getValue().getString(),
//	    weightCBName.getValue().getString());
	
	if(findCBName.getValue() != SbString("none")){
//	    printf("installing find callback \n");
	    setRMFindCallback(getFilterCBByName(findCBName.getValue()));
	}
	if(filterCBName.getValue() != SbString("none")){
//	    printf("installing filter callback \n");
	    setRMFilterCallback(getFilterCBByName(filterCBName.getValue()));
	}
	if(weightCBName.getValue() != SbString("none")){
//	    printf("installing weight callback \n");
	    setRMWeightCallback(getWeightCBByName(weightCBName.getValue()));
	}
	
	initialMaxValue = maxValue.getValue();
	reinforcedMaxValue.setValue(maxValue.getValue());
	return (1);
    }
    else {
	return (0);
    }
}

void ReleasingMechanism::doLearning(float r, BehaviorUpdateAction *action)
{
    if(r> 1.0e-4){
	float err = r - value.getValue();
	float adj = .5 *err;
	reinforcedMaxValue.setValue(maxValue.getValue() + adj);
	maxValue.setValue(reinforcedMaxValue.getValue());
	if(action->tick - reinforcementReceivedAt>1){
	    printf("TIME SINCE LAST VISIT -> %f FOR %s \n", 
		action->tick-reinforcementReceivedAt,
		pname.getValue().getString()); 
	    printf("  started learning at time %f for %s ->  %f\n",
		action->tick,  
		pname.getValue().getString(), 
		reinforcedMaxValue.getValue());
	}
	else
	    printf("  cont. learning at time %f for %s ->  %f\n",
		action->tick,  
		pname.getValue().getString(), 
		reinforcedMaxValue.getValue());
		
	reinforcementReceivedAt = action->tick;
    }
}

void ReleasingMechanism::forget(BehaviorUpdateAction *action)
{
    int currTick = ((BehaviorUpdateAction *)action)->tick;
    if(reinforcementWindow.getValue()){
	float rw = reinforcementWindow.getValue();
	float factor = ((float)(currTick - reinforcementReceivedAt))/rw;
	if(factor>1.0) 
	    factor = 1.0;
	// use an exponential decay, where factor will range from 0 to 1
	factor = exp((factor -1)*4);
	float rmv = reinforcedMaxValue.getValue();
	float newValue;
	if(fabs(experimentGain.getValue())<1.0e-4)
	    newValue = rmv + factor*(initialMaxValue - rmv);
	else {
	    InternalVariable *mot = action->maxUnderlyingMotivation;
	    float mv = 0;
	    if(mot){
		mv = experimentGain.getValue()*mot->value.getValue();
	    }
	    newValue = rmv + factor *(
		((initialMaxValue - rmv - mv)> 0.0)?
		    (initialMaxValue - rmv - mv):
		    0.0);
	}
	maxValue.setValue(newValue);
    }
}
void ReleasingMechanism::doAction(SoAction *action)
{
    if(!action->isOfType(BehaviorUpdateAction::getClassTypeId())){
	BaseBehaviorNode::doAction(action);
    }
    else {
	forget((BehaviorUpdateAction *)action);
	updateValue((BehaviorUpdateAction *)action);
    }
}

void ReleasingMechanism::setRMFilterCallback(rmFilterCB callback)
{
    rmFilterCallback = callback;
}

void ReleasingMechanism::setRMFindCallback(rmFilterCB callback)
{
    rmFindCallback = callback;
}
void ReleasingMechanism::setRMWeightCallback(rmWeightCB callback)
{
    rmWeightCallback = callback;
}
    
void ReleasingMechanism::adjustValueFromHistoryAt(int t)
{
  int i,ip;
  //float tv = value.getValue();
  float tv = v_tmp;
  float tp = period.getValue();
  
  if(type.getValue() == ReleasingMechanism::SUMMINGRELEASER){
    ip = (int)tp;
    if((t-lastAdjustedAt)>= tp){
	next = 0.0;
	for(i=0;i<ip;i++)
	    hist[i] = 0.0;
    }
    hist[next] = tv;
    next++;
    next = next % ip;
    for(tv = 0,i=0;i<ip;i++)
      tv += hist[i];
  }
  else if(type.getValue() == ReleasingMechanism::AVGINGRELEASER){
    if((t-lastAdjustedAt)>= tp){
	hist[0] = 0;
    }
    tp = 1.0/tp;
    tv = hist[0] *(1-tp) + tp*tv;
    hist[0] = tv;
  }
  else if(type.getValue() == ReleasingMechanism::TRIGGERRELEASER){
	if((hist[0]<= tv)){
	    hist[0] = tv;
	    maxValueAt = t;
	}
	else if(maxValueAt > t - tp)
	    tv = hist[0];
  }
  lastAdjustedAt = t;
  v_tmp = tv;
  //value.setValue(tv); 
}

void ReleasingMechanism::reportStateAt(int t)
{
  printf("name => ( %s ) type => %d value => %f\n",
	 pname.getValue(), type.getValue(),value.getValue());
}

float ReleasingMechanism::linearWeightBasedOnRange(float r)
{
    float w;
    const float *ar = activeRange.getValues(0);
    int arc = activeRange.getNum();
    float mn, mx, best;
    float t = 0;
    
    mn = ar[0];
    if(arc<3){
	best = ar[0];
	mx = ar[1];
    }
    else if(arc == 3){
	best = ar[1];
	mx = ar[2];
    }
    if((r >= best) && (r<=mx))
	t = 1.0-((r-best)/(mx-best));
    else if((r<best) && (r>=mn))
	t = (best-r)/(best-mn);
    return t;
}

float ReleasingMechanism::calcWeightBasedOnRange(float r)
{
    float t = linearWeightBasedOnRange(r);
    float w = (t>0)?3*t*t - 2*t*t*t : 0;
    return w;
}

SbBool ReleasingMechanism::find(BehaviorUpdateAction *action)
{
    SbBool fnd = FALSE;
    if((rmFindCallback  != NULL)){
	fnd = rmFindCallback(this, (BehaviorUpdateAction *)action);
    }
    if(!fnd)
	v_tmp = minValue.getValue();
	//value.setValue(minValue.getValue());
    return fnd;
}

SbBool ReleasingMechanism::filter(BehaviorUpdateAction *action)
{
    SbBool passesFilter = TRUE;
    Pronome *p = &objectOfInterest;
    if((rmFilterCallback  != NULL)){
	passesFilter =rmFilterCallback(this, (BehaviorUpdateAction *)action);
    }
    if(!passesFilter){
	v_tmp = minValue.getValue();
	//value.setValue(minValue.getValue());
	p->status = Pronome::FAILEDFILTER;
    }
    else
	p->status = Pronome::PASSEDFILTER;
    return passesFilter;
}

float ReleasingMechanism::weight(BehaviorUpdateAction *action)
{
    float weightedValue = v_tmp;
    //float weightedValue = value.getValue();
    
    if((rmWeightCallback  != NULL))
	weightedValue = rmWeightCallback(this, (BehaviorUpdateAction *)action);
    v_tmp = weightedValue;
    //value.setValue(weightedValue);
    return weightedValue;
}

void ReleasingMechanism::_updateValue(BehaviorUpdateAction *act)
{
    if(find(act)&& filter(act))
	weight(act);
}

float ReleasingMechanism::updateValue(BehaviorUpdateAction *act)
{
 int t = (int)act->tick;
 if(printDebug.getValue())
    printf("debug at tick %i \n", t);
 _updateValue(act);
 
 if((lastAdjustedAt != t)&&(type.getValue() != ReleasingMechanism::STDRELEASER))
    adjustValueFromHistoryAt(t);
  v_tmp = MIN(v_tmp, maxValue.getValue());
  v_tmp = MAX(v_tmp, minValue.getValue());
  value.setValue(v_tmp);  
  //value.setValue(MIN(value.getValue(), maxValue.getValue()));
  //value.setValue(MAX(value.getValue(), minValue.getValue()));
  if(printDebug.getValue())
    printf("debug value at tick %i %f \n", t, value.getValue());
  return value.getValue();
}

