#include <stdio.h>
#include <tcl.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <Inventor/sensors/SoSensor.h>
#include <Inventor/sensors/SoIdleSensor.h>

#include "globals.h"
extern Globals globals;

/*
 * Declarations for various library procedures and variables (don't want
 * to include tclUnix.h here, because people might copy this file out of
 * the Tcl source directory to make their own modified versions).
 */

extern int		errno;
extern void		exit _ANSI_ARGS_((int status));
extern int		isatty _ANSI_ARGS_((int fd));
extern char *		strcpy _ANSI_ARGS_((char *dst, CONST char *src));

static Tcl_DString command;	/* Used to buffer incomplete commands being
				 * read from stdin. */
char *tcl_RcFileName = NULL;	/* Name of a user-specific startup script
				 * to source if the application is being run
				 * interactively (e.g. "~/.tclshrc").  Set
				 * by Tcl_AppInit.  NULL means don't source
				 * anything ever. */



//
// These variables global so that tclInit, tclDone, and tclCB can access them
//
char buffer[1000], *cmd, *ptr;
int code, gotPartial, tty;
int exitCode = 0;

void
tclDone()
{

    /*
     * Rather than calling exit, invoke the "exit" command so that
     * users can replace "exit" with some other command to do additional
     * cleanup on exit.  The Tcl_Eval call should never return.
     */
    sprintf(buffer, "exit %d", exitCode);
    Tcl_Eval(globals.interp, buffer);
    return;
}

void
tclInit(int argc, char **argv)
{
    char *args, *fileName;

    globals.interp = Tcl_CreateInterp();

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".  If the first argument doesn't start with a "-" then
     * strip it off and use it as the name of a script file to process.
     */

    fileName = NULL;
    if ((argc > 1) && (argv[1][0] != '-')) {
	fileName = argv[1];
	argc--;
	argv++;
    }
    args = Tcl_Merge(argc-1, argv+1);
    Tcl_SetVar(globals.interp, "argv", args, TCL_GLOBAL_ONLY);
    ckfree(args);
    sprintf(buffer, "%d", argc-1);
    Tcl_SetVar(globals.interp, "argc", buffer, TCL_GLOBAL_ONLY);
    Tcl_SetVar(globals.interp, "argv0", (fileName != NULL) ? fileName : argv[0],
	    TCL_GLOBAL_ONLY);

    /*
     * Set the "tcl_interactive" variable.
     */

    tty = isatty(0);
    Tcl_SetVar(globals.interp, "tcl_interactive",
	    ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Invoke application-specific initialization.
     */

    if (Tcl_AppInit(globals.interp) != TCL_OK) {
	fprintf(stderr, "Tcl_AppInit failed: %s\n", globals.interp->result);
    }

    /*
     * If a script file was specified then just source that file
     * and quit.
     */

    if (fileName != NULL) {
	code = Tcl_EvalFile(globals.interp, fileName);
	if (code != TCL_OK) {
	    fprintf(stderr, "%s\n", globals.interp->result);
//		    exitCode = 1;
	}
//	tclDone();
    }

    /*
     * We're running interactively.  Source a user-specific startup
     * file if Tcl_AppInit specified one and if the file exists.
     */

    if (tcl_RcFileName != NULL) {
	Tcl_DString buffer;
	char *fullName;
	FILE *f;

	fullName = Tcl_TildeSubst(globals.interp, tcl_RcFileName, &buffer);
	if (fullName == NULL) {
	    fprintf(stderr, "%s\n", globals.interp->result);
	} else {
	    f = fopen(fullName, "r");
	    if (f != NULL) {
		code = Tcl_EvalFile(globals.interp, fullName);
		if (code != TCL_OK) {
		    fprintf(stderr, "%s\n", globals.interp->result);
		}
		fclose(f);
	    }
	}
	Tcl_DStringFree(&buffer);
    }

    /*
     * Process commands from stdin until there's an end-of-file.
     */

    gotPartial = 0;
    Tcl_DStringInit(&command);


    ptr = buffer;

}

int
nonBlockingReadFromStdin(char *buf, int maxsize)
{
  struct timeval timeout = {0,0};
  fd_set readfds;
  int width;
  int bytes;
  
  width = getdtablesize();

  FD_ZERO(&readfds);
  FD_SET(STDIN_FILENO, &readfds);
  
  if (select(width, &readfds, NULL, NULL, &timeout) < 0)
    {
      return -1;
    }

  if (FD_ISSET(STDIN_FILENO, &readfds))
    {
      ioctl(STDIN_FILENO, FIONREAD, &bytes);
      if (bytes < maxsize)
	{
	  bytes = read(STDIN_FILENO, buf, 256);
	  buf[bytes] = 0;
	  if (buf[bytes - 1] == '\n')
	    buf[bytes - 1] = 0;
	  return bytes;
	}
      else
	return -1;
    }
  else
    {
      return 0;
    }
}
  

void 
tclCB(void *, SoSensor *sensor)
{
   //
   // re-schedule the sensor
   //
   sensor->schedule();

   //
   //  non-blocking read
   //

/*
   char c = ' ';
   int f = fcntl(0, F_GETFL);

   fcntl(0, F_SETFL, f|FNONBLOCK);
   while ((read(0, &c, 1) == 1) && (c != '\n'))
	*ptr++ = c;
   fcntl(0, F_SETFL, f);

   if (c != '\n')
       return;

   *ptr = 0;
    ptr = buffer;
*/
   int read_return;
   if ((read_return = nonBlockingReadFromStdin(buffer, 1000)) < 0)
     printf("tclCB: error reading from standard input\n");
   else
     if (read_return == 0)
       return;
     
   //
   // buffer now contains a command to process
   //
	cmd = Tcl_DStringAppend(&command, buffer, -1);
	if ((buffer[0] != 0) && !Tcl_CommandComplete(cmd)) {
	    gotPartial = 1;
	    return;
	}

	gotPartial = 0;
	code = Tcl_RecordAndEval(globals.interp, cmd, 0);
	Tcl_DStringFree(&command);
	if (code != TCL_OK) {
	    fprintf(stderr, "%s\n", globals.interp->result);
	} else if (tty && (*globals.interp->result != 0)) {
	    printf("%s\n", globals.interp->result);
	}

	clearerr(stdin);
	if (tty) {
	    char *promptCmd;

	    promptCmd = Tcl_GetVar(globals.interp,
		gotPartial ? "tcl_prompt2" : "tcl_prompt1", TCL_GLOBAL_ONLY);
	    if (promptCmd == NULL) {
		defaultPrompt:
		if (!gotPartial) {
		    fputs("% ", stdout);
		}
	    } else {
		code = Tcl_Eval(globals.interp, promptCmd);
		if (code != TCL_OK) {
		    fprintf(stderr, "%s\n", globals.interp->result);
		    Tcl_AddErrorInfo(globals.interp,
			    "\n    (script that generates prompt)");
		    goto defaultPrompt;
		}
	    }
	    fflush(stdout);
	}


    return;
    }
