#ifndef _AVATAR_
#define _AVATAR_

#include <Inventor/Sb.h>
#include <Inventor/fields/SoMFFloat.h>
#include <Inventor/fields/SoMFInt32.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFVec2f.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSubNode.h>

class ArticulatedBody;
class AvatarFrameInfo;
class AvatarCache;
class Avatar;
class SoCoordinate3;
class SoIndexedFaceSet;
class SoInput;
class SoSwitch;
class SoText2;
class SoTransform;
class SoTextureCoordinate2;

void AvatarInit();

class AvatarFrameInfo : public SoNode {
  SO_NODE_HEADER(AvatarFrameInfo);

public:
  AvatarFrameInfo();
  static void initClass();

  SoMFInt32 annotationIndices;

  SoSFBool useLeftArm;
  SoSFBool useRightArm;

  // initial arm state
  SoSFVec3f leftShoulderPos;
  SoSFVec3f rightShoulderPos;
  SoSFVec3f leftElbowPos;
  SoSFVec3f rightElbowPos;
  SoSFVec3f leftEndPos;
  SoSFVec3f rightEndPos;
  SoSFVec2f leftAngles;
  SoSFVec2f rightAngles;

  SoSFVec2f leftShoulderRange;
  SoSFVec2f leftElbowRange;
  SoSFVec2f rightShoulderRange;
  SoSFVec2f rightElbowRange;

  // tracked features from pfinder for this frame

  SoSFVec3f centroid;
  SoSFVec3f head;
  SoSFVec3f leftHand;
  SoSFVec3f rightHand;

  // map from texture to world

  SoSFVec2f texscale;
  SoSFVec2f texoffset;


protected:
  virtual ~AvatarFrameInfo();
#ifdef SO_VERSION_REVISION
  virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
  virtual SbBool readInstance(SoInput *in);
#endif
};

class AvatarCache {
public:
  AvatarFrameInfo *afi;
  SoCoordinate3 *originalCoords;
  SoCoordinate3 *newCoords;
  SoTextureCoordinate2 *texCoords;
  SoIndexedFaceSet *faceSet;
  SoTransform *leftABXform;
  SoTransform *rightABXform;
  ArticulatedBody *leftArm;
  ArticulatedBody *rightArm;
  SoText2 *frameText;
  AvatarCache();
  ~AvatarCache();
};

class Avatar : public SoSeparator {
  SO_NODE_HEADER(Avatar);

public:
  Avatar();



  SbBool buildFromPfinderStatisticsFiles(char *user_raw_filename,
					 char *user_seg_filename,
					 char *user_stat_filename,
					 char *user_output_base_filename);
  SbBool writeToIVFile(char *filename);

  
  void updateToNewPositions(float normal_height,
			    SbVec3f left_hand_pos,
			    SbVec3f right_hand_pos,
			    SbVec3f head_pos,
			    SbVec3f centroid_pos);
  void setAvatarScale(float newScale);
  void switchToFrame(int whichFrame);
  void buildFromPfinderStatisticsFiles(char *user_raw_filename,
				       char *user_seg_filename,
				       char *user_stat_filename,
				       char *user_cont_filename,
				       int user_image_width,
				       int user_image_height,
				       int user_num_frames,
				       int standing_num_frames,
				       char *user_output_base_filename);


  static void initClass();

  SoScale *avatarScale;
  SoSwitch *frameSwitch;

  SoSeparator *tailSep;
  AvatarCache *currentFrame;

  SoSFFloat avatarNormalHeight;

  void normalize(float aspect_ratio);

  void resetArmAngles();
  void buildLeftArm();
  void buildRightArm();
  int annoEditMode;
  int doSelect;
  int doWarp;
  void buildFaceSetsFromAnnotations();

  int getNearestContourPoint(SbVec3f &v);
  SbVec2f curTexLoc(SbVec3f p);
protected:

  SoSFInt32 avatarStandingFrameCount;
  SoSFString sceneGraphFilename;

  SoPath *getPathToName(SoNode *from, const SbName &name);
  SoSwitch *originalCoordinateSwitch;

  AvatarCache *frame(int i);

  SbPList frameCaches;
  char *raw_filename;
  char *seg_filename;
  char *stat_filename;
  char *cont_filename;
  char *output_base_filename;
  int image_width;
  int image_height;
  int num_frames;

  virtual ~Avatar();
  static int lcmp(const void *a, const void *b);
  void sortAnnotationIndices();
  void buildCaches();
  void buildArms();

  void centerHeads();
  void restoreAllCoords();

  SbVec3f constrainedSetGoal(AvatarCache *f, SbVec3f g, int sign);

  SbVec3f setGoal(ArticulatedBody *b, SbVec3f g);
  SbBool inverse2DOF(ArticulatedBody *b, SbVec3f g, int sign, float &a1, float &a2);
  void setLengths(ArticulatedBody *b, float l1, float l2);
  void getLengths(ArticulatedBody *b, float &l1, float &l2);
  void getAngles(ArticulatedBody *b, float &a1, float &a2);
  void setAngles(ArticulatedBody *b, float a1, float a2);
  void removeLimbs(ArticulatedBody *b);
  void setElbowCenter(ArticulatedBody *b, float a);
  SbVec3f getMid(ArticulatedBody *b);

  ArticulatedBody * loadArm();
  SoSeparator * buildArmArticulatedBody();
  void computeAvatarNormalHeight();
  void normalizeAvatarContours(float old_scale,
			       float new_scale,
			       float aspect_ratio);
  void normalizeAvatarAFIs(float old_scale,
			       float new_scale,
			       float aspect_ratio);
  void normalizeAvatarSF(SoSFVec3f *field, 
			       float old_scale,
			       float new_scale,
			       float aspect_ratio);
  void addGroupsToSceneGraph();
  SoNode * getNthChildOfType(SoGroup *parent,
			     int whichChild,
			     SoType classId);
  void buildSceneGraph();
  int selectFrame(float normal_height,
		  SbVec3f left_hand_pos,
		  SbVec3f right_hand_pos,
		  SbVec3f head_pos,
		  SbVec3f centroid_pos);
  void copyCurrentFrameCoordinates();
  void leftArmWarp(SbVec3f newend, SbVec3f newmid);
  void rightArmWarp(SbVec3f newend, SbVec3f newmid);


  void convertPfinderRawImagesToTextures();
  unsigned char getSegPixel(unsigned char *buf, int x, int y);
  int getNumberOfNearestNeighborsInSegmentation(unsigned char *buf, int x, int y);
  unsigned char getFilteredAlpha(unsigned char *buf, int x, int y);
  void combineRawImageWithSegmentation(unsigned char *dumpbuf, unsigned char *segbuf, unsigned short *outbuf);
#ifdef SO_VERSION_REVISION
  virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
  virtual SbBool readInstance(SoInput *in);
#endif
};

#endif
