#include <stdio.h>

#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/fields/SoSFNode.h>

#include "EngineHolder.h"
#include "LimbModel.h"
#include "ArticulatedBody.h"
#include "AnchorManip.h"


#define NUMKITER 5

ABConstraint::ABConstraint(ArticulatedBody *c, SoNode *root, SoNode *an, 
    SbBool applyOnEnd)
{
    client = c;
    if(client == root)
	relativeToPath = client->path;
    else 
	relativeToPath = client->buildPath(root);
    adjustNode = an;
    applyFlag = applyOnEnd;
    weight = SbVec3f(1, 1, 1);
}

void ABConstraint::beginConstraint()
{
    constrainedPt.setValue(0, 0, 0);
    client->vecToBaseUsing(constrainedPt, relativeToPath);
}

void ABConstraint::endConstraint()
{
    newPt = SbVec3f(0, 0, 0);
    client->vecToBaseUsing(newPt, relativeToPath);
    
    delta = constrainedPt-newPt;
    int i;
    for(i=0;i<3;i++)
	delta[i] *= weight[i];
    if(applyFlag == TRUE)
	applyConstraint();
}

void ABConstraint::applyConstraint()
{
    if(adjustNode->isOfType(SoTransform::getClassTypeId())){
	    //printf("satisfying constraint \n");
	    ((SoTransform *)adjustNode)->translation.setValue(
		((SoTransform *)adjustNode)->translation.getValue() + delta);
    }
    else if(adjustNode == client){
	newPt += delta;
	client->vecToEndUsing(newPt, relativeToPath);
	client->vecToBase(newPt);
	client->goal.setValue(newPt);
	client->moveToGoalImmediate();
    }
}
	   
/*********************************************************************
 * 
 *	    ArticulatedBody  Class
 * 
 *********************************************************************/


SO_NODE_SOURCE(ArticulatedBody);

void
ArticulatedBody::initClass()
{
    SO_NODE_INIT_CLASS(ArticulatedBody, SoSeparator, "Separator");
}

ArticulatedBody::ArticulatedBody() : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(ArticulatedBody);
    SO_NODE_ADD_FIELD(goal, (0.0, 0.0, 0.0));
    SO_NODE_ADD_FIELD(tol, (0.01));
    SO_NODE_ADD_FIELD(angleHints, (0.0, 0.0, 0.0));
    SO_NODE_ADD_FIELD(angleHintStrengths, (0.0, 0.0, 0.0));
    SO_NODE_ADD_FIELD(interpolatorCache,(NULL));
    SO_NODE_ADD_FIELD(pickInterpolator, (-1));
    SO_NODE_ADD_FIELD(useCache, (TRUE));

    goalSensor = new SoFieldSensor(&ArticulatedBody::goalSensorCB, this);
    goalSensor->attach( &goal );

    angleHintsSensor = new SoFieldSensor(&ArticulatedBody::hintsSensorCB, this);
    angleHintsSensor->attach( &angleHints );
    
    gainHintsSensor = new SoFieldSensor(&ArticulatedBody::hintsSensorCB, this);
    gainHintsSensor->attach( &angleHintStrengths );
    
    gainHintsSensor->setPriority(STRENGTHSENSORPRIORITY);
    angleHintsSensor->setPriority(HINTSENSORPRIORITY);
    goalSensor->setPriority(GOALSENSORPRIORITY);
    
//   pickInterpolatorSensor = new SoFieldSensor(&ArticulatedBody::pickInterpolatorSensorCB, this);
//   pickInterpolatorSensor->attach( &pickInterpolator );
    interpolatorCache.setValue(new SoCube());
    ivkCache = new SbPList;

    jacobian = NULL;
    rootJoint = NULL;
    firstActiveJoint = NULL;
    lastJoint = NULL;
    myManip = (AnchorManip *) NULL;

}

 
ArticulatedBody::ArticulatedBody(SoTransform *anAnchor) : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(ArticulatedBody);
    SO_NODE_ADD_FIELD(goal, (0.0, 0.0, 0.0));
    SO_NODE_ADD_FIELD(tol, (0.01));
    SO_NODE_ADD_FIELD(angleHints, (0.0, 0.0, 0.0));
    SO_NODE_ADD_FIELD(angleHintStrengths, (0.0, 0.0, 0.0));
    SO_NODE_ADD_FIELD(interpolatorCache,(NULL));
    SO_NODE_ADD_FIELD(pickInterpolator, (-1));
    SO_NODE_ADD_FIELD(useCache, (TRUE));
    
    goalSensor = new SoFieldSensor(&ArticulatedBody::goalSensorCB, this);
    goalSensor->attach( &goal );

    angleHintsSensor = new SoFieldSensor(&ArticulatedBody::hintsSensorCB, this);
    angleHintsSensor->attach( &angleHints );
    
    gainHintsSensor = new SoFieldSensor(&ArticulatedBody::hintsSensorCB, this);
    gainHintsSensor->attach( &angleHintStrengths );
    
    gainHintsSensor->setPriority(STRENGTHSENSORPRIORITY);
    angleHintsSensor->setPriority(HINTSENSORPRIORITY);
    goalSensor->setPriority(GOALSENSORPRIORITY);
    
//   pickInterpolatorSensor = new SoFieldSensor(&ArticulatedBody::pickInterpolatorSensorCB, this);
//   pickInterpolatorSensor->attach( &pickInterpolator );
    interpolatorCache.setValue(new SoCube());
    ivkCache = new SbPList;

    SoTransform *anchor = (anAnchor != NULL) ? anAnchor : new SoTransform;
    addChild(anchor);
    addChild(new SoSeparator);  // place holder until a real joint is added

    jacobian = NULL;
    rootJoint = NULL;
    firstActiveJoint = NULL;
    lastJoint = NULL;

   myManip = (AnchorManip *) NULL;
}


int 
ArticulatedBody::getNumJoints()
{
   if (rootJoint) {
	return rootJoint->numJointsToEnd();
   } else {
	return 0;
   }
}

ArticulatedBody::~ArticulatedBody()
{
   if (goalSensor)
      delete goalSensor;
   if (jacobian)
      delete jacobian;
}


SbBool	
#ifdef SO_VERSION_REVISION
ArticulatedBody::readInstance(SoInput *in, unsigned short flags) 
#else
ArticulatedBody::readInstance(SoInput *in) 
#endif
{
    /* detach sensors */
    goalSensor->detach();
    angleHintsSensor->detach();
    gainHintsSensor->detach();
    //pickInterpolatorSensor->detach();
    
#ifdef SO_VERSION_REVISION
    if (SoSeparator::readInstance(in, flags)) {
#else
    if (SoSeparator::readInstance(in)) {
#endif
	/* now re-attach them */
	goalSensor->attach( &goal );
	angleHintsSensor->attach( &angleHints );
	gainHintsSensor->attach( &angleHintStrengths );
	//pickInterpolatorSensor->attach( &pickInterpolator );

        rootJoint = (JointModel *)getChild(1);
        setFirstActiveJoint(rootJoint);
        lastJoint = rootJoint->findLastJoint();

//	goalSensorCB(this, NULL);

	return(1);
   } else {
	return(0);
   }
}

void
ArticulatedBody::addVisual()
{
   if (getNumChildren() > 2)
	return;
   SoSeparator *cubeRoot  = new SoSeparator;
   SoDrawStyle *cubeStyle = new SoDrawStyle;
   SoMaterial  *cubeMat   = new SoMaterial;
   SoTransform *cubeTrans = new SoTransform;
   addChild(cubeRoot);

   cubeRoot->addChild(cubeMat);
   cubeRoot->addChild(cubeTrans);
   cubeRoot->addChild(new SoCube);
   cubeRoot->addChild(cubeStyle);

   cubeStyle->style.setValue(SoDrawStyle::LINES);
   cubeStyle->lineWidth.setValue(5.0);

   cubeMat->diffuseColor.setValue(.3, 0.7, .3);
   cubeTrans->scaleFactor.setValue(.8, .8, .8);
}

void
ArticulatedBody::removeVisual()
{
   if (getNumChildren() < 3)
	return;
   removeChild(2);
}

void
ArticulatedBody::addAllVisuals()
{
   addVisual();

   JointModel *j = rootJoint;
    
   while(j){
	j->addVisual();
	j = j->getNextJoint();
   }
}

void
ArticulatedBody::removeAllVisuals()
{
   removeVisual();

   JointModel *j = rootJoint;
    
   while(j){
	j->removeVisual();
	j = j->getNextJoint();
   }
}


void
ArticulatedBody::addJointToEnd(JointModel *j)
{
	if (rootJoint != NULL) {
		getLastJoint()->setNextJoint(j);
	  	setFirstActiveJoint(rootJoint);
	} else  {
	  	setFirstActiveJoint(j);
		replaceChild(1, rootJoint);
	}
}


void ArticulatedBody::attachAnchorAndJacobian(JointModel *newroot, 
    JointModel *firstActive)
{
    rootJoint = newroot;
    lastJoint = rootJoint->findLastJoint();

    replaceChild(1, rootJoint);

    setFirstActiveJoint(firstActive);
    
    
}

void ArticulatedBody::setFirstActiveJoint(JointModel *firstActive)
{
    firstActiveJoint = rootJoint = firstActive;
    lastJoint = rootJoint->findLastJoint();
    if (jacobian != NULL) 
	delete jacobian;
    jacobian = new Jacobian(firstActive->numJointsToEnd());
    
    /* build path to tip */
    path = buildPath(rootJoint);

}

SoPath *ArticulatedBody::buildPath(SoNode *startNode)
{
    SoPath *pp;
    
    /* build path to tip */
    SoSearchAction sa;
    SoRotation *rot = lastJoint->getOuter();
    sa.setNode(rot);
    sa.apply(startNode);
    pp = sa.getPath();
    pp->ref();
    if(pp==NULL)
        printf("path wasn't created \n");
//    else
//        printf("path created \n");
    return pp;
}

Vector ArticulatedBody::saveJointAngles()
{   
    JointModel *nj = rootJoint;
    Vector sja(rootJoint->numJointsToEnd());
    int i =0;
    
    while(nj){
	sja.mm[i++] = nj->angle.getValue();
	nj = nj->getNextJoint();
    }
    return sja;
}

void ArticulatedBody::restoreJointAngles(const Vector &sja)
{   
    JointModel *nj = rootJoint;
    int i =0;
    
    while(nj){
	//nj->angle.setValue(sja.mm[i++]);
	nj->setAngleImmediate(sja.mm[i++]);
	nj = nj->getNextJoint();
    }
    return;
}

void
ArticulatedBody::recenterJointLimb(JointModel *j)
{

    SoSearchAction sa;

    sa.setNode(j->getInner());
    sa.apply(this);
    SoPath *p = sa.getPath();

    SbViewportRegion vr;
    SoGetMatrixAction ma(vr);
    SbMatrix m;
    
    ma.apply(p);
    m = ma.getMatrix();
    m = m.inverse();
    j->getLimb()->getTransform()->setMatrix(m);
}

void ArticulatedBody::vecToBase(SbVec3f &ep)
{
   if(!path)
	path = buildPath(rootJoint);
   vecToBaseUsing(ep, path);
}

void ArticulatedBody::vecToEnd(SbVec3f &ep)
{
    if(!path)
	path = buildPath(rootJoint);
    vecToEndUsing(ep, path);
}

void ArticulatedBody::vecToBaseUsing(SbVec3f &ep, SoPath *p)
{
    SbViewportRegion vr;
    static SoGetMatrixAction ma(vr);
    SbMatrix m;
    
    if(!p)
	printf("INVALID PATH \n");
    else {
	ma.apply(p);
	m = ma.getMatrix();
	m.multVecMatrix(ep, ep);
    }
}

void ArticulatedBody::vecToEndUsing(SbVec3f &ep, SoPath *p)
{
    SbViewportRegion vr;
    static SoGetMatrixAction ma(vr);
    SbMatrix m;
    if(!p)
	printf("INVALID PATH \n");
    else {   
	ma.apply(p);
	m = ma.getInverse();
	m.multVecMatrix(ep, ep);
    }
}


void ArticulatedBody::getEndPoint(SbVec3f &ep)
{
    ep.setValue(0, 0, 0);
    vecToBase(ep);
}

float ArticulatedBody::moveEndPointTo(SbVec3f *nep, float testTol)
{
    goal.setValue(*nep);
    return _moveEndPointTo(nep,NUMKITER, testTol);
}

float  ArticulatedBody::_moveEndPointTo(SbVec3f *nep, int iter,  float testTol)
{
    Spv cartVel;
    SbVec3f dist, target, result;
    Vector jointAngles;
    int succeeded = 1;

    // assume nep is in body coordinates
    dist = *nep;
    target = dist;

    // this should give delta in endpoint frame
    vecToEnd(dist);
    cartVel.v0 = dist;

    // calculate error before inverse kinematics
    float perr = dist.length();
    jointAngles = saveJointAngles();
    
    // do inverse kinematics
    jacobian->inverseKinematics(firstActiveJoint, &cartVel);

    // get error after inverse kinematics
    getEndPoint(result);
    float err = (target-result).length();
    
    if((testTol != ACCEPT) 
	&& (iter) 
	&& (err>testTol)
	/* && (dist.length()>testTol) */){
        err =  _moveEndPointTo(nep,iter-1,  testTol);
    }
    // You get here either because it found a solution or it
    // recursed until it reached its iteration limit. err now 
    // represents the minimum error solution from here to the
    // end of the recursion. If the minimum err solution exceeds
    // perr, then restore the jointAngles on the way back.
    if(err>perr){
	 restoreJointAngles(jointAngles);
	 err = perr;
//	printf("restoring joint angles \n");
    }
//    if(iter ==NUMITER){
//	cartVel.v0 = SbVec3f(0,0,0);
//	jacobian->inverseKinematics(firstActiveJoint, &cartVel); 
//    }
	return err;  
}

void ArticulatedBody::setAnchorPosition(SbVec3f np)
{
    getAnchor()->translation.setValue(np);
}

void ArticulatedBody::setPickInterpolator(int which)
{
    pickInterpolator.setValue(which);
    ArticulatedBody::pickInterpolatorSensorCB(this, NULL);
}

void ArticulatedBody::initCenterAngleAndGain(float g)
{
    JointModel *j = rootJoint;
    while(j){
	j->center.setValue(j->angle.getValue());
	j->centerGain.setValue(g);
	j = j->getNextJoint();
    }
}


int ArticulatedBody::checkCache(SbVec3f ng, float cache_tol)
{
    int i;
    int num = ivkCache->getLength();
    SbPList currCache = *ivkCache;
    ABCache *abc;
    
    for(i=0;i<num;i++){
	abc = (ABCache *)currCache[i];
	if((abc->goal- ng).length()<cache_tol){
	    //printf("using cache \n");
	    restoreJointAngles(abc->jointData);
	    break;
	}
    }
    return (i<num)?1:0;
}

void  ArticulatedBody::addToCache(SbVec3f ng)
{
    if(useCache.getValue()){
//	printf("adding to cache \n");
	ABCache *abc = new ABCache(ng, this);
	ivkCache->append(abc);
    }
} 

ArticulatedBody *
ArticulatedBody::isSelection(SoPath *p)
{
   if (p->getNodeFromTail(2)->isOfType(ArticulatedBody::getClassTypeId())) {
	p->pop();
	p->pop();
	return (ArticulatedBody *)p->getTail();
   } else {
	return (ArticulatedBody *) NULL;
   }
}

void
ArticulatedBody::addManipulator(SoPath *p)
{
   if (myManip != NULL)
	return;

   myManip = new AnchorManip;

   p->push(getAnchorIndex());
   manipPath = p;

   myManip->replaceNode(p);

   SoDragger *myDragger;
   myDragger = myManip->getDragger();
   myDragger->addMotionCallback(&ArticulatedBody::dragMotionCallback, this);
   myDragger->addStartCallback(&ArticulatedBody::dragStartCallback, this);
}

void
ArticulatedBody::removeManipulator()
{
   if (myManip == NULL)
	return;

   myManip->replaceManip(manipPath,NULL);
   myManip = (AnchorManip *) NULL;
}


void
ArticulatedBody::dragMotionCallback(void *data, SoDragger *)
{
   ArticulatedBody *t = (ArticulatedBody *)data;

   if (t->rootJoint) {

       SbViewportRegion vr;
       SoGetMatrixAction ma(vr);
       SbMatrix c;
    
       ma.apply(t->getAnchor());
       c = ma.getMatrix();
       c = c.inverse();
       c = c.multLeft(t->saveMat);

       t->rootJoint->getInner()->setMatrix(c);
   }

}

void
ArticulatedBody::dragStartCallback(void *data, SoDragger *)
{
   ArticulatedBody *t = (ArticulatedBody *)data;


    SoSearchAction sa;

    if (t->rootJoint) {
	
        sa.setNode(t->rootJoint->getInner());
        sa.apply(t);
        SoPath *p = sa.getPath();

        SbViewportRegion vr;
        SoGetMatrixAction ma(vr);
        ma.apply(p);
        t->saveMat = ma.getMatrix();
    }
}

void ArticulatedBody::moveToConstrainedXYGoal(SbVec3f &pt)
{
    float maxExtension=0;
    lastJoint->getMaxExtension(maxExtension);
    if(pt.length() > (.95 * maxExtension)){
	float ang = atan2(pt[1], pt[0]);
	pt[0] = cos(ang)*maxExtension*.95;
	pt[1] = sin(ang)*maxExtension*.95;
	pt[2] = 0.0;
    }
//	printf("movetogoal %f %f \n",pt[0],pt[1],pt[3]);
    goal.setValue(pt);
    moveToGoalImmediate();
}

void ArticulatedBody::moveToGoalImmediate()
{
    if(checkCache(goal.getValue(), tol.getValue())){
	gainHintsSensor->unschedule();
	angleHintsSensor->unschedule();
	goalSensor->unschedule();
    }
    else {
	/* synch up these silly sensors */
	if(gainHintsSensor->isScheduled()){
	    ArticulatedBody::hintsSensorCB(this, gainHintsSensor);
	    gainHintsSensor->unschedule();
	}
	    
	if(angleHintsSensor->isScheduled()){
	    ArticulatedBody::hintsSensorCB(this, angleHintsSensor);
	    angleHintsSensor->unschedule();
	}
	    
	if(goalSensor->isScheduled()){
	    ArticulatedBody::goalSensorCB(this, NULL);
	    goalSensor->unschedule();
	}
    }

}

void
ArticulatedBody::goalSensorCB(void *in, SoSensor *)
{
    ArticulatedBody *a = (ArticulatedBody *)in;
    SbVec3f v = a->goal.getValue();
 //   SbVec3f v1 = v;
//	printf("distance -> %f \n",v1.length());
    if(!a->checkCache(v, a->tol.getValue())){
	a->_moveEndPointTo(&v, NUMKITER, a->tol.getValue());
	a->addToCache(v);
    }
    SbVec3f ep;
//    a->getEndPoint(ep);
//	if((ep-v).length()>.001)
//    	printf("ERROR -> %f \n", (ep-v).length());
}

void
ArticulatedBody::pickInterpolatorSensorCB(void *in, SoSensor *)
{
    ArticulatedBody *a = (ArticulatedBody *)in;
    int whichInterpolator = (int)a->pickInterpolator.getValue();
    EngineHolder *eh;
    if(whichInterpolator <= 9){
	eh = (EngineHolder *)a->interpolatorCache.getValue();
	eh->connectEngineTo(whichInterpolator, (SoMFVec3f *)&a->goal);
	eh->connectEngineTo(whichInterpolator+1, &a->angleHints);
	eh->connectEngineTo(whichInterpolator+2, &a->angleHintStrengths);
	a->ivkCache = eh->getJointCache(whichInterpolator/3);
	a->tol = eh->getTolerance(whichInterpolator/3);
    }
}

void
ArticulatedBody::hintsSensorCB(void *in, SoSensor *whichSensor)
{
    ArticulatedBody *a = (ArticulatedBody *)in;
    SbVec3f v = a->goal.getValue();
    const SbVec3f *theData = (whichSensor == a->angleHintsSensor)?
	    a->angleHints.getValues(0) : a->angleHintStrengths.getValues(0);

    int i=0, k=0;
    JointModel *j = a->rootJoint;
    while(j->getNextJoint()){
	if(whichSensor == a->angleHintsSensor)
	    j->center.setValue(theData[i/3][k]);
	else
	    j->centerGain.setValue(theData[i/3][k]);
	j = j->getNextJoint();
	/* increment i and k */
	k = (++k)%3;
	i++;
    }
}

ABCache::ABCache(SbVec3f g, ArticulatedBody *a)
{
    goal = g;
    jointData = a->saveJointAngles();
}

