#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/engines/SoConcatenate.h>

#include "BSInterpV3f.h"
#include "ControlPoint.h"

SO_ENGINE_SOURCE(BSInterpV3f);

void
BSInterpV3f::initClass()
{
   SO_ENGINE_INIT_CLASS(BSInterpV3f, SoInterpolate, "Interpolate");
}


BSInterpV3f::BSInterpV3f() : SoInterpolate()
{
   SO_ENGINE_CONSTRUCTOR(BSInterpV3f);

   SO_ENGINE_ADD_INPUT(alpha,   (0));
   SO_ENGINE_ADD_INPUT(degree,  (3));
   SO_ENGINE_ADD_INPUT(controlPoints,  (0.0, 0.0, 0.0));
   SO_ENGINE_ADD_INPUT(geomSep,  (NULL));
   SO_ENGINE_ADD_INPUT(transform, (
    1, 0, 0, 0, 
    0, 1, 0, 0, 
    0, 0, 1, 0, 
    0, 0, 0, 1));

   SO_ENGINE_ADD_OUTPUT(output, SoSFVec3f);

   mySpline = new BSpline(3, 4);
   inputSep = (SoSeparator *)NULL;
   curveSep = (SoSeparator *)NULL;
}

BSInterpV3f::~BSInterpV3f()
{
  delete mySpline;
}

void 
BSInterpV3f::addEditables()
{
   ControlPoint *cp;
   int n = controlPoints.getNum();

   for(int i=0; i<n; i++) {
	cp = new ControlPoint(controlPoints[i], i, &controlPoints);
	curveSep->addChild(cp);
   }
}

void
BSInterpV3f::transformControlPts(SbVec3f *transformedPts)
{
    int n = controlPoints.getNum();
    int i;
    SbVec3f apt;
    if(n){
        SbMatrix m = transform.getValue();
        for(i=0;i<n;i++){
            apt = controlPoints[i];
            m.multVecMatrix(controlPoints[i], transformedPts[i]);
        }
    }
}

void 
BSInterpV3f::inputChanged(SoField *whichField)
{
  SbVec3f points[64];

  if (whichField == &controlPoints) {
	int n = controlPoints.getNum();
	if (n != mySpline->getNumControlPts()) {

	    delete mySpline;
	    mySpline = new BSpline((int)degree.getValue(), n);

	    if (geomSep.getValue() != NULL) {
	        inputSep->removeChild(curveSep);
	        curveSep = mySpline->generateCurve();
	        inputSep->addChild(curveSep);
	        addEditables();
	    }

	} else if (geomSep.getValue() != NULL)
	        mySpline->updateCurve();
	
	transformControlPts(points);
	mySpline->setControlPts(points);

  } else if (whichField == &degree) {
	int n = controlPoints.getNum();
	delete mySpline;
	mySpline = new BSpline((int)degree.getValue(), n);

	transformControlPts(points);
	mySpline->setControlPts(points);

	if (geomSep.getValue() != NULL) {
	    inputSep->removeChild(curveSep);
	    curveSep = mySpline->generateCurve();
	    inputSep->addChild(curveSep);
	    addEditables();
	}

  } else if(whichField == &transform){
	transformControlPts(points);
	if(mySpline)
	    mySpline->setControlPts(points);

  } else if (whichField == &geomSep) {

	if (geomSep.getValue()) {

	    if (geomSep.getValue() == inputSep) 
		return;

            if (!(geomSep.getValue()->isOfType(SoSeparator::getClassTypeId()))) {
		fprintf(stderr, "BSInterpV3f: geomSep must be a Separator\n");
		return;
	    }

	    inputSep = (SoSeparator *)geomSep.getValue();
	    inputSep->ref();

	    curveSep = mySpline->generateCurve();
	    inputSep->addChild(curveSep);
	    addEditables();

	} else {

	    if (inputSep) {
		inputSep->removeChild(curveSep);
		inputSep->unref();
	    }

	}
  }
}

void 
BSInterpV3f::pointOnCurve(float t,SbVec3f &vec)
{
    float tt = t * (mySpline->getNumControlPts() - degree.getValue());
    mySpline->pointOnCurve(tt, vec);
}

void 
BSInterpV3f::deltaOnCurve(float t1,float t2,SbVec3f &vec)
{
    SbVec3f v1,v2;
    pointOnCurve(t1, v1);
    pointOnCurve(t2, v2);
    vec = v2 - v1;
}

void
BSInterpV3f::evaluate()
{
   SbVec3f vec;
   float t = alpha.getValue() * (mySpline->getNumControlPts() - degree.getValue());

   mySpline->pointOnCurve(t, vec);
   SO_ENGINE_OUTPUT(output, SoSFVec3f, setValue(vec));
}


