#include "BSpline.h"
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoNurbsCurve.h>
#include <Inventor/nodes/SoLineSet.h>

float deboor(int j,int k,Vector knots,float t)
{
    float a,b,c,d;
    int i = j-1;
    if(k==1)
	return ((knots.mm[i] <= t)&&(t < knots.mm[i+1]))?1 :0;
    else{
	a = (t - knots.mm[i]) *deboor(j,k-1,knots,t);
	b = knots.mm[i + k -1] - knots.mm[i];
	a = (b)? a/b : 0.0;
	c = (knots.mm[i+k] -t) * deboor(j+1,k-1,knots,t);
	d = knots.mm[i+k] - knots.mm[i+1];
	c = (d)? c/d : 0.0;
	return a + c;
    }
}

int BSpline::pointOnCurve(float t, SbVec3f &p,  SbBool ifContinuous)
{
    if( ifContinuous){
	int i;
	int kn = order + numControlPts;
	float d;
    
	p.setValue(0, 0, 0);
	if((t<knots.mm[0]) || (t> knots.mm[kn-1])){
	    printf("t outside knot range %f \n", t);
	    return -1;
	}
	else if((order<2)||(order>numControlPts)){
	    printf("order too great for number of points \n");
	    return -1;
	}
	else if(t == knots.mm[kn-1])
	    p = controlPts[numControlPts-1];
	
	for(i=0;i<numControlPts;i++){
	    d = deboor(i+1,order,knots,t);
	    p += controlPts[i]*d;
	}
	return 1;
    }
    else {
	if(ft){
	    cacheVertices();
	    ft = 0;
	}
	float tt = t /(numControlPts - (order-1));
	float vi  = (tt*31);
	int vindex = ((vi- floor(vi))<(ceil(vi)-vi))?floor(vi) : ceil(vi);
	p = vertices[vindex];
	return 1;
    }
}

BSpline::BSpline(int degree, int nControlPts)
{
    int i;
    
    order = 1+degree;
    numControlPts = nControlPts;
    controlPts = new SbVec3f[numControlPts];
    numKnots = order + numControlPts;
    knots = Vector(numKnots);
    ft = 1;
    for(i=0;i<numKnots;i++){
	if(i<order)
	    knots.mm[i] = 0;
	else if(i<numControlPts)
	    knots.mm[i] = i-order+1;
	else
	    knots.mm[i] = numControlPts-order + 1;
    }
    //knots.print("knot vector");

   coords = (SoCoordinate3 *)NULL;
}


BSpline::~BSpline()
{
   delete controlPts;
}

void BSpline::cacheVertices()
{
    float maxValue = knots.mm[numKnots-1];
    float t;
    int i=0;
    
    /* calculate vertices */
    for(i = 0, t=0;i<32;i++,t += maxValue/32.0){
	pointOnCurve(t, vertices[i], TRUE);
    }
}
    
void BSpline::setControlPts(const SbVec3f *data)
{
    int i;
    for(i=0;i<numControlPts;i++)
	controlPts[i].setValue(data[i][0],data[i][1], data[i][2]) ;
//    cacheVertices();
}


SoSeparator *BSpline::generateCurve()
{

    if (coords)   // already have generated a curve
	return (SoSeparator *)NULL;

    coords = new SoCoordinate3;
    SoSeparator *sep = new SoSeparator;
    SoComplexity *comp = new SoComplexity;
    SoNurbsCurve *curve = new SoNurbsCurve;

    SoMaterial *mat = new SoMaterial;
    mat->diffuseColor.setValue(.5, .8, .5);
    mat->ambientColor.setValue(.5, .8, .5);

    comp->value = .8;
    coords->point.setValues(0, numControlPts, controlPts);
    curve->numControlPoints = numControlPts;
    curve->knotVector.setValues(0, numKnots, knots.mm);
    
    sep->addChild(comp);
    sep->addChild(mat);
    sep->addChild(coords);
    sep->addChild(curve);

    return sep;
}

void
BSpline::updateCurve()
{
    if (coords)
        coords->point.setValues(0, numControlPts, controlPts);
}

#if 0

SoSeparator *BSpline::generateMyCurve()
{
    SoSeparator *sep = new SoSeparator;
    SoBaseColor *bc = new SoBaseColor;
    SoCoordinate3 *coords = new SoCoordinate3;
    SoLineSet *curve = new SoLineSet;
    int i=0;
    float maxValue = knots.mm[numKnots-1];
    float t;
    
    
    /* calculate vertices */
    for(t=0;t<=maxValue;t += maxValue/31.0){
	pointOnCurve(t, vertices[i++], TRUE);
    }
    fitToData(32, &vertices[0], 0);
    bc->rgb.setValue(1, 0, 0);
    coords->point.setValues(0, 32, vertices);
    curve->numVertices = 32;
    curve->startIndex = 0;
    
    sep->ref();
    sep->addChild(bc);
    sep->addChild(coords);
    sep->addChild(curve);
    sep->unrefNoDelete();
    return sep;    
}
#endif

void BSpline::fitToData(int numpts, SbVec3f *pts, int useChordLength)
{
    int i, j, k;
    /* build matrix of dim -> numDataPts X numControlPts */
    Matrix N(numpts, numControlPts);
    
    /* now for x,y and z */
    Vector data(numpts);
    Vector b(numControlPts);
    Vector chordLengths(numpts);
    
    if(useChordLength){
	for(i=1;i<numpts;i++){
	    chordLengths.mm[i] = sqrt((pts[i]-pts[i-1]).dot(pts[i]-pts[i-1]));
	}    
	/* figure out chord lengths and parameterize */
	chordLengths.mm[0] = 0;
	float chordLenSum = chordLengths.sum();
	float cumChordLen = 0;
	for(i=1;i<data.m;i++){
	    cumChordLen += fabs(chordLengths.mm[i]);
	    chordLengths.mm[i] = (cumChordLen/chordLenSum)*knots.mm[numKnots-1];
	}
    }
    else {
	for(i=1;i<numpts;i++){
	    chordLengths.mm[i] = (i/(float)(numpts-1))*knots.mm[numKnots-1];
	}   
    }
    //chordLengths.print("parameterized chord lengths \n");
    
    /* ok now fill in matrix */
    for(j = 0;j<numpts;j++)
	for(i=0;i<numControlPts;i++)
	    N.mm[j][i] = deboor(i+1, order, knots, chordLengths.mm[j]);
    N.mm[numpts-1][numControlPts-1] = 1;
    
    //N.print("N");
    /* since it won't be square in general do least squares solution */
    Matrix NtNiNt = ((N.transpose()*N).inverse())*N.transpose();
	
    for(k=0;k<3;k++){
	for(i=0;i<numpts;i++){
	    data.mm[i] = pts[i][k];
	}    
	b = NtNiNt * data;
	for(i=0;i<numControlPts;i++){
	   controlPts[i][k] = b.mm[i];
	}
    }    
}


