#ifndef _BASE_CREATURE_
#define _BASE_CREATURE_

#include <Inventor/SbBasic.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/SbLinear.h>
#include <Inventor/Xt/SoXtRenderArea.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/SbPList.h>
#include <Inventor/SoPath.h>
#include "SpatialAlgebra.h"
#include <Inventor/SoLists.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFString.h>
#include "InventorXDRRoutines.h"

class SoSeparator;
class ArticulatedBody;
class JointModel;
class GenericController;
class CreatureCache;
class TimeDependentComponent;
/*********************************************************************
 * 
 *	    UpdateStateData  Class
 * 
 *********************************************************************/
class UpdateStateData {
    public:
    float time;
    SoNode *currSceneGraph;
    SoXtRenderArea *currViewer;
    CreatureCache *currCreatureCache;
};



/*********************************************************************
 * 
 *	    CreatureCBData  Class
 * 
 *********************************************************************/
class BaseCreature;
typedef void (*creatureCB)(BaseCreature *creature);

class CreatureCBData 
{
public:
  CreatureCBData(const char *name,creatureCB callback)
  :fname(name) {cb = callback;}
  
  creatureCB getCB() {return cb;}
  const SbString &getFname() {return fname;}
  
private:
  SbString fname;
  creatureCB cb;
};


/*********************************************************************
 * 
 *	    BaseCreature  Class
 * 
 *********************************************************************/
class CreatureCache; 
class SoSwitch;


class BaseCreature : public SoSeparator {

    SO_NODE_HEADER(BaseCreature);

public:
    SoSFInt32 tag;
    SoSFInt32 realTag;
    SoSFInt32 registryNumber;
    SoSFString geometryFileName;
    SoSFInt32 defaultSwitchValue;
    SoSFBool isMoving;
    SoSFBool isGrabbed;
    SoPath *bodyPath;
    SbPList updateComponents;
    SbPList repList;
    float heading;
    SoSwitch *switchNode;
    CreatureCache *myCache;
    BaseCreature *grabbedBy;
    
    // method called to change representation
    virtual void changeRep(int newRep);
    virtual void updateRepsAt(UpdateStateData &usd);
    
    // Callback registry service.  hardts 3/9/95
    SoSFString updateCBName; // Called every turn.
    creatureCB updateCB;
    SoSFString isGrabbedCBName; // Called when isGrabbed changes
    creatureCB isGrabbedCB;
    SoSFString resetCBName; // Called for resetting state
    creatureCB resetCB;
    static SbPList cbList;
    
    static void addCallback(const char *name,creatureCB cb) {
      cbList.append(new CreatureCBData(name,cb)); 
    }
    static creatureCB getCreatureCBByName(const SbString &);

    static void isGrabbedSensorCB(void *creature,SoSensor *);
    
    static void initClass();
    static BaseCreature *loadCreature(char *filename);
    
    BaseCreature();

    ArticulatedBody *getArticulatedBody(JointModel *j);
    
    /* Inventor specific methods  */
    SbBool bodyInPath(SoPath *path);
    SoSeparator *getBody();
    SoTransform *getBodyTransform();
    SoPath *getBodyPath(SoNode *root);
    SoSwitch *getSwitch() {return switchNode;}
    void setVisible(int yn);
    virtual void setIsGrabbed(SbBool soIsIt);
    virtual void setIsGrabbed(SbBool soIsIt, BaseCreature *byWhom);
    
    SoNode *findFirstInstanceOf(SoType nodeType);
    
    SoPath *getPathToName(SoNode *from, const SbName &name);
    
    /* methods for setting position and orientation */
    void setPosition(SbVec3f &np);
    void getPosition(SbVec3f &np);
    const SbVec3f &position();
    void setHeading(float t);
    void setHeading(SbRotation rot);
    float getHeading();
    float getHeading(SbVec3f &axis, float &ang);
    void getHeading(SbVec3f &axis);
    
    /* methods for geometric operations including transformations */
    void vecToBody(SbVec3f &pt);
    void vecToWorld(SbVec3f &pt);
    void dirToBody(SbVec3f &pt);
    void dirToWorld(SbVec3f &pt);
    void calcBearingAndRange(SbVec3f pt, float &bearing, float &range);
    void calcBoundingBox(SbBox3f &wbox);
    void calcCenter(SbVec3f &center);
    virtual int testRay(SbLine &theRay, SbVec3f &thePt);
    
    TimeDependentComponent *getFirstTDCOfType(SoType atype);
    TimeDependentComponent *getFirstTDCWithName(SbName &aName);
    
    /* methods used to implement time-step updates*/
    virtual void updateStateAt(UpdateStateData &usd);
    virtual int stepTo(UpdateStateData &usd);
    virtual void restart();

    /* Method to allow synchronization of droids to their masters in Distributed ALIVE.
       Over-ride this method to include subclass-specific stuff. I recommend that you call
       the parent's function at the beginning of your over-riding function (i.e.
       if (!BaseCreature::syncToOrFromStream()) return FALSE;).
       */

    virtual bool_t syncToOrFromStream(XDR *xdrs);

protected:
    virtual ~BaseCreature();
#ifdef SO_VERSION_REVISION
    virtual SbBool  readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool  readInstance(SoInput *in);
#endif
    SoGetBoundingBoxAction *bba; // cached for performance reasons
};

#endif
