#include <math.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/draggers/SoDragger.h>
#include <Inventor/SoPath.h>

#include "JointModel.h"
#include "ArticulatedBody.h"
#include "Vector.h"
#include "lu.h"
#include "JointManip.h"

#define DEBUGPRINT(x, y)if(debug)x.print(y);
static int debug = 0;
#define LINEAR_ONLY 1
/*********************************************************************
 * 
 *	    JointModel Class
 *
 *  NOTE: Field can change and not affect anything. For now.
 * 
 *********************************************************************/


int JointModel::inner_index     = 0;
int JointModel::outer_index     = 1;
int JointModel::limb_index      = 2;
int JointModel::nextJoint_index = 3;


SO_NODE_SOURCE(JointModel);

void
JointModel::initClass()
{
    SO_NODE_INIT_CLASS(JointModel, SoSeparator, "Separator");
}

JointModel::JointModel() : SoSeparator(4)
{
    SO_NODE_CONSTRUCTOR(JointModel);
    SO_NODE_ADD_FIELD(center, (0.0));
    SO_NODE_ADD_FIELD(centerGain, (0.0));
    SO_NODE_ADD_FIELD(angle, (0.0));
    SO_NODE_ADD_FIELD(jointAxis, (0, 0, 1));
    SO_NODE_ADD_FIELD(shouldNotify, (FALSE));
    angleSensor = new SoFieldSensor(&JointModel::angleSensorCB, this);
    angleSensor->attach( &angle );
    angleSensor->setPriority(JOINTSENSORPRIORITY);
    prevJoint = NULL;
   myManip = (JointManip *) NULL; 
}

JointModel::JointModel(LimbModel *l, SoTransform *new_inner) : SoSeparator(4)
{
    SO_NODE_CONSTRUCTOR(JointModel);
    SO_NODE_ADD_FIELD(center, (0.0));
    SO_NODE_ADD_FIELD(centerGain, (0.0));
    SO_NODE_ADD_FIELD(angle, (0.0));

    angleSensor = new SoFieldSensor(&JointModel::angleSensorCB, this);
    angleSensor->setPriority(JOINTSENSORPRIORITY);
    angleSensor->attach( &angle );

    addChild(new_inner);
    addChild(new SoRotation());
    addChild((SoNode *)l);
    addChild(new SoSeparator);  // placeholder for the next joint

    prevJoint = NULL;
   myManip = (JointManip *) NULL;
}

JointModel::~JointModel()
{
}

SbBool
JointModel::isLastJoint()
{
  if (getChild(nextJoint_index)->isOfType(
	JointModel::getClassTypeId()))
     return FALSE;
  else 
     return TRUE;
}

void JointModel::getMaxExtension(float &maxExtension)
{
    if(isLastJoint()){
	maxExtension = getInner()->translation.getValue().length();
    }
    else {
	maxExtension += getInner()->translation.getValue().length();
    }
    if(prevJoint != NULL){
	prevJoint->getMaxExtension(maxExtension);
    }	
}
SbBool	
#ifdef SO_VERSION_REVISION
JointModel::readInstance(SoInput *in, unsigned short flags) 
#else
JointModel::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
   if (SoSeparator::readInstance(in, flags)) {
#else
   if (SoSeparator::readInstance(in)) {
#endif

	if (this->isLastJoint())
	   ;
	else 
	   getNextJoint()->prevJoint = this;
	SbRotation r = getOuter()->rotation.getValue();
	float a;
	SbVec3f xx;
	r.getValue(xx,a);
	if(xx[2]<0){
		angle.setValue(-a);
		center.setValue(-a);
		getOuter()->rotation.setValue(SbRotation(SbVec3f(0,0,1),-a));
	}
	enableNotify(shouldNotify.getValue());
	angleSensorCB(this, NULL);
	    
	return(1);
   } else {
	return(0);
   }
}

void JointModel::setInner(SoTransform *new_inner)
{
   replaceChild(inner_index, (SoNode *)new_inner);
}

void JointModel::setLimb(LimbModel *aLimb)
{
    replaceChild(limb_index, (SoNode *)aLimb);
}

void JointModel::setNextJoint(JointModel *aJoint)
{
    aJoint->prevJoint = this;
    replaceChild(nextJoint_index, (SoNode *)aJoint);
}

JointModel *
JointModel::getNextJoint() {

  if (isLastJoint()) 
	return NULL;
  else 
     	return (JointModel *)getChild(nextJoint_index);
}

void JointModel::setAngleImmediate(float newAng)
{
    if(newAng != angle.getValue()){
	angle.setValue(newAng);
	angleSensor->unschedule();
	JointModel::angleSensorCB(this,angleSensor);
    }
}

JointModel *JointModel::findLastJoint()
{
    JointModel *last = this;
    while(!(last->isLastJoint()))
	last = last->getNextJoint();
    return last;
}

void JointModel::setShouldNotify(SbBool yesOrNo)
{
    JointModel *last = this;
    while(!(last->isLastJoint())){
	last->shouldNotify.setValue(yesOrNo);
	last = last->getNextJoint();
    }
    last->shouldNotify.setValue(yesOrNo);
    return last;
}

int  JointModel::numJointsToEnd()
{
    JointModel *last = this;
    int num=1;

    while (!(last->isLastJoint())) {
	num++;
	last = last->getNextJoint();
    }
    return num;
}


void JointModel::xformSpv(Spv *arg)
{
    SbMatrix e;
    SbRotation rot;
    SbVec3f r;
    SbVec3f r1;
    SbVec3f r2;
    
    
    /* first do inner */
    r = getInner()->translation.getValue();
    rot = getInner()->rotation.getValue();
    rot.getValue(e);
    
    e.multMatrixVec(arg->v0 - r.cross(arg->v),arg->v0);
    e.multMatrixVec(arg->v,arg->v);
    
    /* now do outer (only a rotation ) */
    rot = getOuter()->rotation.getValue();
    rot.getValue(e);
    e.multMatrixVec(arg->v0 ,arg->v0);
    e.multMatrixVec(arg->v,arg->v);
}

void JointModel::computeJacobian(int i, Matrix &jac)
{
    int k, j;
    
    Spv tmp;
    // do previous columns to this frame
    for(j=0;j<i;j++){
	tmp = Spv(SbVec3f(jac.mm[0][j], jac.mm[1][j], jac.mm[2][j]), 
	    SbVec3f(jac.mm[3][j], jac.mm[4][j], jac.mm[5][j]));
	xformSpv(&tmp);
	for(k=0;k<3;k++){
	    jac.mm[k][j] = tmp.v[k];
	    jac.mm[k+3][j] = tmp.v0[k];
	}
    }
    //SbVec3f axis = SbVec3f(0, 0, 1);
    SbVec3f axis = jointAxis.getValue();
    // now add this column
    for(j=0;j<3;j++){
	jac.mm[j][i] = axis[j];
	jac.mm[j+3][i] = 0;
    }
    i++;
    if(i<jac.n)
	getNextJoint()->computeJacobian(i, jac);
}

void
JointModel::angleSensorCB(void *in, SoSensor *)
{
    JointModel *j = (JointModel *)in;
    float newAng = j->angle.getValue();
/*
    if(newAng > M_PI)
	newAng -= 2*M_PI;
    else if(newAng < -M_PI)
	newAng += 2*M_PI;
*/    
    j->getOuter()->rotation.setValue(j->jointAxis.getValue(), newAng);
}


void
JointModel::addVisual()
{
   if (getNumChildren() > 4)
	return;

   SoSeparator *visualRoot = new SoSeparator;

   SoSeparator *limbRoot = new SoSeparator;
     SoMaterial *limbMat = new SoMaterial;
     limbMat->diffuseColor.setValue(.5, .5, .5);
     limbRoot->addChild(limbMat);

     SoTransform *limbTrans = new SoTransform;
     limbTrans->scaleFactor.setValue(1.0, 2.0, 1.0);
     limbTrans->translation.setValue(1.0, 0.0, 0.0);
     limbTrans->rotation.setValue(SbRotation(SbVec3f(0,0, 1), M_PI/2));
     limbRoot->addChild(limbTrans);

     SoCylinder *limbCyl = new SoCylinder;
     limbCyl->height.setValue(1.0);
     limbCyl->radius.setValue(0.1);
     limbRoot->addChild(limbCyl);
   visualRoot->addChild(limbRoot);

   SoSeparator *sphereRoot = new SoSeparator;
     SoMaterial *sphereMat = new SoMaterial;
     sphereMat->diffuseColor.setValue(.6, 1.0, .6);
     sphereRoot->addChild(sphereMat);

     SoSphere *s = new SoSphere;
     s->radius = 0.12;
     sphereRoot->addChild(s);
   visualRoot->addChild(sphereRoot);

   SoSeparator *rodRoot = new SoSeparator;
     SoMaterial *rodMat = new SoMaterial;
     rodMat->diffuseColor.setValue(.6, .6, 1.0);
     rodRoot->addChild(rodMat);

     SoTransform *rodTrans = new SoTransform;
     rodTrans->scaleFactor.setValue(1.0, 3.0, 1.0);
     rodTrans->rotation.setValue(SbRotation(SbVec3f(1, 0, 0), M_PI/2));
     rodRoot->addChild(rodTrans);

     SoCylinder *rodCyl = new SoCylinder;
     rodCyl->height.setValue(1.0);
     rodCyl->radius.setValue(0.05);
     rodRoot->addChild(rodCyl);
   visualRoot->addChild(rodRoot);

   addChild(visualRoot);

   updateVisual();
}

void
JointModel::updateVisual()
{
   if (getNumChildren() < 5)
	return;

   if (isLastJoint())
	return;

   SoSeparator *s  = (SoSeparator *)getChild(4);
	        s  = (SoSeparator *)s->getChild(0);
   SoTransform *lt = (SoTransform *)s->getChild(1);

   SbVec3f     pnt = getNextJoint()->getInner()->translation.getValue();
   float       len = pnt.length();
   float       x, y, z;

   pnt.getValue(x, y, z);
   lt->scaleFactor.setValue(1.0, len, 1.0);
   lt->translation.setValue(x/2.0, y/2.0, z/2.0);

   x /= len;   y /= len;   z /= len;
   lt->rotation.setValue(SbVec3f(z, 0, -x), acosf(y));
}

void
JointModel::removeVisual()
{
   if (getNumChildren() < 5)
	return;
   removeChild(4);
}


JointModel *
JointModel::isSelection(SoPath *p)
{
   if (p->getNodeFromTail(3)->isOfType(JointModel::getClassTypeId())) {
	p->pop();
	p->pop();
	p->pop();
	return (JointModel *)p->getTail();
   } else {
	return (JointModel *) NULL;
   }
}

void
JointModel::addManipulator(SoPath *p, ArticulatedBody *art)
{
   if (myManip != NULL)
	return;

   myManip = new JointManip;

   p->push(getInnerIndex());
   manipPath = p;

   myManip->replaceNode(p);

   SoDragger *myDragger;
   myDragger = myManip->getDragger();
   myDragger->addMotionCallback(&JointModel::dragMotionCallback, this);
   myDragger->addStartCallback(&JointModel::dragStartCallback, this);

   artBody = art;
}

void
JointModel::removeManipulator()
{
   if (myManip == NULL)
	return;

   myManip->replaceManip(manipPath,NULL);
   myManip = (JointManip *) NULL;
}


void
JointModel::dragMotionCallback(void *data, SoDragger *)
{
   JointModel *t = (JointModel *)data;

   t->artBody->recenterJointLimb(t);

   if (t->getNextJoint()) {

       SbViewportRegion vr;
       SoGetMatrixAction ma(vr);
       SbMatrix c;
    
       ma.apply(t->getInner());
       c = ma.getMatrix();
       c = c.inverse();
       c = c.multLeft(t->saveMat);

       t->getNextJoint()->getInner()->setMatrix(c);

       t->updateVisual();
   }

   if (t->prevJoint != NULL)
	t->prevJoint->updateVisual();

}

void
JointModel::dragStartCallback(void *data, SoDragger *)
{
   JointModel *t = (JointModel *)data;

    SoSearchAction sa;

    if (t->getNextJoint()) {
	
        sa.setNode(t->getNextJoint()->getInner());
        sa.apply(t);
        SoPath *p = sa.getPath();

        SbViewportRegion vr;
        SoGetMatrixAction ma(vr);
        ma.apply(p);
        t->saveMat = ma.getMatrix();
    }
}

/*********************************************************************
 * 
 *	    Jacobian  Class
 * 
 *********************************************************************/
  
Jacobian::Jacobian(int d)
{
    dim = d;
    jac = Matrix(6,  dim);
    jTj = Matrix(dim, dim);
    pseudoInverse = Matrix(dim, (LINEAR_ONLY)?3:6);
    jcv = Vector((LINEAR_ONLY)?3:6);
    jv = Vector(dim);
    
}

void Jacobian::updateJacobian(JointModel *start)
{
    /* have joints compute jacobian */
    start->computeJacobian(0, jac);
    
    /* test hack
    int i, j;
    for(i=0;i<3;i++)
	for(j=0;j<jac.n;j++)
	    jac.mm[i][j] = 0.0;
    */
    if(LINEAR_ONLY){
	Matrix j1(3,jac.n);
	int i, j;
	for(i=3;i<6;i++)
	    for(j=0;j<jac.n;j++)
	       j1.mm[i-3][j] = jac.mm[i][j];
	jac = j1;
    }
    
if(debug)
{    
   jac.print("Jacobian");
}    
    /* update jTj */
    jTj = jac.transpose()*jac;
    DEBUGPRINT(jTj, "jTj")
       
    /* find inverse of jtj */
    Matrix jTji = jTj.inverseSVD();
    DEBUGPRINT(jTji, "jTji");

    /* multiply by jt to get pseudo-inverse*/
    pseudoInverse = jTji * jac.transpose();
    DEBUGPRINT(pseudoInverse, "pseudoInverse");
   
    return;
}

void Jacobian::computeJointVector(Spv *cartVel)
{
    int i;
    
    /* copy cartesian velocity into jcv */
    if(LINEAR_ONLY){
	for(i=0;i<3;i++){
	    jcv.mm[i] = cartVel->v0[i];
	}
    }
    else {
	for(i=0;i<3;i++){
	    jcv.mm[i] = cartVel->v[i];
	    jcv.mm[i+3] = cartVel->v0[i];
	}
    }
    DEBUGPRINT(jcv, "jcv");
   
    /* multiply by the psuedoInverse */
    jv = pseudoInverse * jcv;
    DEBUGPRINT(jv, "pseudoInverse * jcv");   

}

void Jacobian::applyJointVector(JointModel *start)
{
    int i = 0;
    for(i=0;i<dim;i++){
	start->angle.setValue(start->angle.getValue() + jv.mm[i]);
	start->getOuter()->rotation.setValue(start->jointAxis.getValue(), start->angle.getValue());
	start = start->getNextJoint();
    } 
}

void Jacobian::motionResolve(JointModel *start)
{
    /* form (I-pseudoInverse*Jac) */
    Matrix mr(dim, dim);
    Matrix I(dim, dim);
    mr = (I.identity() - pseudoInverse*jac);
if(debug) {
    mr.print("I-pseudoInverse*Jac");
} 
    /* get center angle and gain */
    Vector vec(dim);
    JointModel *curr = start;
    int i;
    for(i=0;i<dim;i++){
//	printf("%f %f %f \n", 
//	    curr->centerGain.getValue(), 
//	    curr->angle.getValue(), 
//	    curr->center.getValue());
	vec.mm[i] = 2*curr->centerGain.getValue()*(curr->angle.getValue()
	     - curr->center.getValue());
	curr = curr->getNextJoint();
    }
    
    /* add mr*vec to jv */
    jv = jv - (mr * vec);
}

void Jacobian::inverseKinematics(JointModel *start, Spv *cartVel)
{
    /*
     *	cartVel should be the desired cartesian velocity of the end
     *	limb in the frame of the end limb. Here is an example:
     * 
     *	base->findLastJoint()->localXbase.multVec(&g1);
     *	dist = g1-ep;
     *
     *	cartVel.v0[0] = dist[0]/20;
     *	cartVel.v0[1] = dist[1]/20;
     *	cartVel.v0[2] = dist[2]/20;
     * 
     */
    if(LINEAR_ONLY) 
	jac = Matrix(6, dim);
    updateJacobian(start);
    computeJointVector(cartVel);
    motionResolve(start);
 //   jv.print("after");
    applyJointVector(start);
}

