#ifndef _JOINTMODEL_
#define _JOINTMODEL_

#include <Inventor/Sb.h>
#include <Inventor/SbColor.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoSFShort.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include "../CreatureLib/Vector.h"
#include "SpatialAlgebra.h"

#define JOINTSENSORPRIORITY 10
class SoTransform;
class SoSeparator;
class LimbModel;
class JointModel;
class SoShape;
class SoPath;
class SoDragger;
class Vector;
class Matrix;
class JointManip;
class ArticulatedBody;

/*
 * These classes implement a simple articulated chain. 
 * A limb consists of an innerJoint at its base. The joint implements
 * a right-handed coordinate system with the +z axis the axis of
 * rotation,  the +x axis the major axis of the limb,  and the +y
 * axis the 3rd axis. The connectivity is established via the
 * joints. Joint location and orientation is specified via offset
 * and twist,  both relative to the +x axis of the innerJoint.
 * 
 * Dark gray arrow indicates the +z axis, white arrow the +x axis.
 *
 */

#include "vector.h"
#include "SpatialAlgebra.h"

class JointModel : public SoSeparator {
    friend class Jacobian;
    friend class ArticulatedBody;

    SO_NODE_HEADER(JointModel);

private:
    JointModel *prevJoint;

    static int inner_index;
    static int outer_index;
    static int limb_index; 
    static int nextJoint_index;

    virtual ~JointModel();

    void xformSpv(Spv *arg);
    void computeJacobian(int i, Matrix &jac);

    JointManip *myManip;
    SoPath     *manipPath;

    ArticulatedBody *artBody;
    SbMatrix saveMat;
    
    
protected:
    JointModel();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    virtual SbBool	readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool	readInstance(SoInput *in);
#endif

    SoFieldSensor *angleSensor;

    static void angleSensorCB(void *, SoSensor *);
    static void dragMotionCallback(void *data, SoDragger *);
    static void dragStartCallback(void *data, SoDragger *);

public:

    SoSFFloat angle;    
    SoSFFloat center;
    SoSFFloat centerGain;
    SoSFVec3f jointAxis;    // by default (0,0,1)
    SoSFBool  shouldNotify;
    static void initClass();
    JointModel(LimbModel *l, SoTransform *new_inner);

    JointModel *findLastJoint();

    void setInner(SoTransform *new_inner);
    
    SoRotation  *getOuter() {return (SoRotation *)getChild(outer_index);};
    SoTransform *getInner() {return (SoTransform *)getChild(inner_index);};

    int getInnerIndex() {return inner_index;};
    int getOuterIndex() {return outer_index;};

    JointModel *getNextJoint();
    JointModel *getPrevJoint() { return prevJoint; };
    void getMaxExtension(float &maxExtension);

    SbBool isLastJoint();


    LimbModel *getLimb() { return  (LimbModel *)getChild(limb_index);};
    void setLimb(LimbModel *aLimb);
    void setNextJoint(JointModel *aJoint);
    void setAngleImmediate(float newAng);
    void setShouldNotify(SbBool yesOrNo);
    void addVisual();
    void removeVisual();
    void updateVisual();


    static JointModel *isSelection(SoPath *p);
    void addManipulator(SoPath *p, ArticulatedBody *art);
    void removeManipulator();
    
    int  numJointsToEnd();
};

class Jacobian
{
public:
    int dim;
    Matrix jac;
    Matrix jTj;
    Matrix pseudoInverse;
    Vector jcv;
    Vector jv;

    Jacobian(int dim);
    void updateJacobian(JointModel *start);
    void motionResolve(JointModel *start); 
    void computeJointVector(Spv *cartVel);
    void applyJointVector(JointModel *start);
    void inverseKinematics(JointModel *start, Spv *cartVel);
};

#endif
