#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include "lu.h"
#include "nexttypes.h"

#define TINY 1.0e-20
#define EPS 1.0e-3
#define FREERETURN {free_vector(xj,0,n-1);free_vector(xi,0,n-1);\
	free_vector(h,0,n-1);free_vector(g,0,n-1);return;}

static float at,bt,ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ? \
(ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0))

#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

int svdSolve(matrix a,int n,vector b)
{
    matrix u = new_matrix(1,n,1,n);
    matrix v = new_matrix(1,n,1,n);
    vector w = new_vector(1,n);
    vector bt = new_vector(1,n);
    vector x = new_vector(1,n);
    int i,j,problems=0;
    float wmin;
    
    /* copy in values */
    for(i=0;i<n;i++){
	bt[i+1]=b[i];
	for(j=0;j<n;j++)
	    u[i+1][j+1]=a[i][j];
    }
    /* do svdcmp */
    problems = svdcmp(u,n,n,w,v);
    /* get wmin */
    if(!problems){
	wmin = w[vectorFindMax(n,&w[1])+1]*1.0e-6;
	/* zero out w below wmin */
	for(i=1;i<=n;i++)
	    if(w[i]<wmin)
		w[i] = 0;
	/* do svbksp */
	svbksb(u,w,v,n,n,bt,x);
	/* copy out values */
	vectorAssign(n,&x[1],&b[0]);
    }
    
    /* free resources */
    free_matrix(u,1,n,1,n);
    free_matrix(v,1,n,1,n);
    free_vector(w,1,n);
    free_vector(x,1,n);
    free_vector(bt,1,n);
    
    return problems;
}

void svbksb(matrix u,vector w,matrix v,int m,int n,vector b,vector x)
{
	int jj,j,i;
	float s,*tmp,*vector();

	tmp=new_vector(1,n);
	for (j=1;j<=n;j++) {
		s=0.0;
		if (w[j]) {
			for (i=1;i<=m;i++) s += u[i][j]*b[i];
			s /= w[j];
		}
		tmp[j]=s;
	}
	for (j=1;j<=n;j++) {
		s=0.0;
		for (jj=1;jj<=n;jj++) s += v[j][jj]*tmp[jj];
		x[j]=s;
	}
	free_vector(tmp,1,n);
	return;
}

int svdcmp(matrix a,int m,int n,vector w,matrix v)
{
	int flag,i,its,j,jj,k,l=0,nm=0;
	float c,f,h,s,x,y,z;
	float anorm=0.0,g=0.0,scale=0.0;
	vector rv1;

	if (m < n) {
	printf("SVDCMP: You must augment A with extra zero rows\n");
	return -1;
	}
	rv1=new_vector(1,n);
	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				if (i != n) {
					for (j=l;j<=n;j++) {
						for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
						f=s/h;
						for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}
		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				if (i != m) {
					for (j=l;j<=m;j++) {
						for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
						for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
					}
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=MAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=n;i>=1;i--) {
		l=i+1;
		g=w[i];
		if (i < n)
			for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			if (i != n) {
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
					f=(s/a[i][i])*g;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else {
			for (j=i;j<=m;j++) a[j][i]=0.0;
		}
		++a[i][i];
	}
	for (k=n;k>=1;k--) {
		for (its=1;its<=30;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if (fabs(rv1[l])+anorm == anorm) {
					flag=0;
					break;
				}
				if (fabs(w[nm])+anorm == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					if (fabs(f)+anorm != anorm) {
						g=w[i];
						h=PYTHAG(f,g);
						w[i]=h;
						h=1.0/h;
						c=g*h;
						s=(-f*h);
						for (j=1;j<=m;j++) {
							y=a[j][nm];
							z=a[j][i];
							a[j][nm]=y*c+z*s;
							a[j][i]=z*c-y*s;
						}
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
				}
				break;
			}
			if ((its == 30)||(l==0)){
			if(its==30)
			    printf("No convergence in 30 SVDCMP iterations\n");
			else
			    printf("l went to zero\n");
			return -1;
			}
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=PYTHAG(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=PYTHAG(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y=y*c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=PYTHAG(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=(c*g)+(s*y);
				x=(c*y)-(s*g);
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
	free_vector(rv1,1,n);
	return 0;
}

#undef SIGN
#undef PYTHAG

matrix svdInverse(matrix a,int dim,matrix b)
{
    matrix u = new_matrix(1,dim,1,dim);
    matrix v = new_matrix(1,dim,1,dim);
    vector w = new_vector(1,dim);
    float wmin = 0.0;
    int k,i,j,maxAt,svd=0;
    
    /* copy out a into u */
    for(i=0;i<dim;i++)
	for(j=0;j<dim;j++)
	    u[i+1][j+1] = a[i][j];
    /* do svd */
    i = svdcmp(u,dim,dim,w,v);
    
    if(i<0){
	printf("trouble in svd land \n");
	printf("original matrix\n");
	print_matrix(6,a);
	luinverse(a,dim,b);
	printf("matrix from luinverse\n");
	print_matrix(6,b);
    }
    else {
    
	/* check condition */
	maxAt = vectorFindMax(dim,&w[1])+1;
	
	/* count singular values and zero out ones below tolerance */
	wmin = w[maxAt]*1e-6;
	for(i=1;i<=dim;i++){
	    if(w[i]<wmin)
		w[i] = 0.0;
	    else{
		w[i] = 1/w[i];
		svd++;
	    }
	}
//	if(svd<dim)
//	    printf("ill-conditioned matrix\n");
	    
	/* now calculate inverse */
	for(i=1;i<=dim;i++)
	    for(j=1;j<=dim;j++)
		u[i][j] = w[j]*u[i][j];
	
	for(i=1;i<=dim;i++){
	    for(j=1;j<=dim;j++){
		b[i-1][j-1] = 0.0;
		for(k=1;k<=dim;k++)
		    b[i-1][j-1] += v[i][k]*u[j][k];
	    }
	}
    }
    free_matrix(u,1,dim,1,dim);
    free_matrix(v,1,dim,1,dim);
    free_vector(w,1,dim);
    return b;
}
    
void asub(vector x,vector y)
{
    printf("WARNING UN-IMPLEMENTED \n");
    return;
}

void atsub(vector x,vector y)
{
    printf("WARNING UN-IMPLEMENTED \n");
    return;
}


void sparse(vector b,int n,vector x,float *rsq)
{
	int j,iter,irst=0;
	float aden,anum,bsq,dgg,eps2,gam,gg,rp;
	float *g,*h,*xi,*xj;

	g=new_vector(0,n-1);
	h=new_vector(0,n-1);
	xi=new_vector(0,n-1);
	xj=new_vector(0,n-1);
	eps2=n*EPS*EPS;
	for (;;) {
		++irst;
		asub(x,xi);
		rp=bsq=0.0;
		for (j=0;j<n;j++) {
			bsq += b[j]*b[j];
			xi[j] -= b[j];
			rp += xi[j]*xi[j];
		}
		atsub(xi,g);
		for (j=0;j<n;j++)
			h[j] = g[j] = -g[j];
		for (iter=0;iter<10*n;iter++) {
			asub(h,xi);
			anum=aden=0.0;
			for (j=0;j<n;j++) {
				anum += g[j]*h[j];
				aden += xi[j]*xi[j];
			}
			if (aden == 0.0) printf("Very singular matrix in SPARSE\n");
			anum /= aden;
			for (j=0;j<n;j++) {
				xi[j]=x[j];
				x[j] += anum*h[j];
			}
			asub(x,xj);
			*rsq=0.0;
			for (j=0;j<n;j++) {
				xj[j] -= b[j];
				*rsq += xj[j]*xj[j];
			}
			if (*rsq == rp || *rsq <= bsq*eps2) FREERETURN
			if (*rsq > rp) {
				for (j=0;j<n;j++) x[j]=xi[j];
				if (irst >= 3) FREERETURN
				break;
			}
			rp = *rsq;
			atsub(xj,xi);
			gg=dgg=0.0;
			for (j=0;j<n;j++) {
				gg += g[j]*g[j];
				dgg += (xi[j]+g[j])*xi[j];
			}
			if (gg == 0.0) FREERETURN
			gam=dgg/gg;
			for (j=0;j<n;j++) {
				g[j] = -xi[j];
				h[j]=g[j]+gam*h[j];
			}
		}
		printf("Too many interations in routine SPARSE\n");
	}
}

int 
ludcmp(float **a, int n, int *indx, float *d)
{
    int                 i, imax = -1, j, k;
    float              big, dum, sum, temp;
    float             *vv;

    vv = new_vector(0, n);
    *d = 1.0;
    for (i = 0; i < n; i++) {
	big = 0.0;
	for (j = 0; j < n; j++)
	    if ((temp = fabs(a[i][j])) > big)
		big = temp;
	if (big == 0.0) {
	    printf("singular matrix -ending routine\n");
	    return 0;
	}
	vv[i] = 1.0 / big;
    }
    for (j = 0; j < n; j++) {
	for (i = 0; i < j; i++) {
	    sum = a[i][j];
	    for (k = 0; k < i; k++)
		sum -= a[i][k] * a[k][j];
	    a[i][j] = sum;
	}
	big = 0.0;
	for (i = j; i < n; i++) {
	    sum = a[i][j];
	    for (k = 0; k < j; k++)
		sum -= a[i][k] * a[k][j];
	    a[i][j] = sum;
	    if ((dum = vv[i] * fabs(sum)) >= big) {
		big = dum;
		imax = i;
	    }
	}
	if (j != imax) {
	    for (k = 0; k < n; k++) {
		dum = a[imax][k];
		a[imax][k] = a[j][k];
		a[j][k] = dum;
	    }
	    *d = -(*d);
	    vv[imax] = vv[j];
	}
	indx[j] = imax;
	if (a[j][j] == 0.0)
	    a[j][j] = TINY;
	if (j != (n-1)) {
	    dum = 1./ (a[j][j]);
	    for (i = j +1; i < n; i++)
		a[i][j] *= dum;
	}
    }
    free_vector(vv, 0, n);
    return 1;
}

void 
lubksb(float **a, int n, int *indx, float *b)
{
    int                 i, ii = 0, ip, j,flag=0;
    float              sum;

    for (i = 0; i < n; i++) {
	ip = indx[i];
	sum = b[ip];
	b[ip] = b[i];
	if (flag)
	    for (j = ii; j <= i - 1; j++)
		sum -= a[i][j] * b[j];
	else if (sum){
	    flag = 1;
	    ii = i;
	}
	b[i] = sum;
    }
    for (i = n-1; i >= 0; i--) {
	sum = b[i];
	for (j = i + 1; j < n; j++)
	    sum -= a[i][j] * b[j];
	b[i] = sum / a[i][i];
    }
    return;
}

int  luinverse(float **a_orig, int n, float **y)
{
    int                 i, j;
    float              d;
    float             *col = new_vector(0, n);
    int                *indx = (int *)malloc((n) * sizeof(int));
    matrix a = copyMatrix(n, a_orig);

    i = ludcmp(a, n, indx, &d);
    if (i == 0) {
	printf("singular matrix, inverse not found\n");
	return 0;
    }
    for (j = 0; j < n; j++) {
	for (i = 0; i < n; i++)
	    col[i] = 0.0;
	col[j] = 1.0;
	lubksb(a, n, indx, col);
	for (i = 0; i < n; i++)
	    y[i][j] = col[i];
    }
    free_vector(col, 0, n);
    free_matrix(a,0,n-1,0,n-1);
    free(indx);
    return 1;
}

int luSolve(float **a_orig, int n, float *b)
{
    int      i;
    int      *indx = (int *)malloc((n + 1) * sizeof(int));
    float              d;
    matrix a = copyMatrix(n, a_orig);
    i = ludcmp(a, n, indx, &d);
    if (i == 0){
	free(indx);
	free_matrix(a,0,n-1,0,n-1);
	return i;
    }else {
	lubksb(a, n, indx, b);
	free(indx);
	free_matrix(a,0,n-1,0,n-1);
	return 1;
    }
}
