#include <iostream.h>
#include <unistd.h>
#include <bstring.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "AliveClient.h"
#include "AliveDispatch.h"

const int MAX_NAME_LENGTH = 128;
const float MAX_SERVER_DELAY = 1.0;

AliveClient::AliveClient()
{
  tcp_sc = new SocketClient(ALIVE_TCP_PORT);
  udp_sc = new SocketClient(ALIVE_MULTICAST_PORT);
  getMyAddress();
}

AliveClient::~AliveClient()
{
  delete tcp_sc;
  delete udp_sc;
}

int
AliveClient::connectToServer(char *server_name)
{
  struct timeval t1, t2;
  struct timezone tz = {0,0};

  if (getServerAddress(server_name) < 0)
    {
      cerr << "AliveClient::connectToServer: error getting server address" << endl;
      return -1;
    }
  
  // get ready to find out if we can use multicasting

  udp_sc->setPort(ALIVE_MULTICAST_PORT);
  udp_sc->setUsingMulticast(1);
  udp_sc->setSuppressingMulticastLoopback(0);
  udp_sc->connectToServer(ALIVE_GROUP);

  if (tcp_sc->connectToServer(server_name) != SocketClient::MSG_NO_ERROR)
    {
      cerr << "AliveClient::connectToServer: Error making TCP connecton to server!" << endl;
      udp_sc->closeConnection();
      return -1;
    }
  
  // the server should respond to our connection by
  // multicasting a CLIENT_CONNECTION_ACKNOWLEDGE message.

  gettimeofday(&t1, &tz);
  gettimeofday(&t2, &tz);
  cerr << "Starting wait" << endl;
  while (AliveDispatch::timevalDifference(t2,t1) < MAX_SERVER_DELAY)
    {
      SocketBuf *sb = udp_sc->readWithPollAndTimeout();
      if (!sb)
	{
	  if ((udp_sc->getGlobalMessage() == SocketClient::ERR_CURRENT_SOCKET_CLOSED) ||
	      (udp_sc->getGlobalMessage() == SocketClient::ERR_CURRENT_SOCKET_NOT_OPEN))
	    {
	      cerr << "AliveClient::connectToServer: Error: multicast socket closed unexpectedly!" << endl;
	      tcp_sc->closeConnection();
	      udp_sc->closeConnection();
	      return -1;
	    }
	}
      else
	{
	  // check to see if it's the server validating our connection
	  if (serverIsValidatingMe(sb))
	    {
	      delete sb;
	      cerr << "Server validated me" << endl;
	      return 0;
	    }
	  cerr << "Server didn't validate me" << endl;
	  delete sb;
	}
      gettimeofday(&t2, &tz);
    }
  cerr << "Ended wait" << endl;
  // we didn't get validated, so we have to use a UDP connection for updates
  udp_sc->closeConnection();
  udp_sc->setUsingMulticast(0);
  udp_sc->setPort(ALIVE_UDP_PORT);
  if (udp_sc->connectToServer(server_name) != SocketClient::MSG_NO_ERROR)
    {
      cerr << "AliveClient::connectToServer: Error making UDP connection to server!" << endl;
      tcp_sc->closeConnection();
      return -1;
    }
  cerr << "Connected (UDP) to server instead" << endl;
  return 0;
}

int
AliveClient::getMyAddress()
{
  char                myname[MAX_NAME_LENGTH];
  int                 client_len;
  struct hostent      *my_info;
  struct sockaddr_in  client_info;
  
  client_len = sizeof(client_info);
  gethostname(myname, MAX_NAME_LENGTH);
  my_info = (struct hostent *)gethostbyname(myname);
  if (my_info == NULL)
    {  
      cerr << "AliveClient::getMyAddress: I don't know my own name: \"" << myname << "\"" << endl;
      return -1;
    }
  
  bzero((char *)&client_info, client_len);
  bcopy(my_info->h_addr, 
	(char *)&(client_info.sin_addr), 
	my_info->h_length);
  my_address = client_info.sin_addr.s_addr;
  return 0;
}
  
int
AliveClient::getServerAddress(char *server_name)
{
  struct sockaddr_in server;
  struct hostent     *host_info;

  host_info = (struct hostent *)gethostbyname((char *)server_name);
  if (host_info == NULL)
    {  
      cerr << "AliveClient::getServerAddress: unknown host \"" << server_name << "\"" << endl;
      return -1;
    }
  bzero((char *)&server, sizeof(server));
  bcopy(host_info->h_addr, (char *)&(server.sin_addr), host_info->h_length);
  server_address = server.sin_addr.s_addr;
  return 0;
}

int
AliveClient::serverIsValidatingMe(SocketBuf *sb)
{
  cerr << "Packet is from " << AliveDispatch::longToDotFormat(sb->from_address) << ", and"
    << " server is " << AliveDispatch::longToDotFormat(server_address) << endl;

  if (sb->from_address == server_address)
    {
      // where should this go?
      // AliveDispatch, AliveBase?

      if (AliveDispatch::getAliveDataType(sb) == AliveDispatch::CLIENT_CONNECTION_ACKNOWLEDGE)
	{
	  struct ConnectionAcknowledge *ca = AliveDispatch::getConnectionAcknowledge(sb);
	  cerr << "Server validated address " << AliveDispatch::longToDotFormat(ca->connection_address) << endl;
	  if (ca->connection_address == my_address)
	    {
	      delete ca;
	      return 1;
	    }
	  else
	    {
	      delete ca;
	    }
	}
    }
  return 0;
}
  
