#ifndef _ALIVE_DISPATCH_
#define _ALIVE_DISPATCH_

#include <sys/time.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include "AliveDispatchInternals.h"  // see comments on data structures here

class AliveServer;
class AliveClient;
class UpdateStateData;
class BaseCreature;
class MotorCommandBlock;
class CreatureCache;
class SocketBuf;

class AliveDispatch {
private:
  static AliveServer *as;
  static AliveClient *ac;

  // static (big) SocketBuf for all serialization
  static SocketBuf *sb;

  // static XDR stream for writing to this buffer
  static XDR xdrs;

  // statics for MotorCommandBlock serialization
  static int mcb_count;
  static unsigned int mcb_xdr_repos;

  static CreatureCache *creature_cache;
  static struct timeval time_of_last_sync;
  static float time_between_syncs;

public:
  static int getAliveDataType(SocketBuf *sb);

  // Functions for wrapping/unwrapping data types for transmittal

  static SocketBuf *wrapConnectionAcknowledge(ConnectionAcknowledge *ca);
  static ConnectionAcknowledge *getConnectionAcknowledge(SocketBuf *sb);
  static SocketBuf *wrapCreatureRegistryRequest(CreatureRegistryRequest *crr);
  static CreatureRegistryRequest *getCreatureRegistryRequest(SocketBuf *sb);
  static SocketBuf *wrapCreatureRegistryAcceptance(CreatureRegistryAcceptance *cra);
  static CreatureRegistryAcceptance *getCreatureRegistryAcceptance(SocketBuf *sb);
  static SocketBuf *wrapCreatureDownloadRequest(CreatureDownloadRequest *cdr);
  static CreatureDownloadRequest *getCreatureDownloadRequest(SocketBuf *sb);
  static SocketBuf *wrapCreatureDownloadService(CreatureDownloadService *cds);
  static CreatureDownloadService *getCreatureDownloadService(SocketBuf *sb);

  // Function that initializes the buffer for all writing serialization
  static void initialize();

  // How to tell the AliveDispatcher whether it is a
  // client or a server
  static int setAliveServer(AliveServer *as);
  static int setAliveClient(AliveClient *ac);

  // How to set the creature cache that the AliveDispatcher
  // is looking at
  static int setCreatureCache(CreatureCache *cc);

  // Polls the multicast socket (on either client or server)
  // to see if any update has come in
  static SocketBuf *pollForUpdate();

  // Main dispatch routine.
  // Called inside CreatureCache.
  static void mainDispatch(UpdateStateData &usd); 

  // Frees all char ptrs in an MCB
  static void flushMCB(MotorCommandBlock *mcb);

  // Dispatches an MCB series from the net to a droid
  static void dispatchMCBSeries(SocketBuf *sb, 
				UpdateStateData &usd,
				int parent_only = FALSE);

  // Called inside CreatureCache::addBaseCreature
  // if a client adds a new creature to the world
  static int registerCreatureWithServer(BaseCreature *bc);
  // Server side function that does the registration
  static int serviceCreatureRegistryRequest(SocketBuf *sb);

  // Server side function that multicasts new creature announcement
  static void broadcastCreatureAnnouncement(BaseCreature *bc);

  // Client calls this to download a creature from the server
  static int downloadCreatureFromServer(CreatureDownloadRequest *cdr);
  // Server side function which services download requests
  static int serviceCreatureDownloadRequest(SocketBuf *sb);
  
  // Client calls this to synchronize contents of CreatureCache with server
  static int synchronizeCreatureDatabaseWithServer();
  // Server calls this to service requests from above function
  static int serviceCreatureDatabaseSynchronizationRequest(SocketBuf *sb);

  // This is called from mainDispatch to send out periodic synchronization
  // packets for each master creature in the local CreatureCache.
  static void conditionalSendSyncForAllMasterCreatures();
  // This is called to dispatch incoming synchronization packets to
  // droids on the local machine.
  static void dispatchSyncToCreature(SocketBuf *sb);

  // Called when a Master MotorCreature begins its behavior
  // system update. Initializes variables in preparation for
  // serializing MotorCommandBlocks.
  static void beginMCBSerialization(BaseCreature *bc);
  // Serializes one MotorCommandBlock
  static int serializeMCB(MotorCommandBlock *mcb);
  // Called when a Master MotorCreature ends its behavior
  // system update cycle. Dispatches the MCB series over the net.
  static void endMCBSerialization();

  // Method to send a MotorCommand over the net to a parent creature only.
  static void sendMCBToParentCreature(BaseCreature *bc, 
				      MotorCommandBlock *mcb);

  // utility functions
  static float timevalDifference(const struct timeval &tv1,
			  const struct timeval &tv2);
  static char *longToDotFormat(int32_t);

enum AliveDataType {
  CLIENT_CONNECTION_ACKNOWLEDGE,
  CREATURE_REGISTRY_REQUEST,
  CREATURE_REGISTRY_ACCEPTANCE,
  CREATURE_DOWNLOAD_REQUEST,
  CREATURE_DOWNLOAD_SERVICE,
  MCB_SERIALIZATION,
  CREATURE_DATABASE_SYNCHRONIZATION_REQUEST,
  CREATURE_DATABASE_SYNCHRONIZATION_SERVICE,
  CREATURE_SYNCHRONIZATION_REQUEST,
  CREATURE_SYNCHRONIZATION_SERVICE,
  MCB_SERIALIZATION_FOR_PARENT
    // more data types to follow
  };
};

#endif

  
