#ifndef _MOTOR_CREATURE_
#define _MOTOR_CREATURE_

#include <Inventor/SbBasic.h>
#include <Inventor/fields/SoMFString.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/SbPList.h>
#include <Inventor/SoPath.h>
#include "CreatureLib.h"

class SoSeparator;
class GenericController;
class MotorSkill;
class MotorCommandBlock;

class MotorCreature : public BaseCreature {

    SO_NODE_HEADER(MotorCreature);

public:
    enum MotorCreatureType{
	MASTER, 
	DROID 
    }; 
    
    SoSFEnum motorCreatureType;
    SoSFFloat lastRemoteUpdateTick;
    SbPList remoteMCBList;
    int updatedRemotelyThisTick;
        
    SoSFBool shouldTrace;
    
    /* these guys are for the behavior system */
    SoSFBool runBehavior;
    SoSFBool traceBehavior;
    SoSFString behaviorFileName;
    SoNode *behaviorRoot;
    SoSFFloat turnBase;	    // equivalent of wheel-base
    SoMFString sounds;
    SoSFBool useRealTime;
    SoSFFloat defaultFwdSpeed;
    SoSFFloat defaultAngSpeed;
    SoSFBool updateReps;
    Spv vel;
    float speed;
    float angle;
    
    SbPList dofs;
    SbPList motorSkills;
    GenericController *controller;
    
    static void initClass();
    MotorCreature();

    //
    // 
    MotorSkill *getMotorSkillByName(const char *name);
    
    /* methods for motor system and controller */
    /*
     *  It should not be necessary to over-ride these methods
     */
    GenericController *getController() { return controller; };
    
    virtual void buildMotorSystem();
    
    virtual void updateMotorSkillsAt(UpdateStateData &usd);
    
    void installXYZInterpolator(const char *name);
    
    SoNode *getSensorOfType(SoType stype);
    SoNode *getSensorWithName(SbName &aName);
    /*
     * OVER-RIDE THESE METHODS
     */
    /* use this method to add class-specific controller */
    virtual GenericController *addController();
    
    /* use this class to add class-specific motor skills */
    virtual void addMotorSkills();
    
    /* use this class to add class-specific behavior system */
    virtual void buildBehaviorSystem();
    virtual SoAction *getBehaviorUpdateAction();
    
    /* use this method to install class-specific interpolators */
    virtual void getDOFInterpolators();
    
    /* use this method to update class-specific behavior system */
    virtual SbBool beginBehaviorSystemUpdate(UpdateStateData &usd);
    virtual void updateBehaviorSystemAt(UpdateStateData &usd);
    virtual void endBehaviorSystemUpdate(UpdateStateData &usd);
    virtual void traceBehaviorSystemAt(UpdateStateData &usd);
    
    /* these can be over-ridden to set the cmd in the mcb to the
     * appropriate action to get it to move,  or to get it to a
     * canonical posture (such as standing).
     */
    virtual int checkIfReadyToMove(MotorCommandBlock &mcb);
    virtual int checkIfNeedToChangePosture(MotorCommandBlock &mcb, int cmd);
    virtual int checkIfNeedToChangePosture(MotorCommandBlock &mcb, MotorSkill *thisOwner);

    
    /* methods used to implement time-step updates */
    
    /* this calls updateBehaviorSystemAt() and updateMotorSystemAt()
     * by default,  over-ride only if you want additional functionality. 
     */
    virtual void updateStateAt(UpdateStateData &usd);
    
    /*
     * over-ride as necessary for any post-updateStateAt() changes to state
     */
    virtual int stepTo(UpdateStateData &usd);
    void restart();
    
    /* general interrogation methods */
    void getVelocity(Spv &v);
//    int isMoving();

    void buildSounds();

    /* Contains creature-specific synchronization information for Distributed ALIVE
       (see BaseCreature.h) */

    virtual bool_t syncToOrFromStream(XDR *xdrs);

protected:
    virtual ~MotorCreature();
#ifdef SO_VERSION_REVISION
    SbBool readInstance(SoInput *in, unsigned short flags);
#else
    SbBool readInstance(SoInput *in);
#endif
};

#endif
