#include <iostream.h>
#include "MotorSkill.h"
#include <Inventor/SbPList.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/engines/SoInterpolate.h>
#include "MotorCreature.h"

SoEngine *getConnectedEngine(SoField *sof)
{
    SoEngineOutput *eo;
    if(sof->getConnectedEngine(eo))
	return(eo->getContainer());
    else
	return NULL;
}


float cycleWrap(float x)
{
    float r = fmod(x, 1);
    if(r<0)
	r += 1.0;
    return r;
}    
    
Dof::Dof(void *cont)
{
    interpolator = cont;
    auxInterpolator0 = auxInterpolator1= NULL;
    use = Dof::RELEASED;
    currentTarget = 1.0e6;
    owner = NULL;
}

Dof::Dof(void *cont, void *geo)
{
    interpolator = cont;
    auxInterpolator0 = auxInterpolator1= NULL;
    use = Dof::RELEASED;
    geometry = geo;
    vel0 = 0;
    currentTarget = 1.0e6;
    owner = NULL;
}

Dof *Dof::grab(void *grabber)
{
    if(use == Dof::RELEASED){
	owner = (MotorSkill *)grabber;
	use = Dof::GRABBED;
	return this;
    }
    else return NULL;
}

Dof *Dof::release(void *grabber)
{
    if((use == Dof::GRABBED)&&(owner == (MotorSkill *)grabber)){
	use = Dof::RELEASED;
	owner = NULL;
	return this;
    }
    else if(use == Dof::RELEASED)
	return this;
    else
	return NULL;
}

int Dof::isGrabbed()
{
    return(use == Dof::GRABBED)?1:0;
}

void Dof::setAlpha(float alpha, SbBool doAll)
{
    SoInterpolate *tmp;
    
//    alpha += .01 * randFloatInInterval(-1.0,1.0); 
    /* main interpolator */
    tmp = (SoInterpolate *)interpolator;
    if(tmp &&(fabs(tmp->alpha.getValue()-alpha)> 1.0e-5))
	tmp->alpha.setValue(alpha);
    if(doAll== TRUE){
	/* 1st aux interpolator */
	tmp = (SoInterpolate *)auxInterpolator0;
	if(tmp &&(fabs(tmp->alpha.getValue()-alpha)> 1.0e-5))
	    tmp->alpha.setValue(alpha);
	
	/* 2nd aux  interpolator */
	tmp = (SoInterpolate *)auxInterpolator1;
	if(tmp &&(fabs(tmp->alpha.getValue()-alpha)> 1.0e-5))
	    tmp->alpha.setValue(alpha);
    }
}

float Dof::setAlphaFromSpeedAndTarget(float speed, float targVal, SbBool doAll)
{
    SoInterpolate *interp = (SoInterpolate *)interpolator;
    float cv = ((SoSFFloat *)interp->getField("alpha"))->getValue();
    float sp;
    float amtToMove;
    sp = (cv<targVal)?speed : -speed;
    if(sp>0)
	amtToMove = (cv+sp>=targVal)?targVal:cv+sp;
    else
	amtToMove = (cv+sp<=targVal)?targVal:cv+sp;
    setAlpha(amtToMove, doAll);
    return (amtToMove);
}

void Dof::initializeCubic(float cv, float tv, float dur)
{
    if(dur<=1.0){
	coeff[0] = tv;
	coeff[1] = coeff[2] = coeff[3] = 0.0;
	duration = 1.0;
	vel0 = tv - cv;
    }
    else {
	duration = (int)ceil(dur);
	coeff[0] = cv;
	coeff[1] = vel0;
	coeff[2] = (3.0/(dur*dur))*(currentTarget -cv) - (2.0/dur)*vel0;
	coeff[3] = (-2.0/(dur*dur*dur))*(currentTarget -cv) 
	    + (1.0/(dur*dur))*vel0;
    }
}

float Dof::getNextAlpha()
{
    delta++;
    float ct;
    
    if(delta>=((int)duration))
	ct =  currentTarget;
    else {
	ct = coeff[0] + coeff[1]*delta 
	    + coeff[2]*delta*delta + coeff[3]*delta*delta*delta;
    }
    if(ct>1.0)
	ct = 1.0;
    else if(ct<0)
	ct = 0.0;
    return ct;
}

float Dof::setAlphaUsingSpline(float speed, float targVal, SbBool doAll)
{
    SoInterpolate *interp = (SoInterpolate *)interpolator;
    float cv = ((SoSFFloat *)interp->getField("alpha"))->getValue();
    
    if((fabs(currentTarget - targVal)>2.0e-2) 
	 || (fabs(speed - dvel)>1.0e-2)){
	dvel = speed;
	currentTarget = targVal;
	delta = 0;
	initializeCubic(cv, targVal, (fabs(targVal-cv)/speed));
    }
    float amtToMove = getNextAlpha();
    vel0 = amtToMove - cv;
    setAlpha(amtToMove, doAll);
    return (amtToMove);
}

bool_t
Dof::toOrFromStream(XDR *xdrs, MotorCreature *parent)
{
  float alpha;
  int int_owner;  // index of owner skill in parent MotorCreature's SbPList

  if (!xdr_int(xdrs, &use))
    return FALSE;
  if (use == Dof::GRABBED)  // there is a valid owner
    {
      if (xdrs->x_op == XDR_ENCODE)
	{
	  int_owner = parent->motorSkills.find(owner);
	  if (int_owner < 0)
	    {
	      cerr << "Dof::toOrFromStream: owner was not found!" << endl;
	      return FALSE;
	    }
	}
      
      if (!xdr_int(xdrs, &int_owner))
	return FALSE;
      
      if (xdrs->x_op == XDR_DECODE)
	{
	  owner = (MotorSkill *) parent->motorSkills[int_owner];
	}
    }

  if (!xdr_SbVec4f(xdrs, &coeff))
    return FALSE;
  if (!xdr_float(xdrs, &currentTarget))
    return FALSE;
  if (!xdr_int(xdrs, &delta))
    return FALSE;
  if (!xdr_float(xdrs, &duration))
    return FALSE;
  if (!xdr_float(xdrs, &vel0))
    return FALSE;
  if (!xdr_float(xdrs, &dvel))
    return FALSE;
  if (interpolator)
    {
      if (xdrs->x_op == XDR_ENCODE)
	alpha = ((SoInterpolate *)interpolator)->alpha.getValue();
      
      if (!xdr_float(xdrs, &alpha))
	return FALSE;

      if (xdrs->x_op == XDR_DECODE)
	((SoInterpolate *)interpolator)->alpha.setValue(alpha);
    }

  if (auxInterpolator0)
    {
      if (xdrs->x_op == XDR_ENCODE)
	alpha = ((SoInterpolate *)auxInterpolator0)->alpha.getValue();
      
      if (!xdr_float(xdrs, &alpha))
	return FALSE;

      if (xdrs->x_op == XDR_DECODE)
	((SoInterpolate *)auxInterpolator0)->alpha.setValue(alpha);
    }

  if (auxInterpolator1)
    {
      if (xdrs->x_op == XDR_ENCODE)
	alpha = ((SoInterpolate *)auxInterpolator1)->alpha.getValue();
      
      if (!xdr_float(xdrs, &alpha))
	return FALSE;

      if (xdrs->x_op == XDR_DECODE)
	((SoInterpolate *)auxInterpolator1)->alpha.setValue(alpha);
    }

  return TRUE;
}

MotorSkill::MotorSkill(MotorCreature *p, char *pname)
{
    parent = p;
    name = new SbName(pname);
    parent->motorSkills.append(this);
    counter = 0;
    numDofs = 0;
    isImmediate = FALSE;
    state = MotorSkill::SKILL_OFF;
}

float MotorSkill::getDeltaTowardTarget(SoInterpolateFloat *interp, float targVal)
{
    float cv = ((SoSFFloat *)interp->getField("alpha"))->getValue();
    float sp;
    float amtToMove;
    sp = (cv<targVal)?speed : -speed;
    if(sp>0)
	amtToMove = (cv+sp>targVal)?targVal:cv+sp;
    else
	amtToMove = (cv+sp<targVal)?targVal:cv+sp;
    return amtToMove;
}

float MotorSkill::getDeltaTowardTarget(float cv, float targVal)
{
    float sp;
    float amtToMove;
    sp = (cv<targVal)?speed : -speed;
    if(sp>0)
	amtToMove = (cv+sp>targVal)?targVal:cv+sp;
    else
	amtToMove = (cv+sp<targVal)?targVal:cv+sp;
    return amtToMove;
}

int MotorSkill::dofsAvailable()
{
    int i;
    for(i=0;i<numDofs;i++){
	if((dofs[i]->use == Dof::GRABBED)&&
	    (dofs[i]->owner != this)){
//		printf("skill %s wants dof %d not not released by %s \n", 
//		    name->getString(), 
//		    i, 
//		    dofs[i]->owner->name->getString());
		return MotorSkill::DOF_ALREADY_GRABBED;
	    }
    }
    return MotorSkill::DOFS_AVAILABLE;
}

int MotorSkill::grabDofs()
{
    int i;
    for(i=0;i<numDofs;i++){
	dofs[i]->grab(this);
    }
    return MotorSkill::DOFS_AVAILABLE;
}

void MotorSkill::releaseDofs()
{
    int i;
    for(i=0;i<numDofs;i++){
	dofs[i]->release(this);
    }
}

int MotorSkill::turnOn()
{
    return turnOn(NULL);
}

int MotorSkill::turnOn(void *arg)
{
    if(isImmediate == TRUE){
	return _doImmediate(arg);
    }
    else 
	return _turnOn(arg);
}

int MotorSkill::_turnOn(void *arg)
{
    requestedThisTick = 1;
    if (state == MotorSkill::SKILL_ON){
	modifyRun(arg);
	return MotorSkill::SKILL_ALREADY_ON;
    }
    else {
	/* check if dofs are available */
	if(dofsAvailable() == MotorSkill::DOFS_AVAILABLE){
	    if(prepareToRun(arg)){
		grabDofs();
		state = MotorSkill::SKILL_ON;
		return MotorSkill::SKILL_ON;
	    }
	    else{ 
		return MotorSkill::SKILL_NOT_APPLICABLE;
	    }
	}
	else {
	    state = MotorSkill::SKILL_OFF;
	   // printf("dof already grabbed, request refused \n");
	    return MotorSkill::DOF_ALREADY_GRABBED;
	}
    }
}

void MotorSkill::updateImmediate(void *)
{
    return;
}

int MotorSkill::_doImmediate(void *arg)
{
    int s = _turnOn(arg);
    if((s == MotorSkill::SKILL_ALREADY_ON)||
	(s == MotorSkill::SKILL_ON)){
	updateImmediate(arg);
	s = turnOff(arg);
    }
    return s;
}

int MotorSkill::turnOff()
{
   return turnOff(&defaultsForOff);
}

int MotorSkill::turnOff(void *arg)
{
    requestedThisTick = 0;
    if(state == MotorSkill::SKILL_OFF)
	return MotorSkill::SKILL_OFF;
    else if(prepareToTurnOff(arg)){
	state = MotorSkill::SKILL_OFF;
	releaseDofs();
	return MotorSkill::SKILL_OFF;
    }
    else {
	state = MotorSkill::SKILL_OFF_BLOCKED;
	return MotorSkill::SKILL_OFF_BLOCKED;
    }
}
int MotorSkill::modifyRun(void *)
{
    return 1;  
}

int MotorSkill::prepareToRun(void *)
{
    return 1;  
}

int MotorSkill::prepareToTurnOff(void *)
{
    return 1;
}

void MotorSkill::updateStateAt(UpdateStateData &)
{
    requestedThisTick = 0;
    if(state == MotorSkill::SKILL_OFF_BLOCKED)
	turnOff();
}

void
MotorSkill::staticSetup()
{
}

bool_t
MotorSkill::toOrFromStream(XDR *xdrs)
{
  static int c1;

  if (!xdr_int(xdrs, &state))
    return FALSE;
  if ((xdrs->x_op == XDR_DECODE) && ((state == SKILL_ON) || (state == SKILL_ALREADY_ON)))
    {
      staticSetup();
    }
  if (!xdr_int(xdrs, &requestedThisTick))
    return FALSE;
  if (!xdr_float(xdrs, &counter))
    return FALSE;
  if (!xdr_float(xdrs, &speed))
    return FALSE;
  for (c1 = 0; c1 < numDofs; c1++)
    {
      if (!dofs[c1]->toOrFromStream(xdrs, parent))
	return FALSE;
    }
  return TRUE;
}


