#ifndef MOTORSKILL
#define MOTORSKILL
#include "CreatureLib.h"
#include "GenericMotorControl.h"
#include <rpc/xdr.h>

class MotorCreature;
class SBPList;
class MotorSkill;
class SoInterpolate;
class SoInterpolateFloat;
class SoEngine;
class SoField;

#define MAXDOFS 10

/* convenience function */
SoEngine *getConnectedEngine(SoField *sof);

class Dof {
    public:
    enum State{
	GRABBED,		
	RELEASED
    };
    void *interpolator;
    void *auxInterpolator0;
    void *auxInterpolator1;
    void *geometry;
    int use;
    MotorSkill *owner;
    Dof(void *cont);
    Dof(void *cont, void *geo);
    Dof *grab(void *grabber);
    Dof *release(void *grabber);
    int isGrabbed(); 
    void setAlpha(float alpha, SbBool doAll = FALSE);
    float setAlphaFromSpeedAndTarget(float speed, float targVal, SbBool doAll = FALSE);
    float setAlphaUsingSpline(float speed, float targVal, SbBool doAll = FALSE);
    
    // function to read in or write out the state of the Degree of Freedom
    // to allow droids to sync up with a master in Distributed ALIVE.
    virtual bool_t toOrFromStream(XDR *xdrs, MotorCreature *parent);

    /* fields and methods below used for spline interpolation */
    private: 
    SbVec4f coeff;
    float currentTarget;
    int delta;
    float duration;
    float vel0;
    float dvel;
    void initializeCubic(float cv, float tv, float dur);
    float getNextAlpha();

};

class MotorSkill {
    public:
    enum State {
	SKILL_ON, 
	SKILL_ALREADY_ON, 
	SKILL_OFF, 
	DOF_ALREADY_GRABBED,
	DOFS_AVAILABLE,
	SKILL_NOT_APPLICABLE,
	SKILL_OFF_BLOCKED   
    };
    int state;
    SbName *name;

    // this can be used to terminate a continuous activity by testing its
    // value in updateStateAt(). See WaggingAndMovingTail for example.
    int requestedThisTick;
    float counter;
    float speed;
    Dof *dofs[MAXDOFS];
    MotorCreature *parent;
    int numDofs;
    SbBool isImmediate;
    
    MotorSkill(MotorCreature *p, char *pname);
    int grabDofs();
    int dofsAvailable();
    void releaseDofs();
    // shouldn't have to over ride these
    int turnOn();
    int turnOn(void *arg);
    int turnOff();
    int turnOff(void *arg);

    //
    // when turn off is call iwth out any args this is sent.
    // EMPTY by default sub-class must over-ride if desired.
    //
    MotorCommandBlock defaultsForOff;
    
    // convenience function 
    float getDeltaTowardTarget(SoInterpolateFloat *interp, float targVal);
    float getDeltaTowardTarget(float cv, float targVal);
    
    // This function gets called when skill is sent turnOn and it is already
    // on. Can be used to modify parameters (speed, target etc...)
    virtual int modifyRun(void *arg);
    
    // This function gets called on every tick whether it is active or not.
    // Set the interpolator's value here. You should call MotorSkill::updateStateAt()
    // at the end of your over-ridden method
    virtual void updateStateAt(UpdateStateData &usd);
    
    // This function is called when the MotorSkill is turned on and
    // isImmediate is true. It will only be called if the Skill can be
    // turned on, and does a turn off at the end.This is intended for
    // motor skills which are synchronous (for example, accessor or
    // setter functions).
    virtual void updateImmediate(void *arg);
    
    // This gets called by turnOn and (1) resources are available and
    // (2) not already running. Can be used as additional test of whether
    // skill can be executed, or to read arguments. 
    virtual int prepareToRun(void *arg);// returns 1 if ok to run
    
    // This gets called to see if it is ok to turnOff
    virtual int prepareToTurnOff(void *arg);	// returns 1 if ok to end
    
    // This function is called during prepareToRun to set up internal
    // pointers, etc. necessary for the motor skill to run. As an example,
    // in DogMotorSkills.c++, many motor skills require interpolators to
    // be swapped in. This swapping would be placed in the over-riding
    // function in the subclass. The primary use of this is for Distributed
    // ALIVE to allow synchronization of the motor systems of droids to
    // their masters. Call the parent's setup function at the beginning
    // of your over-riding function.
    virtual void staticSetup();

    // function for reading and writing state of the MotorSkill
    // to allow droids to sync with the master in Distributed ALIVE.
    // You should add whatever fields are in your subclass and then
    // call MotorSkill::toOrFromStream() at the end of your
    // over-ridden method.
    virtual bool_t toOrFromStream(XDR *xdrs);

    private:
    int _turnOn(void *arg);
    int _doImmediate(void *arg);
};

float cycleWrap(float x);

#endif
