#include <stdio.h>

#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/actions/SoRayPickAction.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/fields/SoSFNode.h>
#include "BaseSensor.h"
#include "MotorCreature.h"
#include "TimeStat.h"

/*********************************************************************
 * 
 *	    BaseSensor  Class
 * 
 *********************************************************************/


SO_NODE_SOURCE(BaseSensor);

void
BaseSensor::initClass()
{
    SO_NODE_INIT_CLASS(BaseSensor, TimeDependentComponent, "TimeDependentComponent");
}

BaseSensor::BaseSensor() : TimeDependentComponent()
{
    SO_NODE_CONSTRUCTOR(BaseSensor);
    SO_NODE_ADD_FIELD(shouldTime, (FALSE));
    swNode = NULL;
    timer = NULL;
}

BaseSensor::~BaseSensor()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
BaseSensor::readInstance(SoInput *in, unsigned short flags) 
#else
BaseSensor::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoSeparator::readInstance(in, flags)) {
#else
    if (SoSeparator::readInstance(in)) {
#endif
	SoSearchAction sa;
	sa.setType(SoTransform::getClassTypeId());
	sa.setInterest(SoSearchAction::FIRST);
	sa.apply(this);
	if(sa.getPath() != NULL){
	    sensorFrame = (SoTransform *)sa.getPath()->getTail();
	    pathToCreature = NULL;
	    pathToWorld = NULL;
	    if(shouldTime.getValue() == TRUE){
		timer = new Timestat("sensor timing", 20);
	    }
	    return (1);
	}
	else {
	    printf("NO TRANSFORM FOUND FOR BASESENSOR \n");
	    return (0);
	}
   } else {
	return(0);
   }
}

void BaseSensor::initializeComponent(SoGroup *root)
{
    TimeDependentComponent::initializeComponent(root);
    pathToWorld = buildPath(root, sensorFrame);
    swNode = (SoSwitch *)((BaseCreature *)root)->findFirstInstanceOf(SoSwitch::getClassTypeId());
}

void BaseSensor::vecFromSensorFrameUsing(SbVec3f &ep, SoPath *p)
{
    SbViewportRegion vr;
    static SoGetMatrixAction ma(vr);
    SbMatrix m;
    
    if(!p)
	printf("INVALID PATH \n");
    else {
	ma.apply(p);
	m = ma.getMatrix();
	m.multVecMatrix(ep, ep);
    }
}

SbVec3f BaseSensor::calcBearingAndRange(SbVec3f pt, float &bearing, float &range)
{
    SbVec3f localPt;
    lXw.multVecMatrix(pt, localPt);
    
    /* calc range */
    range = localPt.length();
    
    /* calc bearing */
    bearing = atan2(localPt[0], -localPt[2]);
    if(bearing>= -(M_PI/2.0))
	bearing -= M_PI/2.0;
    else 
	bearing += M_PI*1.5;
    return localPt;
}

SbVec3f BaseSensor::getVecInSensorFrame(const SbVec3f pt)
{
    SbVec3f localPt;
    lXw.multVecMatrix(pt, localPt);
    return localPt;
}

SbVec3f BaseSensor::getVecInWorldFrame(const SbVec3f pt)
{
    SbVec3f wPt;
    wXl.multVecMatrix(pt, wPt);
    return wPt;
}

void BaseSensor::vecToSensorFrameUsing(SbVec3f &ep, SoPath *p)
{
    SbViewportRegion vr;
    static SoGetMatrixAction ma(vr);
    SbMatrix m;
    if(!p)
	printf("INVALID PATH \n");
    else {   
	ma.apply(p);
	m = ma.getInverse();
	m.multVecMatrix(ep, ep);
    }
}

void BaseSensor::cacheMatrices()
{
    SbViewportRegion vr;
    static SoGetMatrixAction ma(vr);
    if(!pathToWorld)
	printf("INVALID PATH TO WORLD\n");
    else {   
	ma.apply(pathToWorld);
	wXl = ma.getMatrix();
	lXw = ma.getInverse();
    }
}

void BaseSensor::prepareForUpdateAt(UpdateStateData &usd)
{
    cacheMatrices();
    if(swNode){
	cachedSwitchValue = swNode->whichChild.getValue();
	swNode->whichChild.setValue(SO_SWITCH_NONE);
    }
}

void BaseSensor::endUpdateAt(UpdateStateData &usd)
{
    if(swNode){
	swNode->whichChild = cachedSwitchValue;
    }
}
void BaseSensor::_updateStateAt(UpdateStateData &usd)
{
    
}

void BaseSensor::updateStateAt(UpdateStateData &usd)
{
    /*
     *	    Sensors will not update themselves if the creature
     *	    is a droid
     *
     */
    
    if(((MotorCreature *)base)->runBehavior.getValue() == TRUE &
	((MotorCreature *)base)->motorCreatureType.getValue() 
	!= MotorCreature::DROID){
	if(timer)
	    timer->startTime();
	prepareForUpdateAt(usd);
	_updateStateAt(usd);
	endUpdateAt(usd);
	if(timer)
	    timer->endTime();
    }
    return;
}
