#include <stdio.h>

#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/actions/SoRayPickAction.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/fields/SoSFNode.h>
#include "RaySensor.h"
#include "TimeStat.h"
#include "../CreatureLib/CreatureLib.h"

#define RAYSENSOR_ANY_TAG -1

/*********************************************************************
 * 
 *	    RayData  Class
 * 
 *********************************************************************/
 
RayData::RayData()
{
    marker = NULL;
    validVel = FALSE;
    useRealTag = 0;
}

SoSeparator *RayData::makeMarker(SbColor col)
{
    SoSeparator *sep = new SoSeparator;
    SoBaseColor *bc = new SoBaseColor;
    bc->rgb.setValue(col);
    SoTransform *xf = new SoTransform;
    SoPickStyle *ps = new SoPickStyle;
    ps->style = SoPickStyle::UNPICKABLE;
    SoCube *cube = new SoCube;
    cube->depth = cube->height = cube->width = 2.0;
    sep->addChild(xf);
    sep->addChild(ps);
    sep->addChild(bc);
    sep->addChild(cube);
    return sep;
}

RayData::RayData(SbColor color)
{
    marker = makeMarker(color);
    validVel = FALSE;
    useRealTag = 0;
}

void RayData::setMarkerPosition(SbVec3f pos)
{
    if(marker){
	SoTransform *xf = (SoTransform *)marker->getChild(0);
	xf->translation.setValue(pos);
    }
}
  
int RayData::shootRay(SbMatrix &wXl, float min, float max, 
    SoSeparator *sep, BaseCreature *base)
{
    static Timestat *ts0 = NULL;
    static Timestat *ts1 = NULL;
    static Timestat *ts2 = NULL;
    SbViewportRegion vpr;
    static SoRayPickAction *rpa = NULL;
    
    if(!rpa){
	rpa = new SoRayPickAction(vpr);
	rpa->setPickAll(FALSE);
	ts0 = new Timestat("total time", 10);
	ts1 = new Timestat("ray cast time", 10);
	ts2 = new Timestat("ray cast", 10);
    }
    /* transform ray into world frame */
    wXl.multLineMatrix(rayLocal, rayWorld);
 
    /* create SoRayPickAction and send ray into world */
    rpa->setRay(rayWorld.getPosition(), rayWorld.getDirection(), min, max);

    rpa->apply(sep);    
    
    /* get point list */
    SoPickedPoint *pp = rpa->getPickedPoint(0);
    if(!pp){
	pt = SbVec3f(1e6,1e6, 1e6);
	return 0; 
    }
    else {
	pt = pp->getPoint();
	range = (pt - rayWorld.getPosition()).length();
//	printf("intersection found at %f %f %f \n", pt[0],
//	    pt[1], pt[2]);
    return 1; 
    }
}

int RayData::shootRayUsingCache(SbMatrix &wXl, float min, float max, 
    CreatureCache *cc, BaseCreature *base)
{
    /* transform ray into world frame */
    wXl.multLineMatrix(rayLocal, rayWorld);
 
    int i;
    BaseCreature *bc;
    BaseCreature *oldWho;
    SbVec3f cpt, vpt;
    float td;
 
    range = max;
    // save previous position for velocity, if valid
    oldWho = who;
    vpt = pt;
	
    pt = SbVec3f(1000, 1000, 1000);
    who = NULL;
    density = 0;
    
    for(i=0;i<cc->creatureCount;i++){
	bc = (BaseCreature *)cc->creatures[i];
	if((bc != base)&&(testFunction(bc, cpt))){
	    td = (cpt - rayWorld.getPosition()).length();
	    if((td < range)&&(td<max)&&(td>min)){
		pt = cpt;
		range = td;
		who = bc;
	    }
	}
    }
    if(who && oldWho){
	vel = pt - vpt;
	validVel = TRUE;
    }
    else 
	validVel = FALSE;
    
    density = 1.0 - ((range - min)/(max-min));
    return (range<max)?1:0;
}

SbBool RayData::testFunction(BaseCreature *bc,  SbVec3f &cpt)
{
    return (bc->testRay(rayWorld, cpt)>0)? TRUE:FALSE;
}

/*********************************************************************
 * 
 *	    TagRayData  Class
 * 
 *********************************************************************/
TagRayData::TagRayData()
    :RayData() 
{
    filterTag = -100;
} 

SbBool TagRayData::testFunction(BaseCreature *bc,  SbVec3f &cpt)
{
    int tagToUse = (useRealTag)
	    ? bc->realTag.getValue() 
	    : bc->tag.getValue();
	
    if(tagToUse != filterTag)
	return FALSE;
    else {
	cpt = bc->position(); // bc->calcCenter(cpt);
	return TRUE;
    }
}

/*********************************************************************
 * 
 *	    RaySensor  Class
 * 
 *********************************************************************/


SO_NODE_SOURCE(RaySensor);

void
RaySensor::initClass()
{
    SO_NODE_INIT_CLASS(RaySensor, BaseSensor, "BaseSensor");
}

RaySensor::RaySensor() : BaseSensor()
{
    SO_NODE_CONSTRUCTOR(RaySensor);
    SO_NODE_ADD_FIELD(leftLimit, (-M_PI/2.0));
    SO_NODE_ADD_FIELD(rightLimit, (M_PI/2.0));
    SO_NODE_ADD_FIELD(sensorAxis, (0, 1, 0));
    SO_NODE_ADD_FIELD(sensorDir, (1, 0, 0));
    SO_NODE_ADD_FIELD(numRays, (15));
    SO_NODE_ADD_FIELD(min, (1.0));
    SO_NODE_ADD_FIELD(max, (100.0));
    SO_NODE_ADD_FIELD(showRays, (FALSE));
    SO_NODE_ADD_FIELD(useRealTag, (FALSE));
    swNode = NULL;
    markerRoot = NULL;
}

RaySensor::~RaySensor()
{
}

void RaySensor::buildRays()
{
    SbRotation r;
    int i;
    float delta;
    float ang = leftLimit.getValue();
    SbVec3f xx;
    RayData *rd;
    if((int)numRays.getValue() > 1)
	delta = (rightLimit.getValue() - leftLimit.getValue())/
	    (numRays.getValue()-1);
    
    float cdelta = 1.0/numRays.getValue();
    SbColor col;
    sensorInfo = new RayData[numRays.getValue()];
    for(i=0;i<numRays.getValue();i++){
	r = SbRotation(sensorAxis.getValue(), -ang);
	r.multVec(sensorDir.getValue(), xx);
	//printf("%f %f %f %f \n",ang, xx[0], xx[1], xx[2]);
	col.setHSVValue(cdelta*i, .8, .8);
	sensorInfo[i].marker = sensorInfo[i].makeMarker(col);
	sensorInfo[i].bearing = ang;
	sensorInfo[i].rayLocal = SbLine(SbVec3f(0, 0, 0), xx);
	markerRoot->addChild(sensorInfo[i].marker);
	
	ang += delta; 
    }
}

SbBool	
#ifdef SO_VERSION_REVISION
RaySensor::readInstance(SoInput *in, unsigned short flags) 
#else
RaySensor::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (BaseSensor::readInstance(in, flags)) {
#else
    if (BaseSensor::readInstance(in)) {
#endif
	markerRoot = (SoSeparator *)new SoSwitch;
	if(showRays.getValue()!= FALSE)
	    ((SoSwitch *)markerRoot)->whichChild.setValue(SO_SWITCH_ALL);
	buildRays();
	return (1);
    }
    else {
	return(0);
   }
}

RayData *RaySensor::getDataForClosestWithTag(int tag)
{
    return getDataForClosestWithTag(tag, 0, numRays.getValue());
}

RayData *RaySensor::getDataForClosest()
{
    return getDataForClosestWithTag(RAYSENSOR_ANY_TAG, 0, numRays.getValue());
}

RayData *RaySensor::getDataForClosest(int startRay, int endRay)
{
    return getDataForClosestWithTag(RAYSENSOR_ANY_TAG, startRay, endRay);
}

RayData *RaySensor::getDataForClosestWithTag(int tag, int startRay, int endRay)
{
    RayData *brd=NULL;
    float md = 1.0e6;
    int i;
    int tagToUse;
    
    for(i=startRay;i<endRay;i++){
	if((sensorInfo[i].who != NULL) &&
	    (sensorInfo[i].range < md)){
		tagToUse = (useRealTag.getValue())
		    ? sensorInfo[i].who->realTag.getValue()
		    : sensorInfo[i].who->tag.getValue();
		if((tag == RAYSENSOR_ANY_TAG) ||
		    (tagToUse == tag)){
		    md = sensorInfo[i].range;
		    brd = &sensorInfo[i];
		}
	}
    }
    return brd;
}

int RaySensor::indexOfClosestRayToBearing(float whatBear, float &delta)
{
    float bd = 1.0e6, td;
    int ti, i;  
    
    for(i=0;i<numRays.getValue();i++){
	td = fabs(sensorInfo[i].bearing - whatBear);
	if(td<bd){
	    ti = i;
	    bd = td;
	}
    }
    delta = bd;
    return ti;	    
}

float RaySensor::calcAvgDensity(int startRay, int endRay)
{
    int i;
    float ad = 0;
    
    for(i=startRay;i< endRay;i++){
	ad +=  sensorInfo[i].density;
    }
    return ad /(float)(endRay-startRay);      
}

float RaySensor::calcAvgExpDensity(int startRay, int endRay)
{
    int i;
    float ad = 0, td;
    
    for(i=startRay;i< endRay;i++){
	td = sensorInfo[i].density;
	ad += (exp(-(1.0-td))-.364)/(1.0-.364);
    }
    return ad /(float)(endRay-startRay);      
}

void RaySensor::calcSummaryData()
{
    int i, ci = -1;
    float mr = max.getValue();
    for(i=0;i<numRays.getValue();i++){
	if(sensorInfo[i].range<mr)
	    ci = i;
    }
    if(ci > -1){
	detected = TRUE;
	range = sensorInfo[ci].range;
	bearing = sensorInfo[ci].bearing;
	closestPt = sensorInfo[ci].pt;
    }
    else{
	detected = FALSE;
	range = max.getValue();
    }
}

void RaySensor::_updateStateAt(UpdateStateData &usd)
{
    int i;
    RayData *rd;
    static int ft = 1;
    
    if(ft){
	((SoSeparator *)usd.currSceneGraph)->addChild(markerRoot);
	ft = 0;
    }
    
    currCreatureCache = usd.currCreatureCache;  
    
    for(i=0;i<numRays.getValue();i++){
	if(sensorInfo[i].shootRayUsingCache(wXl, min.getValue(),max.getValue(), 
	    currCreatureCache, (BaseCreature *)base)){
		sensorInfo[i].setMarkerPosition(sensorInfo[i].pt);
	    }
	else {
	    sensorInfo[i].setMarkerPosition(SbVec3f(1000, 1000, 1000));
	}
    }
    calcSummaryData();
    return;
}


/*********************************************************************
 * 
 *	    SniffSensor  Class
 * 
 *********************************************************************/


SO_NODE_SOURCE(SniffSensor);

void
SniffSensor::initClass()
{
    SO_NODE_INIT_CLASS(SniffSensor, RaySensor, "RaySensor");
}

SniffSensor::SniffSensor() : RaySensor()
{
    SO_NODE_CONSTRUCTOR(SniffSensor);
    SO_NODE_ADD_FIELD(tags, (900));
}

SniffSensor::~SniffSensor()
{
}

void SniffSensor::buildRays()
{
    int i;
    SbColor col;
    const int32_t *theTags = tags.getValues(0);
    sensorInfo = new TagRayData[getNumTags()];
    float cdelta = 1.0/(float)getNumTags();
    numRays.setValue(getNumTags());
    
    for(i=0;i<getNumTags();i++){
	col.setHSVValue(cdelta*i, .8, .8);
	sensorInfo[i].rayLocal = SbLine(SbVec3f(0, 0, 0), SbVec3f(1, 0, 0));
	sensorInfo[i].marker = sensorInfo[i].makeMarker(col);
	markerRoot->addChild(sensorInfo[i].marker);
	sensorInfo[i].setFilterTag(theTags[i]);
	sensorInfo[i].useRealTag = (useRealTag.getValue() == TRUE)?1:0; 
    }
}

SbBool	
#ifdef SO_VERSION_REVISION
SniffSensor::readInstance(SoInput *in, unsigned short flags) 
#else
SniffSensor::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (BaseSensor::readInstance(in, flags)) {
#else
    if (BaseSensor::readInstance(in)) {
#endif
	markerRoot = (SoSeparator *)new SoSwitch;
	if(showRays.getValue()!= FALSE)
	    ((SoSwitch *)markerRoot)->whichChild.setValue(SO_SWITCH_ALL);
	buildRays();
	return (1);
    }
    else {
	return(0);
   }
}

void SniffSensor::_updateStateAt(UpdateStateData &usd)
{
    int i;
    static int ft = 1;
    
    if(ft){
	((SoSeparator *)usd.currSceneGraph)->addChild(markerRoot);
	ft = 0;
    }
    
    currCreatureCache = usd.currCreatureCache;  
    
    for(i=0;i<getNumTags();i++){
	if(sensorInfo[i].shootRayUsingCache(wXl, min.getValue(),max.getValue(), 
	currCreatureCache, (BaseCreature *)base)){
	    sensorInfo[i].setMarkerPosition(sensorInfo[i].pt);
	    
	    /* do bearing */
	    SbVec3f localPt;
	    float tmpbearing;
	    
	    lXw.multVecMatrix(sensorInfo[i].pt, localPt);
	    sensorInfo[i].bearing =tmpbearing = atan2(localPt[0], -localPt[2]);
	    if(tmpbearing>= -(M_PI/2.0))
		sensorInfo[i].bearing -= M_PI/2.0;
	    else 
		sensorInfo[i].bearing += M_PI*1.5;
	}
    }
    calcSummaryData();
    return;
}
