#ifndef _RAYSENSOR_
#define _RAYSENSOR_

#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/fields/SoMFVec3f.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoMFInt32.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/nodes/SoTransform.h>
#include "BaseSensor.h"
#include <Inventor/SbColor.h>
class SoPath;
class SoSwitch;
class CreatureCache;

class RayData  {
    public:
    float bearing;
    SbLine rayLocal;
    SbLine rayWorld;
    SbVec3f pt;
    SbVec3f vel;
    SbBool validVel;
    BaseCreature *who;
    float range;
    float density;  // varies between 1 (range = min) and 0 (range = max)
    SoSeparator *marker;
    int useRealTag;
    
    RayData();
    RayData(SbColor color);
    SoSeparator *makeMarker(SbColor col);
    void setMarkerPosition(SbVec3f pos);
    int shootRay(SbMatrix &wXl, float min, float max, SoSeparator *sep, BaseCreature *base);
    int shootRayUsingCache(SbMatrix &wXl, float min, float max, 
	CreatureCache *cc, BaseCreature *base);
    virtual SbBool testFunction(BaseCreature *bc,  SbVec3f &pt);
    
    int filterTag; // not used by RayData but is by TagRayData below
    void setFilterTag(const int32_t whichFilter){filterTag = (int)whichFilter;}
};

class TagRayData : public RayData  {
    public:
    TagRayData();
    virtual SbBool testFunction(BaseCreature *bc,  SbVec3f &pt);
};

class RaySensor : public BaseSensor {
    SO_NODE_HEADER(RaySensor);

public:
    SoSFFloat leftLimit;
    SoSFFloat rightLimit;
    SoSFVec3f sensorAxis;
    SoSFVec3f sensorDir;
    SoSFInt32 numRays;
    SoSFFloat min;
    SoSFFloat max;
    SoSFBool showRays;
    SoSFBool useRealTag;
    
    RayData *sensorInfo;
    
    static void initClass();
    
    int indexOfClosestRayToBearing(float whatBear, float &delta);
    
    /* these methods check all rays, they call virtual function below */
    RayData *getDataForClosestWithTag(int tag);
    RayData *getDataForClosest();
    
    /* these methods only check within a range */
    virtual RayData *getDataForClosestWithTag(int tag, int startRay, int endRay);
    RayData *getDataForClosest(int startRay, int endRay);
    
    /* these methods calc the average density on the selected rays.
     * The density ranges from 1 to 0.
     */
    float calcAvgDensity(int startRay, int endRay);
    float calcAvgExpDensity(int startRay, int endRay);
    
    /* this method is called auto-magically and causes all rays in the rayCache
     * to be updated. The methods above use the most recent data in the rayCache
     * and do not actually do any new sensing.
    */
    virtual void _updateStateAt(UpdateStateData &usd); // over-ride this method
    virtual void buildRays();
    

protected:
    SoSeparator *markerRoot;
    CreatureCache *currCreatureCache;
    
    RaySensor();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    virtual SbBool	readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool	readInstance(SoInput *in);
#endif
    void calcSummaryData();
    virtual ~RaySensor();

};

class SniffSensor : public RaySensor {
    SO_NODE_HEADER(SniffSensor);

public:
    SoMFInt32 tags;
    int getNumTags() {return tags.getNum();}
    
    static void initClass();
    
    virtual void _updateStateAt(UpdateStateData &usd); // over-ride this method
    virtual void buildRays();
    

protected:
    SniffSensor();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    virtual SbBool	readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool	readInstance(SoInput *in);
#endif
    virtual ~SniffSensor();

};    
#endif

