#ifdef __GNUG__
extern "C" {
#endif
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <strings.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netdb.h>
#ifdef __GNUG__
}
#endif
#ifdef __sgi
#include <bstring.h>
#endif
#include <iostream.h>
#include "SocketBase.h"
#include "SocketGlobals.h"

#define SB_DEBUG_VERBOSE 0

/**************************************************
  SocketBase::SocketBase              *** PROTECTED

  Default constructor for the SocketBase class.
  Only used implicitly in constructors for
  derived classes (SocketMan, SocketClient).
  *************************************************/

SocketBase::SocketBase()
{  
  global_message = 0;
  quiet = 0;
  use_udp = 0;

  if (sixsixsix != htonl(sixsixsix))
    {
      // I am little endian.
#if SB_DEBUG_VERBOSE
      cerr << "I am little endian." << endl;
#endif
      little_endian = 1;
    }
  else
    {
      little_endian = 0;
    }
}

/**************************************************
  SocketBase::~SocketBase             *** PROTECTED

  Null destructor.
  *************************************************/

SocketBase::~SocketBase()
{
}

/**************************************************
  SocketBase::timevalDifference       *** PROTECTED

  Passed in timeval structures tv2 and tv1, returns
  the number of seconds (as a float) elapsed
  between them.
  *************************************************/

float
SocketBase::timevalDifference(const struct timeval& tv2,
			      const struct timeval& tv1)
{
  return (float) (tv2.tv_sec - tv1.tv_sec) + (float) (tv2.tv_usec - tv1.tv_usec) * (float) (1.0e-6);
}

/**************************************************
  SocketBase::longToDotFormat           *** PRIVATE

  Converts an internet address (normally stored in
  a long) into dot format. For diagnostic purposes
  only.
  *************************************************/

char *
SocketBase::longToDotFormat(long addr)
{
  char *temp_s = new char[20];

  sprintf(temp_s, "%d.%d.%d.%d", ((addr & 0xff000000) >> 24),
	  ((addr & 0x00ff0000) >> 16), ((addr & 0x0000ff00) >> 8),
	  (addr & 0x000000ff));

  return temp_s;
}

/**************************************************
  SocketBase::pollOnFDWithTimeout     *** PROTECTED

  Executes a non blocking select on the file
  descriptor given in poll_fd. Returns
  MSG_SOCKET_READY_TO_READ or
  MSG_SOCKET_NOT_READY_TO_READ depending on the
  condition of the socket.

  Errors:
  ERR_DURING_SELECT
  *************************************************/

int
SocketBase::pollOnFDWithTimeout(int poll_fd, struct timeval timeout)
{
  struct fd_set readfds;

  FD_ZERO(&readfds);
  FD_SET(poll_fd, &readfds);
  
#ifndef __hpux
  if (select(1+poll_fd, &readfds, NULL, NULL, &timeout) < 0)
#else
  if (select(1+poll_fd, (int *) &readfds, NULL, NULL, &timeout) < 0)
#endif
    {
      if (!quiet)
	{
	  cerr << "SocketBase::pollOnFDWithTimeout: ";
	  cerr << "error during select()" << endl;
	}
      global_message = ERR_DURING_SELECT;
      return ERR_DURING_SELECT;
    }
  else
    {
      if (FD_ISSET(poll_fd, &readfds))
	{
	  global_message = MSG_SOCKET_READY_TO_READ;
	  return MSG_SOCKET_READY_TO_READ;
	}
      else
	{
	  global_message = MSG_SOCKET_NOT_READY_TO_READ;
	  return MSG_SOCKET_NOT_READY_TO_READ;
	}
    }
}

/**************************************************
  SocketBase::bytesAvailableOnFD      *** PROTECTED

  Returns the number of bytes available to be read
  on the given socket.
  *************************************************/

int
SocketBase::bytesAvailableOnFD(int fd)
{
  int how_many_bytes;

  ioctl(fd, FIONREAD, &how_many_bytes);
  return how_many_bytes;
}

int
SocketBase::byteSwapLongBufferFromNet(SocketBuf *sb,
				      int data_size)
{
  int index;

  for (index = 0; index < sb->num_bytes; index += data_size)
    {
      *((int *) (sb->buf + index)) = ntohl(*((int *) (sb->buf + index)));
    }

  if (sb->num_bytes % data_size)
    {
      return ERR_DATA_TYPE_DOES_NOT_MATCH_NUM_BYTES;
    }
  else
    {
      return 0;
    }
}

int
SocketBase::byteSwapLongBufferToNet(SocketBuf *sb,
				    int data_size)
{
  int index;

  for (index = 0; index < sb->num_bytes; index += data_size)
    {
      *((int *) (sb->buf + index)) = htonl(*((int *) (sb->buf + index)));
    }

  if (sb->num_bytes % data_size)
    {
      return ERR_DATA_TYPE_DOES_NOT_MATCH_NUM_BYTES;
    }
  else
    {
      return 0;
    }
}

/**************************************************
  SocketBase::readPacketLength        *** PROTECTED

  Attempts to read the length of the packet coming
  down the socket (sizeof(int) bytes). Returns a
  positive number indicating the number of bytes to
  follow if successful, or -1 otherwise. Returns
  MSG_NO_ERROR in global_message and in the current
  socket's last_message field if successful.

  Errors stored in last_message:
  ERR_TIMEOUT_READING_PACKET_LENGTH

  Errors stored ONLY in global_message:
  ERR_SIGPIPE_READING_PACKET_LENGTH
  *************************************************/

int
SocketBase::readPacketLength(SocketStruct *current_socket,
			      float timeout)
{
  int nbytes;
  int bufptr = 0;
  int left;
  static sm_byte lengthbuf[sizeof(int)];
  int expiring;
  struct timeval current_time;
  struct timezone current_tz = {0, 0};
  float time_difference;

  expiring = 0;

  left = sizeof(int);
#if SB_DEBUG_VERBOSE
  cerr << "Reading packet length..." << endl;
#endif
  while(left > 0 && !global_sigpipe_raised)
    {
      nbytes = read(current_socket->fd, lengthbuf+bufptr, left);
#if SB_DEBUG_VERBOSE
      if (nbytes >= 0)
	cerr << "nbytes = " << nbytes << endl;
#endif
      if (nbytes <= 0 && !expiring)
	{
	  gettimeofday(&current_socket->last_time_accessed, &current_tz);
	  expiring = 1;
#if SB_DEBUG_VERBOSE
	  cerr << "nbytes = " << nbytes << ", expiring..." << endl;
#endif
	}
      else
	{
	  if (nbytes > 0)
	    {
	      expiring = 0;
	      bufptr += nbytes;
	      left -= nbytes;
	    }
	}
      if (expiring)
	{
	  gettimeofday(&current_time, &current_tz);
	  time_difference = timevalDifference(current_time, current_socket->last_time_accessed);
	  if (time_difference > timeout)
	    {
	      global_message = ERR_TIMEOUT_READING_PACKET_LENGTH;
	      current_socket->last_message = ERR_TIMEOUT_READING_PACKET_LENGTH;
	      SocketBuf* tbuf = new SocketBuf;
	      tbuf->num_bytes = sizeof(int);
	      tbuf->num_bytes_successful = tbuf->num_bytes - left;
	      tbuf->buf = new sm_byte[tbuf->num_bytes];
	      bcopy(lengthbuf, tbuf->buf, tbuf->num_bytes_successful);
	      current_socket->error_buf = tbuf;
#if SB_DEBUG_VERBOSE
	      cerr << "expired!" << endl;
	      cerr << "read packet length failed" << endl;
#endif
	      return -1;
	    }
	}
    }
  
  if (global_sigpipe_raised)
    {
      if (!quiet)
	cerr << "read packet length interrupted by SIGPIPE" << endl;
      global_message = ERR_SIGPIPE_READING_PACKET_LENGTH;
      return -1;
    }

  left =  *((int *) lengthbuf);

  if (little_endian)
    left = ntohl(left);

  return left;
}

SocketBuf *
SocketBase::readPacketDataType(SocketStruct *current_socket,
			       int bytes_in_packet,
			       float timeout)
{
  int nbytes;
  int left;
  int expiring;
  char temp;
  struct timeval current_time;
  struct timezone current_tz = {0, 0};
  float time_difference;

  SocketBuf *tbuf = new SocketBuf;
  tbuf->buf = new sm_byte[bytes_in_packet];
  tbuf->num_bytes = bytes_in_packet;
  tbuf->num_bytes_successful = 0;
  tbuf->type = -1;

  left = sizeof(sm_byte);   // had better be 1

  expiring = 0;

  while (left > 0 && !global_sigpipe_raised)
    {
      nbytes = read(current_socket->fd, &temp, left);

      if (nbytes <= 0 && !expiring)
	{
	  gettimeofday(&current_socket->last_time_accessed, &current_tz);
	  expiring = 1;
#if SB_DEBUG_VERBOSE
	  cerr << "nbytes = " << nbytes << ", expiring..." << endl;
#endif
	}
      else
	{
	  if (nbytes > 0)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "nbytes = " << nbytes << endl;
#endif
	      expiring = 0;
	      left -= nbytes;
	    }
	}
      if (expiring)
	{
	  gettimeofday(&current_time, &current_tz);
	  time_difference = timevalDifference(current_time, current_socket->last_time_accessed);
	  if (time_difference > timeout)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "expired!" << endl;
#endif
	      current_socket->error_buf = tbuf;
	      global_message = ERR_TIMEOUT_READING_PACKET_DATA_TYPE;
	      current_socket->last_message = ERR_TIMEOUT_READING_PACKET_DATA_TYPE;
	      return NULL;
	    }
	}
    }  

  if (global_sigpipe_raised)
    {
      if (!quiet)
	cerr << "readPacketDataType was interrupted by SIGPIPE" << endl;
      global_message = ERR_SIGPIPE_READING_PACKET_DATA_TYPE;
      return NULL;
    }

  tbuf->type = temp;

  return tbuf;
}

/**************************************************
  SocketBase::readPacketData          *** PROTECTED

  Attempts to read sbuf->num_bytes bytes from the
  socket contained in current_socket within
  timeout seconds. Returns a pointer to the buffer
  the data was read into if successful; returns
  NULL otherwise and sets global_message or
  current_socket->last_message appropriately.

  Errors stored in last_message:
  ERR_TIMEOUT_READING_PACKET

  Errors stored ONLY in global_message:
  ERR_SIGPIPE_READING_PACKET
  *************************************************/

SocketBuf*
SocketBase::readPacketData(SocketStruct *current_socket,
			   SocketBuf *sbuf,
			   float timeout)
{
  int nbytes;
  int bufptr = 0;
  int left;
  int expiring;
  struct timeval current_time;
  struct timezone current_tz = {0, 0};
  float time_difference;

  left = sbuf->num_bytes;
  
  expiring = 0;
  
#if SB_DEBUG_VERBOSE
  cerr << "Reading packet data" << endl;
#endif
  
  while(left > 0 && !global_sigpipe_raised)
    {
      nbytes = read(current_socket->fd, sbuf->buf+bufptr, left);
      if (nbytes <= 0 && !expiring) 
	{
	  gettimeofday(&current_socket->last_time_accessed, &current_tz);
	  expiring = 1;
#if SB_DEBUG_VERBOSE
	  cerr << "nbytes = " << nbytes << ", expiring..." << endl;
#endif
	}
      else
	{
	  if (nbytes > 0)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "nbytes = " << nbytes << endl;
#endif
	      expiring = 0;
	      bufptr += nbytes;
	      left -= nbytes;
	    }
	}
      if (expiring)
	{
	  gettimeofday(&current_time, &current_tz);
	  time_difference = timevalDifference(current_time, current_socket->last_time_accessed);
	  if (time_difference > timeout)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "expired!" << endl;
#endif
	      sbuf->num_bytes_successful = sbuf->num_bytes - left;
	      global_message = ERR_TIMEOUT_READING_PACKET;
	      current_socket->last_message = ERR_TIMEOUT_READING_PACKET;
	      current_socket->error_buf = sbuf;
	      return NULL;
	    }
	}
    }

  if (global_sigpipe_raised)
    {
      if (!quiet)
	cerr << "readPacketData was interrupted by SIGPIPE" << endl;
      global_message = ERR_SIGPIPE_READING_PACKET;
      return NULL;
    }

  sbuf->num_bytes_successful = sbuf->num_bytes;
		  
  if (little_endian)
    {
      switch (sbuf->type)
	{
	case SOCKET_CHAR:
	  {
	    break;   // no byte swapping necessary
	  }
	  
	case SOCKET_INT:
	case SOCKET_FLOAT:
	  {
	    // Need to swap.
	    byteSwapLongBufferFromNet(sbuf, sizeof(int));
	    break;
	  }
	}
    }
  return sbuf;
}  

/**************************************************
  SocketBase::_readWithPollAndTimeout *** PROTECTED

  Base function for implementation of SocketMan's
  and SocketClient's readWithPollAndTimeout and
  readWithTimeout. Necessary for generality in
  length of timeout during the poll and higher-
  level error handling (such as SIGPIPE). Returns
  a pointer to the data buffer if successful, or
  NULL otherwise (with error in global_message or
  current_socket->last_message). If socket was
  not ready to read, returns 
  MSG_SOCKET_NOT_READY_TO_READ in last_message.

  Errors stored in last_message:
  ERR_DURING_SELECT
  ERR_TIMEOUT_READING_PACKET_LENGTH  
  ERR_TIMEOUT_READING_PACKET
  ERR_UDP_RECEIVE_FAILED

  Errors stored ONLY in global_message:
  ERR_CURRENT_SOCKET_NOT_OPEN
  ERR_CURRENT_SOCKET_CLOSED
  ERR_SIGPIPE_READING_PACKET
  ERR_SIGPIPE_READING_PACKET_LENGTH
  *************************************************/

SocketBuf*
SocketBase::_readWithPollAndTimeout(SocketStruct *current_socket,
				    struct timeval poll_timeout,
				    float read_timeout_float,
				    int use_udp_for_read,
				    int use_multicast_for_read)
{
  int select_result;
  int bytes_in_packet;
  static sm_byte udp_buf[SB_MAX_UDP_PACKET_SIZE];
  static int udp_bytes_received;
  struct timeval tv1, tv2;
  struct timezone tz = {0, 0};
  float time_difference;
  int expiring;
  static struct sockaddr_in address;
  static int length;

  if (current_socket)
    {
      if (current_socket->fd <= 0)
	{
	  if (!quiet)
	    {
	      cerr << "SocketBase::readWithPollAndTimeout: ";
	      cerr << "no current socket" << endl;
	    }
	  global_message = ERR_CURRENT_SOCKET_NOT_OPEN;
	  return NULL;
	}
    }
  else
    {
      if (!quiet)
	{
	  cerr << "SocketBase::_readWithPollAndTimeout: ";
	  cerr << "no current socket" << endl;
	}
      global_message = ERR_CURRENT_SOCKET_NOT_OPEN;
      return NULL;
    }

  if (((select_result = pollOnFDWithTimeout(current_socket->fd, poll_timeout)) == ERR_DURING_SELECT) &&
      !global_sigpipe_raised)
    {
      if (!quiet)
	{
	  cerr << "SocketBase::_readWithPollAndTimeout: ";
	  cerr << "error during select()" << endl;
	}
      global_message = ERR_DURING_SELECT;
      current_socket->last_message = ERR_DURING_SELECT;
      return NULL;
    }
  
  if (select_result == MSG_SOCKET_READY_TO_READ && !global_sigpipe_raised)
    {
#if SB_DEBUG_VERBOSE
      cerr << "SocketBase::_rWPAT: socket " << current_socket->fd << " ready to read" << endl;
#endif

      int how_many_bytes;
      how_many_bytes = bytesAvailableOnFD(current_socket->fd);
#if SB_DEBUG_VERBOSE
      cerr << "Number of bytes available: " << how_many_bytes << endl;
#endif
      if (how_many_bytes == 0)
	{
	  if (!quiet)
	    {
	      cerr << "Socket " << current_socket->fd << " appears to have closed" << endl;
	    }
	  global_message = ERR_CURRENT_SOCKET_CLOSED;
	  return NULL;
	}
      
      // For cleanliness, erase the error buffer if there was one.
      // The user should have called the recover function if necessary.

      if (current_socket->error_buf)
	{
	  delete current_socket->error_buf;
	  current_socket->error_buf = NULL;
	}

      if (!use_udp_for_read)
	{
	  /* Find out how many bytes are in the packet
	     coming down the socket */
	  
	  bytes_in_packet = readPacketLength(current_socket, read_timeout_float);
	  
#if SB_DEBUG_VERBOSE
	  cerr << "bytes_in_packet: " << bytes_in_packet << endl;
#endif

	  if (bytes_in_packet > 0)
	    {
	      // Now find out what data type it is.

	      SocketBuf *tbuf;
	      tbuf = readPacketDataType(current_socket, bytes_in_packet, read_timeout_float);

	      if (tbuf)
		{
		  tbuf = readPacketData(current_socket,
					tbuf,
					read_timeout_float);
		  if (tbuf)   // read was completely successful
		    {
		      global_message = MSG_NO_ERROR;
		      current_socket->last_message = MSG_NO_ERROR;
		    }
					   
		  return tbuf;
		}
	      else
		{
		  return NULL;
		}
	    }
	  else
	    {
	      return NULL;
	    }
	}
      else
	{
	  // Using UDP, so we read the entire packet as a chunk
	  // This is assuming messages are less than the size of a UDP packet

	  expiring = 0;

	  do
	    {
	      if (use_multicast_for_read)
		{
		  length = sizeof(address);
		  udp_bytes_received = recvfrom(current_socket->fd, 
						udp_buf,
						SB_MAX_UDP_PACKET_SIZE,
						0,
						(struct sockaddr *)&address,
						&length);
		}
	      else
		udp_bytes_received = recv(current_socket->fd, udp_buf, SB_MAX_UDP_PACKET_SIZE, 0);

	      if (udp_bytes_received <= 0)
		{
		  if (!expiring)
		    {
		      gettimeofday(&tv1, &tz);
		      expiring = 1;
		    }
		  else
		    {
		      gettimeofday(&tv2, &tz);
		      time_difference = timevalDifference(tv2, tv1);
		      if (time_difference > read_timeout_float)
			{
			  if (!quiet)
			    {
			      cerr << "UDP recv failed" << endl;
			    }
			  global_message = ERR_UDP_RECEIVE_FAILED;
			  current_socket->last_message = ERR_UDP_RECEIVE_FAILED;
			  return NULL;
			}
		    }
		}
	    }
	  while(udp_bytes_received <= 0);
//	  cerr << "UDP recv succeeded: was " << udp_bytes_received << endl;
	  
	  // Got the entire packet, split it up into a SocketBuf

	  SocketBuf *sb = new SocketBuf;
	  
	  sb->num_bytes = *((int *) udp_buf);  // int is first in the buffer

	  if (use_multicast_for_read)
	    {
	      // set the from address to where it came from
	      sb->from_address = address.sin_addr.s_addr;
	    }

	  // take care of byte swapping

	  if (little_endian)
	    sb->num_bytes = ntohl(sb->num_bytes);

	  sb->type = *((char *) (udp_buf + sizeof(int)));

#if SB_DEBUG_VERBOSE
	  cerr << "UDP recv: num_bytes = " << sb->num_bytes << endl;
	  cerr << "          type = " << sb->type << endl;
#endif

	  sb->buf = new sm_byte[sb->num_bytes];
	  bcopy(udp_buf + sizeof(int) + sizeof(sm_byte), sb->buf, sb->num_bytes);

	  if (little_endian)
	    {
	      switch (sb->type)
		{
		case SOCKET_CHAR:
		  {
		    break;   // no byte swapping necessary
		  }
		  
		case SOCKET_INT:
		case SOCKET_FLOAT:
		  {
		    // Need to swap.
		    byteSwapLongBufferFromNet(sb, sizeof(int));
		    break;
		  }
		}
	    }

	  sb->num_bytes_successful = sb->num_bytes;
	  return sb;
	}
    }
  else
    {
      global_message = MSG_SOCKET_NOT_READY_TO_READ;
      current_socket->last_message = MSG_SOCKET_NOT_READY_TO_READ;
      return NULL;
    }
}

/**************************************************
  SocketBase::recoverReadPacketLength *** PROTECTED

  Recovers reading the packet length, if timeout
  occured earlier reading it. If successful, returns
  an integer > 0 indicating the number of bytes to
  follow. Otherwise returns -1 and sets one or more
  of the message fields.

  Errors stored in last_message:
  ERR_TIMEOUT_READING_PACKET_LENGTH  

  Errors stored ONLY in global_message:
  ERR_SIGPIPE_READING_PACKET_LENGTH
  *************************************************/

int
SocketBase::recoverReadPacketLength(SocketStruct *current_socket,
				     float timeout)
{
  int nbytes;
  int bufptr = 0;
  int left;
  int expiring;
  struct timeval current_time;
  struct timezone current_tz;
  float time_difference;
  SocketBuf *ptr;
  
  ptr = current_socket->error_buf;
  left = ptr->num_bytes - ptr->num_bytes_successful;
  bufptr = ptr->num_bytes_successful;
  expiring = 0;
#if SB_DEBUG_VERBOSE
  cerr << "Recovering header..." << endl;
#endif

  while(left > 0 && !global_sigpipe_raised)
    {
      nbytes = read(current_socket->fd, ptr->buf + bufptr, left);
      if (nbytes <= 0 && !expiring)
	{
	  gettimeofday(&current_socket->last_time_accessed, &current_tz);
	  expiring = 1;
#if SB_DEBUG_VERBOSE
//	  cerr << "nbytes = " << nbytes << ", expiring..." << endl;
#endif
	}
      else
	{
	  if (nbytes > 0)
	    {
	      expiring = 0;
	      bufptr += nbytes;
	      left -= nbytes;
	    }
	}
      if (expiring)
	{
	  gettimeofday(&current_time, &current_tz);
	  time_difference = timevalDifference(current_time, current_socket->last_time_accessed);
	  if (time_difference > timeout)
	    {
	      global_message = ERR_TIMEOUT_READING_PACKET_LENGTH;

	      // update the values in the (already existing) error buffer

	      current_socket->last_message = ERR_TIMEOUT_READING_PACKET_LENGTH;
	      current_socket->error_buf->num_bytes_successful = sizeof(int) - left;
#if SB_DEBUG_VERBOSE
	      cerr << "expired!" << endl;
#endif
	      return -1;
	    }
	}
    }
  
  if (global_sigpipe_raised)
    {
      if (!quiet)
	cerr << "recoverReadPacketLength was interrupted by SIGPIPE" << endl;
      global_message = ERR_SIGPIPE_READING_PACKET_LENGTH;
      return -1;
    }

  // We were successful in reading the packet length, so
  // we'll delete the error buffer.
  left =  *((int *) ptr->buf);

  global_message = MSG_NO_ERROR;
  current_socket->last_message = MSG_NO_ERROR;
  delete current_socket->error_buf;
  current_socket->error_buf = NULL;

  if (little_endian)
    left = ntohl(left);

#if SB_DEBUG_VERBOSE
  cerr << "Number of bytes in coming packet: " << left << endl;
#endif
  return left;
}

SocketBuf *
SocketBase::recoverReadPacketDataType(SocketStruct *current_socket,
				      float timeout)
{
  int nbytes;
  int left;
  int expiring;
  char temp;
  struct timeval current_time;
  struct timezone current_tz = {0, 0};
  float time_difference;

#if SB_DEBUG_VERBOSE
  cerr << "Recovering packet data type..." << endl;
#endif

  left = sizeof(sm_byte);   // had better be 1

  expiring = 0;

  while (left > 0 && !global_sigpipe_raised)
    {
      nbytes = read(current_socket->fd, &temp, left);

      if (nbytes <= 0 && !expiring)
	{
	  gettimeofday(&current_socket->last_time_accessed, &current_tz);
	  expiring = 1;
#if SB_DEBUG_VERBOSE
//	  cerr << "nbytes = " << nbytes << ", expiring..." << endl;
#endif
	}
      else
	{
	  if (nbytes > 0)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "nbytes = " << nbytes << endl;
#endif
	      expiring = 0;
	      left -= nbytes;
	    }
	}
      if (expiring)
	{
	  gettimeofday(&current_time, &current_tz);
	  time_difference = timevalDifference(current_time, current_socket->last_time_accessed);
	  if (time_difference > timeout)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "expired!" << endl;
#endif
	      global_message = ERR_TIMEOUT_READING_PACKET_DATA_TYPE;
	      current_socket->last_message = ERR_TIMEOUT_READING_PACKET_DATA_TYPE;
	      return NULL;
	    }
	}
    }  

  if (global_sigpipe_raised)
    {
      if (!quiet)
	cerr << "recoverReadPacketDataType was interrupted by SIGPIPE" << endl;
      global_message = ERR_SIGPIPE_READING_PACKET_DATA_TYPE;
      return NULL;
    }

  SocketBuf *ptr;
  ptr = current_socket->error_buf;
  current_socket->error_buf = NULL;
  ptr->type = temp;
  return ptr;
}

/**************************************************
  SocketBase::recoverReadPacketData   *** PROTECTED

  Recovers reading the packet length, if timeout
  occured earlier reading it. If successful, returns
  an integer > 0 indicating the number of bytes to
  follow. Otherwise returns -1 and sets one or more
  of the message fields.

  Errors stored in last_message:
  ERR_TIMEOUT_READING_PACKET

  Errors stored ONLY in global_message:
  ERR_SIGPIPE_READING_PACKET
  *************************************************/

SocketBuf*
SocketBase::recoverReadPacketData(SocketStruct *current_socket, 
				  float timeout)
{
  int nbytes;
  int bufptr = 0;
  int left;
  int expiring;
  struct timeval current_time;
  struct timezone current_tz = {0, 0};
  float time_difference;
  SocketBuf *ptr;
  
  ptr = current_socket->error_buf;
  left = ptr->num_bytes - ptr->num_bytes_successful;
  bufptr = ptr->num_bytes_successful;
  expiring = 0;
#if SB_DEBUG_VERBOSE
  cerr << "Recovering packet data..." << endl;
#endif
  
  while(left > 0 && !global_sigpipe_raised)
    {
      nbytes = read(current_socket->fd, ptr->buf+bufptr, left);
      if (nbytes <= 0 && !expiring) 
	{
	  gettimeofday(&current_socket->last_time_accessed, &current_tz);
	  expiring = 1;
#if SB_DEBUG_VERBOSE
//	  cerr << "nbytes = " << nbytes << ", expiring..." << endl;
#endif
	}
      else
	{
	  if (nbytes > 0)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "nbytes = " << nbytes << endl;
#endif
	      expiring = 0;
	      bufptr += nbytes;
	      left -= nbytes;
	    }
	}
      if (expiring)
	{
	  gettimeofday(&current_time, &current_tz);
	  time_difference = timevalDifference(current_time, current_socket->last_time_accessed);
	  if (time_difference > timeout)
	    {
#if SB_DEBUG_VERBOSE
	      cerr << "expired!" << endl;
#endif
	      
	      // update the values in the (already existing) error buffer

	      ptr->num_bytes_successful = ptr->num_bytes - left;
	      global_message = ERR_TIMEOUT_READING_PACKET;
	      current_socket->last_message = ERR_TIMEOUT_READING_PACKET;
	      return NULL;
	    }
	}
    }

  if (!global_sigpipe_raised)
    {
      global_message = MSG_NO_ERROR;
      current_socket->last_message = MSG_NO_ERROR;
      current_socket->error_buf = NULL;	// reassign the error buffer to NULL
      					// now that we've completed the read

      ptr->num_bytes_successful = ptr->num_bytes;

      if (little_endian)
	{
	  switch (ptr->type)
	    {
	    case SOCKET_CHAR:
	      {
		break;   // no byte swapping necessary
	      }
	      
	    case SOCKET_INT:
	    case SOCKET_FLOAT:
	      {
		// Need to swap.
		byteSwapLongBufferFromNet(ptr, sizeof(int));
		break;
	      }
	    }
	}

      return ptr;
    }
  else
    {
      if (!quiet)
	cerr << "recoverPacketData was interrupted by SIGPIPE" << endl;
      global_message = ERR_SIGPIPE_READING_PACKET;
      return NULL;
    }
}  

/**************************************************
  SocketBase::_recoverFromReadError   *** PROTECTED

  Attempts to recover from the previous read error
  on the current socket, if there was one. Timeouts
  are the only error from which you can recover.
  Returns a pointer to the buffer in which the
  data was read if successful. If there was no
  data to read, returns NULL and sets last_message
  to MSG_SOCKET_NOT_READY_TO_READ.

  NOTE: Only works with TCP sockets.

  Errors stored in last_message:
  ERR_DURING_SELECT
  ERR_TIMEOUT_READING_PACKET
  ERR_TIMEOUT_READING_PACKET_LENGTH

  Errors stored ONLY in global_message:
  ERR_CURRENT_SOCKET_NOT_OPEN
  ERR_CURRENT_SOCKET_CLOSED
  ERR_SIGPIPE_READING_PACKET
  ERR_SIGPIPE_READING_PACKET_LENGTH
  *************************************************/

SocketBuf*
SocketBase::_recoverFromReadError(SocketStruct *current_socket,
				  float read_timeout_float)

{
  int bytes_in_packet;
  int select_result;
  struct timeval poll_timeout = {0, 0};

  if (current_socket)
    {
      if (current_socket->fd <= 0)
	{
	  if (!quiet)
	    {
	      cerr << "SocketMan::recoverFromReadError: ";
	      cerr << "no current socket" << endl;
	    }
	  global_message = ERR_CURRENT_SOCKET_NOT_OPEN;
	  return NULL;
	}
    }
  else
    {
      if (!quiet)
	{
	  cerr << "SocketMan::recoverFromReadError: ";
	  cerr << "no current socket" << endl;
	}
      global_message = ERR_CURRENT_SOCKET_NOT_OPEN;
      return NULL;
    }
  
  switch (current_socket->last_message)
    {
    case MSG_NO_ERROR:
      {
	return NULL;
      }
      
    case ERR_TIMEOUT_READING_PACKET_LENGTH:
      {
	// need to finish reading the packet length
	// and the data from the socket if we can

	// see if any data is ready to read
	
	if ((select_result = pollOnFDWithTimeout(current_socket->fd, poll_timeout)) == ERR_DURING_SELECT)
	  {
	    if (!quiet)
	      {
		cerr << "SocketBase::_recoverFromReadError: ";
		cerr << "error during select()" << endl;
	      }
	    global_message = ERR_DURING_SELECT;
	    current_socket->last_message = global_message;
	    return NULL;
	  }

	if (select_result == MSG_SOCKET_READY_TO_READ)
	  {
#if SB_DEBUG_VERBOSE
	    cerr << "SocketBase::_recoverFromReadError: socket " << current_socket->fd << " ready to read" << endl;
#endif	    
	    int how_many_bytes;
	    how_many_bytes = bytesAvailableOnFD(current_socket->fd);
#if SB_DEBUG_VERBOSE
	    cerr << "Number of bytes available: " << how_many_bytes << endl;
#endif
	    if (how_many_bytes == 0)
	      {
		if (!quiet)
		  {
		    cerr << "Socket " << current_socket->fd << " appears to have closed" << endl;
		  }
		global_message = ERR_CURRENT_SOCKET_CLOSED;
		return NULL;
	      }
	    
	    bytes_in_packet = recoverReadPacketLength(current_socket,
						      read_timeout_float);

	    if (bytes_in_packet > 0)
	      {
		SocketBuf *tbuf;
		tbuf = readPacketDataType(current_socket, bytes_in_packet, read_timeout_float);

		if (tbuf)
		  {
		    tbuf = readPacketData(current_socket,
					  tbuf,
					  read_timeout_float);
		    if (tbuf)   // read was completely successful
		      {
			global_message = MSG_NO_ERROR;
			current_socket->last_message = MSG_NO_ERROR;
		      }
		    
		    return tbuf;
		  }
		else
		  {
		    return NULL;
		  }
	      }
	    else
	      {
		return NULL;
	      }
	  }
	else
	  {
//	    global_message = current_socket->last_message = MSG_SOCKET_NOT_READY_TO_READ;
	    return NULL;
	  }
	break;
      }
	    
    case ERR_TIMEOUT_READING_PACKET_DATA_TYPE:
      {
	// see if any data is ready to read
	
	if ((select_result = pollOnFDWithTimeout(current_socket->fd, poll_timeout)) == ERR_DURING_SELECT)
	  {
	    if (!quiet)
	      {
		cerr << "SocketBase::_recoverFromReadError: ";
		cerr << "error during select()" << endl;
	      }
	    global_message = ERR_DURING_SELECT;
	    current_socket->last_message = global_message;
	    return NULL;
	  }

	if (select_result == MSG_SOCKET_READY_TO_READ)
	  {
#if SB_DEBUG_VERBOSE
	    cerr << "SocketBase::_recoverFromReadError: socket " << current_socket->fd << " ready to read" << endl;
#endif	    
	    int how_many_bytes;
	    how_many_bytes = bytesAvailableOnFD(current_socket->fd);
#if SB_DEBUG_VERBOSE
	    cerr << "Number of bytes available: " << how_many_bytes << endl;
#endif
	    if (how_many_bytes == 0)
	      {
		if (!quiet)
		  {
		    cerr << "Socket " << current_socket->fd << " appears to have closed" << endl;
		  }
		global_message = ERR_CURRENT_SOCKET_CLOSED;
		return NULL;
	      }
	    
	    // try to get the data type again
	    
		SocketBuf *tbuf;
	    
	    tbuf = recoverReadPacketDataType(current_socket, read_timeout_float);
	    
	    if (tbuf)   // got it successfully
	      {
		// now get the data as above
		
		tbuf = readPacketData(current_socket,
				      tbuf,
				      read_timeout_float);
		
		if (tbuf)  // read finished successfully
		  {
		    global_message = MSG_NO_ERROR;
		    current_socket->last_message = MSG_NO_ERROR;
		  }
		
		return tbuf;
	      }
	    else
	      {
		return NULL;
	      }
	  }
	else
	  {
	    return NULL;
	  }
	break;
      }
	    
    case ERR_TIMEOUT_READING_PACKET:
      {
	// we need to finish reading the packet in
	// see if any data is ready to be read

	if ((select_result = pollOnFDWithTimeout(current_socket->fd, poll_timeout)) == ERR_DURING_SELECT)
	  {
	    if (!quiet)
	      {
		cerr << "SocketBase::_recoverFromReadError: ";
		cerr << "error during select()" << endl;
	      }
	    global_message = ERR_DURING_SELECT;
	    return NULL;
	  }

	if (select_result == MSG_SOCKET_READY_TO_READ)
	  {
#if SB_DEBUG_VERBOSE
	    cerr << "SocketBase::_recoverFromReadError: socket " << current_socket->fd << " ready to read" << endl;
#endif	    
	    int how_many_bytes;
	    how_many_bytes = bytesAvailableOnFD(current_socket->fd);
#if SB_DEBUG_VERBOSE
	    cerr << "Number of bytes available: " << how_many_bytes << endl;
#endif	    
	    if (how_many_bytes == 0)
	      {
		if (!quiet)
		  {
		    cerr << "Socket " << current_socket->fd << " appears to have closed" << endl;
		  }
		global_message = ERR_CURRENT_SOCKET_CLOSED;
		return NULL;
	      }
	    
	    SocketBuf *tbuf;
	    tbuf = recoverReadPacketData(current_socket, read_timeout_float);
	    if (tbuf)
	      {
		global_message = MSG_NO_ERROR;
		current_socket->last_message = MSG_NO_ERROR;
	      }
	    return tbuf;
	  }
	else
	  {
//	    global_message = current_socket->last_message = MSG_SOCKET_NOT_READY_TO_READ;
	    return NULL;
	  }
	break;
      }

    default:
      {
	return NULL;
      }
    }
}

/**************************************************
  SocketBase::_writeWithTimeout       *** PROTECTED

  Base function for implementation of derived
  classes' writeWithTimeout function. This enables
  correct error handling for both client and server.
  Returns MSG_NO_ERROR if the write succeeded.

  Errors stored in last_message:
  ERR_DURING_SELECT
  ERR_TIMEOUT_WRITING_PACKET
  ERR_UDP_PACKET_TOO_LARGE
  ERR_UDP_SEND_FAILED

  Errors stored ONLY in global_message:
  ERR_CURRENT_SOCKET_NOT_OPEN
  ERR_SIGPIPE_WRITING_PACKET
  *************************************************/

int
SocketBase::_writeWithTimeout(SocketStruct *current_socket,
			      SocketBuf *sbuf,
			      float timeout,
			      int use_udp_for_write,
			      int use_multicast,
			      struct sockaddr *address)
{
  int temp_bytes;
  int bytes_left;
  int bytes_sent;
  int bytes_written;
  int expiring;
  struct timeval current_time;
  struct timezone tz = {0, 0};
  float time_difference;
  
#if SB_DEBUG_VERBOSE
  cerr << "_write timeout is " << timeout << endl;
#endif

  if (current_socket)
    {
      if (current_socket->fd <= 0)
	{
	  if (!quiet)
	    {
	      cerr << "SocketBase::_writeWithTimeout: ";
	      cerr << "no current socket" << endl;
	    }
	  global_message = ERR_CURRENT_SOCKET_NOT_OPEN;
	  return ERR_CURRENT_SOCKET_NOT_OPEN;
	}
    }
  else
    {
      if (!quiet)
	{
	  cerr << "SocketBase::_writeWithTimeout: ";
	  cerr << "no current socket" << endl;
	}
      global_message = ERR_CURRENT_SOCKET_NOT_OPEN;
      return ERR_CURRENT_SOCKET_NOT_OPEN;
    }

  if (use_udp_for_write)
    {
      if ((sbuf->num_bytes + sizeof(int)) > SB_MAX_UDP_PACKET_SIZE)
	{
	  if (!quiet)
	    {
	      cerr << "Buffer too large to fit in UDP packet" << endl;
	    }
	  global_message = ERR_UDP_PACKET_TOO_LARGE;
	  current_socket->last_message = global_message;
	  return global_message;
	}
    }

  // if we're little endian, swap bytes in the buffer if necessary

  if (sbuf->type == SOCKET_INT || sbuf->type == SOCKET_FLOAT)
    if (little_endian)
      byteSwapLongBufferToNet(sbuf, sizeof(int));

  sm_byte *sendbuf = new sm_byte[sbuf->num_bytes + sizeof(int) + sizeof(sm_byte)];

  temp_bytes = ntohl(sbuf->num_bytes);

  bcopy((sm_byte *) &(temp_bytes), sendbuf, sizeof(int));
  bcopy(&(sbuf->type), sendbuf + sizeof(int), sizeof(sm_byte));
  bcopy(sbuf->buf, sendbuf+sizeof(int)+sizeof(sm_byte), sbuf->num_bytes);

  bytes_sent = 0;
  bytes_left = sbuf->num_bytes + sizeof(int) + sizeof(sm_byte);

  // put the bytes back in order for our use if necessary

  if (sbuf->type == SOCKET_INT || sbuf->type == SOCKET_FLOAT)
    if (little_endian)
      byteSwapLongBufferFromNet(sbuf, sizeof(int));


#if SB_DEBUG_VERBOSE
  cerr << "Plan to send " << bytes_left << " bytes" << endl;
#endif

  // Clear the global SIGPIPE flag
  global_sigpipe_raised = 0;

  if (!use_udp_for_write)
    {
      expiring = 0;
      
      while (bytes_left > 0 && !global_sigpipe_raised)
	{
	  bytes_written = write(current_socket->fd, sendbuf+bytes_sent, bytes_left);
	  if (bytes_written > 0)
	    {
	      bytes_sent += bytes_written;
	      bytes_left -= bytes_written;
	      expiring = 0;
#if SB_DEBUG_VERBOSE
	      cerr << "bytes_written = " << bytes_written << endl;	  
#endif
	    }
	  else
	    {
	      if (!expiring)
		{
#if SB_DEBUG_VERBOSE
		  cerr << "bytes_written = " << bytes_written << ", expiring..." << endl;
#endif
		  gettimeofday(&current_socket->last_time_accessed, &tz);
		  expiring = 1;
		}
	      else
		{
		  gettimeofday(&current_time, &tz);
		  time_difference = timevalDifference(current_time, current_socket->last_time_accessed);
		  if (time_difference > timeout)
		    {
#if SB_DEBUG_VERBOSE
		      cerr << "write expired!" << endl;
#endif
			delete[] sendbuf;
		      global_message = ERR_TIMEOUT_WRITING_PACKET;
		      current_socket->last_message = global_message;
		      return global_message;
		    }
		}
	    }
	}
    }
  else
    {
      // Using UDP, so use send instead of write
      // If we're using multicast (SocketClient only), use sendto

      if (use_multicast)
	{
	  if (sendto(current_socket->fd, sendbuf, bytes_left, 0, address, sizeof(sockaddr)) <= 0)
	    {
	      if (!quiet)
		{
		  cerr << "Multicast send failed" << endl;
		}
	      delete[] sendbuf;
	      global_message = ERR_MULTICAST_SEND_FAILED;
	      current_socket->last_message = global_message;
	      return global_message;
	    }
	}	    
      else
	{
	  if (send(current_socket->fd, sendbuf, bytes_left, 0) <= 0)
	    {
	      if (!quiet)
		{
		  cerr << "UDP send failed" << endl;
		}
	      delete[] sendbuf;
	      global_message = ERR_UDP_SEND_FAILED;
	      current_socket->last_message = global_message;
	      return global_message;
	    }
	}
    }

  if (global_sigpipe_raised)
    {
      if (!quiet)
	cerr << "Write was interrupted by SIGPIPE" << endl;
      delete[] sendbuf;
      return ERR_SIGPIPE_WRITING_PACKET;
    }

#if SB_DEBUG_VERBOSE
  cerr << "Wrote " << bytes_sent << " bytes total" << endl;
#endif
  delete[] sendbuf;
  return MSG_NO_ERROR;
}
  
/**************************************************
  SocketBase::getGlobalMessage

  Returns the last global message set by an
  executed function.
  *************************************************/

int
SocketBase::getGlobalMessage()
{
  return global_message;
}

/**************************************************
  SocketBase::getQuiet

  Returns 1 if the SocketBase is in quiet mode
  (no printed output), 0 otherwise.
  *************************************************/

int
SocketBase::getQuiet()
{
  return (quiet ? 1 : 0);
}

/**************************************************
  SocketBase::setQuiet

  Turns on quiet mode if the argument is nonzero.
  *************************************************/

void
SocketBase::setQuiet(int on_or_off)
{
  quiet = on_or_off;
}

/**************************************************
  SocketBase::getUsingSysVSignals

  Returns 1 if the SocketBase is using System V
  signals for its internal SIGPIPE handler, 0
  otherwise.
  *************************************************/

int
SocketBase::getUsingSysVSignals()
{
  return (global_sm_sys_v_signals ? 1 : 0);
}

/**************************************************
  SocketBase::setUsingSysVSignals

  If the argument is nonzero, will reinstall the
  internal SIGPIPE signal handler after it is
  called (System V signal behavior).
  *************************************************/

void
SocketBase::setUsingSysVSignals(int on_or_off)
{
  global_sm_sys_v_signals = on_or_off;
}
