#ifdef __GNUG__  // most platforms which use GNU C don't
extern "C" {     // have this in their system header files
#endif
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <strings.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netdb.h>
#ifdef __GNUG__
}
#endif
#ifdef __sgi
#include <bstring.h>
#endif
#include <iostream.h>
#include "SocketMan.h"

#define SM_DEBUG_VERBOSE 0

/**************************************************
  SocketMan::SocketMan

  Instantiates the Socket Manager. Fixes the port
  on which the returned object listens for
  connections and opens the listening socket.
  
  Optional arguments:
  user_read_timeout_secs	Number of seconds before reads time out.
  				Default value is 1 second.
  user_read_timeout_usecs	Number of microseconds before reads time out.
				Total timeout is (secs + (usecs / 1,000,000)).
				Default value is 0 microseconds.
  user_write_timeout_secs	Number of seconds before writes time out.
  				Default value is 1 second.
  user_write_timeout_usecs	Number of microseconds before writes time out.
				Total timeout is (secs + (usecs / 1,000,000)).
				Default value is 0 microseconds.
  user_quiet			If nonzero, suppresses all printed output from
  				the object.
				Default value is 0.
  user_limit_connections	If greater than zero, limits the number of
				connections which the Socket Manager can have
				open at any one time.
				Default value is 0.
  user_use_sigio		If nonzero, sets the appropriate flags for
				each data socket so that they will deliver
				SIGIO to the calling process. The user is
				responsible for setting up the signal handler
				to catch SIGIO.
				Default value is 0.
  user_sys_v_signals		A flag used internally to the SocketMan structure
				telling the system whether to reinstall the
				SIGPIPE signal handler after it is called.
				Should be 1 if your system supports only Sys V
				signals (i.e. HP/UX). Default value is 0.
				(For SGIs, should be 0.)
  user_use_udp			If nonzero, SocketMan uses UDP sockets for all
				connections. NOTE that you can not have a UDP
				client and server on the same machine running
				on the same port.
				Default value is 0.
  *************************************************/

SocketMan::SocketMan(unsigned short port, 
		     int user_read_timeout_secs,
		     int user_read_timeout_usecs,
		     int user_write_timeout_secs,
		     int user_write_timeout_usecs,
		     int user_quiet,
		     int user_limit_connections,
		     int user_use_sigio,
		     int user_sys_v_signals,
		     int user_use_udp)
{
  sl = new Dlist< SocketStruct >;
  sli = new Dlist_iter< SocketStruct >(*sl);

  pl = new Dlist< unsigned short >;
  pli = new Dlist_iter< unsigned short >(*pl);

  current_socket = 0;
  last_accepted_socket = 0;

  quiet = user_quiet;
  use_udp = user_use_udp;
  udp_last_port = port;
  udp_base_port = port;

  use_sigio = user_use_sigio;
  global_sm_sys_v_signals = user_sys_v_signals;

  while (createListeningSocket(port) != 0)
    {
      if (!quiet)
	cerr << "Error instantiating Socket Manager" << endl;
      sleep(2);
    }

  read_timeout.tv_sec = user_read_timeout_secs;
  read_timeout.tv_usec = user_read_timeout_usecs;
  write_timeout.tv_sec = user_read_timeout_secs;
  write_timeout.tv_usec = user_read_timeout_usecs;
  read_timeout_float = (float) user_read_timeout_secs + (float) user_read_timeout_usecs * (float) (1.0e-6);
  write_timeout_float = (float) user_write_timeout_secs + (float) user_write_timeout_usecs * (float) (1.0e-6);
  global_message = 0;

  if (user_limit_connections)
    {
      if (user_limit_connections > 0)
	limit_connections = user_limit_connections;
      else
	{
	  if (!quiet)
	    {
	      cerr << "SocketMan: ERROR: ";
	      cerr << "You can't limit the number of connections to a negative number!" << endl;
	    }
	}
    }
  else
    limit_connections = 0;
  signal(SIGPIPE, handleSIGPIPE);
}

/**************************************************
  SocketMan::~SocketMan

  Destructor. Closes listening socket and all
  connected sockets.
  *************************************************/

SocketMan::~SocketMan()
{
  closeListeningSocket();
  closeConnectedSockets();

#if SM_DEBUG_VERBOSE  
  cerr << "Destructor called" << endl;
#endif
}

/**************************************************
  SocketMan::closeListeningSocket       *** PRIVATE

  Closes listening socket. This method was removed
  from the destructor and separated from 
  closeConnectedSockets, below, to allow switching
  UDP connections on and off.
  *************************************************/

int
SocketMan::closeListeningSocket()
{
  if (shutdown(listening_socket_fd, 2) == -1)
    {
      if (!quiet)
	cerr << "SocketMan::closeListeningSocket: error shutting down listening socket" << endl;
      global_message = ERR_SHUTTING_DOWN_LISTENING_SOCKET;
      return global_message;
    }

  if (close(listening_socket_fd) == -1)
    {
      if (!quiet)
	cerr << "SocketMan::closeListeningSocket: error closing listening socket" << endl;
      global_message = ERR_CLOSING_LISTENING_SOCKET;
      return global_message;
    }
  
  global_message = MSG_NO_ERROR;
  return global_message;
}

/**************************************************
  SocketMan::closeConnectedSockets      *** PRIVATE

  Closes all connected sockets. Called in 
  destructor and if necessary when switching UDP
  use on or off.
  *************************************************/

int
SocketMan::closeConnectedSockets()
{
  rewindSockets();
  if (!quiet)
    cerr << "Closing connected sockets:" << endl;
  while (nextSocket())
    {
      if (shutdown(current_socket->fd, 2) == -1)
	{
	  if (!quiet)
	    cerr << "SocketMan::closeConnectedSockets: error shutting down socket " << current_socket->fd << endl;
	  global_message = ERR_SHUTTING_DOWN_CONNECTED_SOCKET;
	  return global_message;
	}
      
      if (close(current_socket->fd) == -1)
	{
	  if (!quiet)
	    cerr << "SocketMan::closeConnectedSockets: error closing socket " << current_socket->fd << endl;
	  global_message = ERR_CLOSING_CONNECTED_SOCKET;
	  return global_message;
	}
      if (!deleteSocket())
	{
	  if (!quiet)
	    cerr << "SocketMan::closeConnectedSockets: error deleting socket" << current_socket->fd << endl;
	  global_message = ERR_DELETING_CONNECTED_SOCKET;
	  return global_message;
	}
    }

  global_message = MSG_NO_ERROR;
  return global_message;
}


/**************************************************
  SocketMan::addToSocketList            *** PRIVATE

  Adds a socket to the internal list structure. If
  UDP socket, also stores the port to which the
  socket is bound so it can be reused if the socket
  closes. Modifies last_accepted_socket to be the
  socket added to the list.
  *************************************************/

void
SocketMan::addToSocketList(int sock_fd, 
			   unsigned short sock_port,
			   unsigned long address)
{
  SocketStruct *temp = new SocketStruct;

  temp->fd = sock_fd;
  temp->port = sock_port;
  temp->address = address;
  temp->last_time_accessed.tv_sec = 0;
  temp->last_time_accessed.tv_usec = 0;

  sl->append(*temp);
  last_accepted_socket = temp;
}

/**************************************************
  SocketMan::createListeningSocket      *** PRIVATE

  Makes a socket to which other machines can connect;
  accepting a connection on this socket hands off another,
  data, socket through which communication can occur.
  Called in the constructor. Returns 0 if successful.

  ERRORS:
  ERR_UNKNOWN_HOST
  ERR_CREATING_SOCKET
  ERR_BINDING_SOCKET
  ERR_LISTENING_ON_SOCKET
  *************************************************/

int
SocketMan::createListeningSocket(unsigned short port)
{
  char                myname[SB_MAXHOSTNAME];
  int                 ret, server_len;
  struct hostent      *my_info;
  struct sockaddr_in  server_info;
  struct linger       ling;

  server_len = sizeof(server_info);

  gethostname(myname, SB_MAXHOSTNAME);

#if SM_DEBUG_VERBOSE
  cerr << "Getting hostname" << endl;
#endif
  my_info = (struct hostent *)gethostbyname(myname);
  if (my_info == NULL)
    {  
      if (!quiet)
	cerr << "Attempt to connect to unknown host \"" << myname << "\"" << endl;
      global_message = ERR_UNKNOWN_HOST;
      return global_message;
    }
#if SM_DEBUG_VERBOSE
  cerr << "Got hostname" << endl;
#endif

  bzero((char *)&server_info, server_len);
  bcopy(my_info->h_addr, 
	(char *)&(server_info.sin_addr), 
	my_info->h_length);
  server_info.sin_family = my_info->h_addrtype;
  server_info.sin_port = htons(port);

  if ((listening_socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {  
      listening_socket_fd = 0;
      if (!quiet)
	cerr << "Error creating listening socket" << endl;
      global_message = ERR_CREATING_SOCKET;
      return global_message;
    }

#if SM_DEBUG_VERBOSE
  cerr << "Created listening socket" << endl;
#endif

  ret = bind(listening_socket_fd, (struct sockaddr *) &server_info, server_len);
  if (ret == -1)
    {  
      close(listening_socket_fd);
      listening_socket_fd = 0;
      if (!quiet)
	cerr << "Error binding socket to address" << endl;
      global_message = ERR_BINDING_SOCKET;
      return global_message;
    }
  else
    {
#if SM_DEBUG_VERBOSE
      cerr << "Bound socket to port " << ntohs(server_info.sin_port) << endl;
      cerr << "On machine " << longToDotFormat(server_info.sin_addr.s_addr) << endl;
#endif
    }

#if SM_DEBUG_VERBOSE
  cerr << "Bound listening socket" << endl;
#endif

  ret = listen(listening_socket_fd, 5);
  if (ret == -1)
    {  
      close(listening_socket_fd);
      listening_socket_fd = 0;
      if (!quiet)
	cerr << "Listen failed" << endl;
      global_message = ERR_LISTENING_ON_SOCKET;
      return global_message;
    }
  
#if SM_DEBUG_VERBOSE
  cerr << "Listen succeeded" << endl;
#endif

  ling.l_onoff = 0;
  ling.l_linger = 0;

  setsockopt(listening_socket_fd, SOL_SOCKET, SO_LINGER, &ling, sizeof(struct linger));
  
#if SM_DEBUG_VERBOSE
  cerr << "Set no linger" << endl;
#endif

  global_message = MSG_NO_ERROR;
  return global_message;
}  

/**************************************************
  SocketMan::acceptConnectionWithPoll

  Polls on the listening socket and, if there are
  waiting connections, accepts the first one off
  the queue. 
  Returns MSG_NO_ERROR if successful.
  If there were no connections pending, returns
  MSG_LISTENING_SOCKET_NO_PENDING_CONNECTIONS.
  If the maximum number of connections has been
  reached (see the constructor), returns
  MSG_MAXIMUM_NUMBER_OF_CONNECTIONS.

  Errors:
  ERR_LISTENING_SOCKET_NOT_OPEN
  ERR_LISTENING_SOCKET_FAILED_ACCEPT
  ERR_UNKNOWN_HOST
  ERR_UDP_CREATING_SOCKET
  ERR_UDP_BINDING_SOCKET
  ERR_UDP_CONNECTING_SOCKET
  *************************************************/

int
SocketMan::acceptConnectionWithPoll()
{
  int temp_fd;
  unsigned short temp_port = 0;
  struct sockaddr_in peer_addr;
  int peer_addrlen;
  struct timeval timeout = {0, 0};

  if (listening_socket_fd <= 0)
    {
      if (!quiet)
	cerr << "Can't accept connections; listening socket not open" << endl;
      global_message = ERR_LISTENING_SOCKET_NOT_OPEN;
      return global_message;
    }

  if (!(limit_connections && (number_of_connections == limit_connections)))
    {
      if (pollOnFDWithTimeout(listening_socket_fd, timeout) == MSG_SOCKET_READY_TO_READ)
	{
	  peer_addrlen = sizeof(peer_addr);
	  temp_fd = accept(listening_socket_fd, (sockaddr *) &peer_addr, &peer_addrlen);
	  if (temp_fd == -1)
	    {
	      if (!quiet)
		cerr << "SocketMan::aCWP: failed accept" << endl;
	      global_message = ERR_LISTENING_SOCKET_FAILED_ACCEPT;
	      return global_message;
	    }
	  else
	    { 
	      if (!quiet)
		cerr << "SocketMan::aCWP: accepted successfully" << endl;
	      // enable non blocking IO
	      int flags;
	      flags = fcntl(temp_fd, F_GETFL, 0);
	      
	      if (use_udp)
		{
		  unsigned short *udp_port;
		  int udp_sock;
		  
		  // set non blocking reads on the temporary TCP socket

		  flags |= O_NDELAY;
		  fcntl(temp_fd, F_SETFL, flags);

		  // find out which local port we're going to use
		  pli->rewind();

		  if ((udp_port = pli->next()) == 0)
		    {
		      udp_port = new unsigned short;
		      *udp_port = udp_last_port++;
		    }
		  else
		    pli->deleteCurrentElement();

		  // make a udp socket
		  udp_sock = socket(AF_INET, SOCK_DGRAM, 0);
		  if (udp_sock == -1)
		    {
		      if (!quiet)
			cerr << "SocketMan::aCWP: error creating UDP socket" << endl;
		      global_message = ERR_UDP_CREATING_SOCKET;
		      return global_message;
		    }

		  // get our name so we can bind the
		  // socket to our local port

		  char                myname[SB_MAXHOSTNAME];
		  int                 ret, server_len;
		  struct hostent      *my_info;
		  struct sockaddr_in  server_info;

		  server_len = sizeof(server_info);

		  gethostname(myname, SB_MAXHOSTNAME);

		  my_info = (struct hostent *)gethostbyname(myname);
		  if (my_info == NULL)
		    {  
		      if (!quiet)
			cerr << "Attempt to connect to unknown host \"" << myname << "\"" << endl;
		      global_message = ERR_UNKNOWN_HOST;
		      return global_message;
		    }
		  
		  bzero((char *)&server_info, server_len);
		  bcopy(my_info->h_addr, 
			(char *)&(server_info.sin_addr), 
			my_info->h_length);
		  server_info.sin_family = my_info->h_addrtype;
		  server_info.sin_port = htons(*udp_port);
		  
		  // bind the UDP socket

		  ret = bind(udp_sock, (struct sockaddr *) &server_info, server_len);
		  if (ret == -1)
		    {  
		      close(udp_sock);
		      if (!quiet)
			cerr << "Error binding UDP socket to port" << endl;
		      global_message = ERR_UDP_BINDING_SOCKET;
		      return global_message;
		    }

		  // tell the client we're using UDP on this port

		  SocketStruct tmp;
		  tmp.fd = temp_fd;

		  SocketBuf *sbuf = new SocketBuf;
		  sbuf->num_bytes = sizeof(unsigned short);
		  sbuf->buf = new sm_byte[sizeof(unsigned short)];
		  *((unsigned short *) sbuf->buf) = htons(*udp_port);
		  _writeWithTimeout(&tmp, sbuf, 2.0, 0);
		  delete sbuf;

		  // connect the UDP socket to the client, as he
		  // has already created his and connected to us

		  peer_addr.sin_port = htons(udp_base_port);  // reuse his address information

		  int connect_val;
		  if ((connect_val = connect(udp_sock, (const sockaddr *) &peer_addr, peer_addrlen)) != 0)
		    {
		      if (!quiet)
			{
			  cerr << "SocketMan::aCWP: UDP connect failed" << endl;
			  perror("SocketMan::aCWP");
			}
		      global_message = ERR_UDP_CONNECTING_SOCKET;
		      return global_message;
		    }
#if SM_DEBUG_VERBOSE
		  else
		    {
		      cerr << "Succeeded connecting UDP socket " << udp_sock << " to port " << ntohs(peer_addr.sin_port) << endl;
		      cerr << "on machine " << longToDotFormat(ntohl(peer_addr.sin_addr.s_addr)) << endl;
		    }
#endif

/*		  // now we should have a UDP socket connected to both machines.
		  // test it by writing a short burst of data to it

		  tmp.fd = udp_sock;
		  *((int *) sbuf.buf) = htonl(sixsixsix);
		  _writeWithTimeout(&tmp, &sbuf, 2.0, 1);
		  
		  // now wait and see if the client comes back
		  
		  SocketBuf *sbufp;
		  struct timeval poll_timeout = {2, 0};

		  sbufp = _readWithPollAndTimeout(&tmp, poll_timeout, 2.0, 1);

		  if (!sbufp)
		    {
		      if (!quiet)
			{
			  cerr << "UDP client failed to respond" << endl;
			}
		      close(udp_sock);
		      close(temp_fd);
		      global_message = ERR_UDP_CLIENT_NO_RESPONSE;
		      return global_message;
		    }
*/
		  // OK. We've got the UDP socket set up.
		  // Clean up after ourselves

		  close(temp_fd);   // close down the TCP socket
		  temp_fd = udp_sock;
		  temp_port = *udp_port;
		  flags = fcntl(temp_fd, F_GETFL, 0);
		}
	      else
		{
		  // The client is expecting a response on whether
		  // we're using UDP. Tell him we're not.
		  
		  SocketBuf *sbuf = new SocketBuf;
		  SocketStruct tmp;
		  
		  sbuf->num_bytes = sizeof(unsigned short);
		  sbuf->buf = new sm_byte[sizeof(unsigned short)];
		  *((unsigned short *) sbuf->buf) = 0;
		  tmp.fd = temp_fd;
		  
		  _writeWithTimeout(&tmp, sbuf, 2.0, 0);
		  delete sbuf;
		}

	      // enable SIGIO if desired
	      if (use_sigio)
		{
#ifndef __hpux
		  fcntl(temp_fd, F_SETOWN, getpid());
		  flags |= FASYNC;
#else
		  int processID;
		  int sig_status = 1;

		  processID = getpid();
		  if (ioctl(temp_fd, SIOCSPGRP, &processID) < 0)
		    perror("SocketMan::aCWP");
		  if (ioctl(temp_fd, FIOASYNC, &sig_status) < 0)
		    perror("SocketMan::aCWP");
#endif
		}

	      flags |= O_NDELAY;

	      fcntl(temp_fd, F_SETFL, flags);
	      addToSocketList(temp_fd, temp_port, peer_addr.sin_addr.s_addr);
	      number_of_connections++;
	      global_message = MSG_NO_ERROR;
	      return global_message;
	    }
	}
      else
	{
	  global_message = MSG_LISTENING_SOCKET_NO_PENDING_CONNECTIONS;
	  return global_message;
	}
    }
  else
    {
      if (!quiet)
	{
	  cerr << "SocketMan::acceptConnectionWithPoll:" << endl;
	  cerr << "Reached maximum number of connections" << endl;
	}
      global_message = MSG_MAXIMUM_NUMBER_OF_CONNECTIONS;
      return global_message;
    }
}

/**************************************************
  SocketMan::rewindSockets

  "Rewinds" the internal "list" of sockets to
  allow iteration over all connected sockets.
  *************************************************/

void
SocketMan::rewindSockets()
{
  sli->rewind();
}

/**************************************************
  SocketMan::ffwdSockets                *** PRIVATE

  Goes to the end of the internal socket list. This
  function is unused. Was originally intended to
  allow iteration to go in both directions along
  the list.
  *************************************************/

void
SocketMan::ffwdSockets()
{
  sli->ffwd();
}

/**************************************************
  SocketMan::nextSocket

  Changes the (internal) current socket to the next
  one on the (internal) list of connected sockets,
  if available. Returns 1 if succeeded; returns 0
  if end of the list.
  *************************************************/

int
SocketMan::nextSocket()
{
  SocketStruct *temp;

  temp = sli->next();
  if (temp)
    {
      current_socket = temp;
      return 1;
    }
  else
    return 0;
}

/**************************************************
  SocketMan::prevSocket                 *** PRIVATE

  Changes the current socket to the previous one
  on the internal list of sockets. Only used during
  deletions of closed sockets (by deleteSocket,
  called from readWith***) to keep iterations along
  the socket list in order. Returns 1 if successful,
  0 otherwise.
  *************************************************/

int
SocketMan::prevSocket()
{
  SocketStruct *temp;

  temp = sli->prev();
  if (temp)
    {
      current_socket = temp;
      return 1;
    }
  else
    return 0;
}

/**************************************************
  SocketMan::deleteSocket               *** PRIVATE

  Deletes the current socket from the internal list
  of sockets. Called when readWith*** decides the
  socket has closed, or when SIGPIPE is raised.
  Returns 1 on success, 0 otherwise.
  *************************************************/

int
SocketMan::deleteSocket()
{
  int temp;

  if (use_udp)
    {
      if (current_socket->port > 0)
	{
	  pl->append(current_socket->port);
	}
    }

  if (!quiet)
    {
      cerr << "Shutdown returned " << shutdown(current_socket->fd, 2) << endl;
      cerr << "Close returned " << close(current_socket->fd) << endl;
    }
  else
    {
      shutdown(current_socket->fd, 2);
      close(current_socket->fd);
    }

  temp = sli->deleteCurrentElement();
  number_of_connections--;

  if (!sli->prev())
    sli->rewind();    // Keeps iterations in order
  return temp;
}

/**************************************************
  SocketMan::setFDSet

  Sets a user's passed in fd_set with all connected
  sockets and the listening socket of the Socket
  Manager. Used to support an external select()
  over several SocketMans or SocketClients; also
  used in smartSelect, below.
  *************************************************/

void
SocketMan::setFDSet(fd_set *fds)
{
  // add the listening socket to the file descriptor set
  FD_SET(listening_socket_fd, fds);

  // now add all the data sockets' descriptors to the fd_set

  rewindSockets();

  while (nextSocket())
    {
      FD_SET(current_socket->fd, fds);
    }
}  

/**************************************************
  SocketMan::getPostSelectReadAction

  Calling function passes in an fd_set which is
  assumed to contain the result of a select()
  prepared by setFDSet() (and possibly other
  functions as well), above. If listening socket is
  ready to read, returns MSG_SS_ACCEPT_CONNECTION
  (an indication to call 
  acceptConnectionWithPoll()). If another socket is
  ready to read, sets the current socket 
  appropriately and returns 
  MSG_SS_READ_FROM_CURRENT_SOCKET. If no socket is
  ready to read, returns 
  MSG_NO_SOCKET_READY_TO_READ.
  *************************************************/

int
SocketMan::getPostSelectReadAction(fd_set *readfds)
{
  if (FD_ISSET(listening_socket_fd, readfds))
    {
      global_message = MSG_SS_ACCEPT_CONNECTION;
      return global_message;
    }
  else
    {
      // step through the entire socket list.
      // leave the socket that's ready to read
      // as the current socket.
      
      rewindSockets();
      
      while (nextSocket())
	{
	  if (FD_ISSET(current_socket->fd, readfds))
	    {
	      global_message = MSG_SS_READ_FROM_CURRENT_SOCKET;
	      return global_message;
	    }
	}
      
      global_message = MSG_NO_SOCKET_READY_TO_READ;
      return global_message;
    }
}

/**************************************************
  SocketMan::smartSelect

  Blocks on a select. Returns 
  MSG_SS_ACCEPT_CONNECTION if the user should call
  acceptConnectionWithPoll() after it returns.
  Returns MSG_SS_READ_FROM_CURRENT_SOCKET if the
  user should call readWith*** after it returns.
  Uses above functions setFDSet() and 
  getPostSelectReadAction(). 
  
  Errors:
  ERR_DURING_SELECT
  ERR_NO_SOCKET_READY_TO_READ (should not happen as
                     a result of using this method)
  *************************************************/

int
SocketMan::smartSelect()
{
  fd_set readfds;
  int width;
  int result;
  
  // find the maximum number of open file descriptors
#ifndef __hpux
  width = getdtablesize();
#else
  width = FD_SETSIZE;
#endif

  // get file descriptor set containing listening socket
  // and all connected clients' sockets

  FD_ZERO(&readfds);
  setFDSet(&readfds);

  // go into a blocking select on all these sockets

#ifndef __hpux
  result = select(width, &readfds, NULL, NULL, 0);
#else
  result = select(width, (int *) &readfds, NULL, NULL, 0);
#endif  

  if (result < 0)
    {
      global_message = ERR_DURING_SELECT;
      return global_message;
    }
  else
    {
      return getPostSelectReadAction(&readfds);
    }
}

/**************************************************
  SocketMan::readWithPollAndTimeout

  Polls the current socket and, if ready to read,
  attempts to read a packet from it.

  If successful, returns a pointer to a SocketBuf
  and sets global_message and last_message to
  MSG_NO_ERROR. If there was nothing to read,
  returns NULL and sets global_message and
  last_message to MSG_SOCKET_NOT_READY_TO_READ.
  If the read timed out, returns NULL and returns
  the appropriate error shown below.
  recoverFromReadError may then be called to try
  to finish the read.
  
  Errors stored in last_message:
  ERR_DURING_SELECT
  ERR_UDP_RECEIVE_FAILED
  ERR_TIMEOUT_READING_PACKET_LENGTH
  ERR_TIMEOUT_READING_PACKET_DATA

  Errors stored ONLY in global_message:
  ERR_CURRENT_SOCKET_NOT_OPEN
  ERR_CURRENT_SOCKET_CLOSED
  ERR_SIGPIPE_READING_PACKET_LENGTH
  ERR_SIGPIPE_READING_PACKET
  *************************************************/

SocketBuf*
SocketMan::readWithPollAndTimeout()
{
  struct timeval timeout = {0, 0};
  SocketBuf* retval;
      
  global_sm = (void *) this;
  global_sm_or_sc = SM_GLOBAL_SWITCH;
  global_sigpipe_raised = 0;

  retval = _readWithPollAndTimeout(current_socket,
				   timeout,
				   read_timeout_float,
				   use_udp);

  if (retval == NULL && global_message == ERR_CURRENT_SOCKET_CLOSED)
    {
      if (!quiet)
	{
	  cerr << "Deleting it from the list" << endl;
	}
      deleteSocket();
    }

  global_sm = NULL;

  return retval;
}

/**************************************************
  SocketMan::readWithTimeout

  Attempts to read data (without polling) from the
  current socket for the number of seconds 
  specified by the user in the read-timeout
  options in the constructor.

  If successful, returns a pointer to a SocketBuf
  and sets global_message and last_message to
  MSG_NO_ERROR. If there was nothing to read,
  returns NULL and sets global_message and
  last_message to MSG_SOCKET_NOT_READY_TO_READ.
  If the read timed out, returns NULL and returns
  the appropriate error shown below.
  recoverFromReadError may then be called to try
  to finish the read.
  
  Errors stored in last_message:
  ERR_DURING_SELECT
  ERR_UDP_RECEIVE_FAILED
  ERR_TIMEOUT_READING_PACKET_LENGTH
  ERR_TIMEOUT_READING_PACKET_DATA

  Errors stored ONLY in global_message:
  ERR_CURRENT_SOCKET_NOT_OPEN
  ERR_CURRENT_SOCKET_CLOSED
  ERR_SIGPIPE_READING_PACKET_LENGTH
  ERR_SIGPIPE_READING_PACKET
  *************************************************/

SocketBuf*
SocketMan::readWithTimeout()
{
  SocketBuf* retval;

      
  global_sm = (void *) this;
  global_sm_or_sc = SM_GLOBAL_SWITCH;
  global_sigpipe_raised = 0;

  retval = _readWithPollAndTimeout(current_socket,
				   read_timeout,
				   read_timeout_float,
				   use_udp);

  if (retval == NULL && global_message == ERR_CURRENT_SOCKET_CLOSED)
    {
      if (!quiet)
	{
	  cerr << "Deleting it from the list" << endl;
	}
      deleteSocket();
    }

  global_sm = NULL;

  return retval;
}

/**************************************************
  SocketMan::recoverFromReadError

  If the previous read on the current socket timed
  out, attempts to finish it by polling to see if
  new data has come in and reading it if it has.

  If successful, returns a pointer to a SocketBuf
  and sets global_message and last_message to
  MSG_NO_ERROR. If there was nothing to read,
  returns NULL and sets global_message and
  last_message to MSG_SOCKET_NOT_READY_TO_READ.
  If the read timed out, returns NULL and returns
  the appropriate error shown below.
  recoverFromReadError may then be called to try
  to finish the read.
  
  NOTE: Only works for TCP sockets.

  Errors stored in last_message:
  ERR_DURING_SELECT
  ERR_TIMEOUT_READING_PACKET_LENGTH
  ERR_TIMEOUT_READING_PACKET_DATA

  Errors stored ONLY in global_message:
  ERR_CURRENT_SOCKET_NOT_OPEN
  ERR_CURRENT_SOCKET_CLOSED
  ERR_SIGPIPE_READING_PACKET_LENGTH
  ERR_SIGPIPE_READING_PACKET
  *************************************************/

SocketBuf*
SocketMan::recoverFromReadError()
{
  SocketBuf *retval;

  global_sm = (void *) this;
  global_sm_or_sc = SM_GLOBAL_SWITCH;
  global_sigpipe_raised = 0;

  retval = _recoverFromReadError(current_socket,
				 read_timeout_float);

  if (retval == NULL && global_message == ERR_CURRENT_SOCKET_CLOSED)
    {
      if (!quiet)
	{
	  cerr << "Deleting it from the list" << endl;
	}
      deleteSocket();
    }

  global_sm = NULL;

  return retval;
}

/**************************************************
  SocketMan::writeWithTimeout

  Attempts to write the data stored in the 
  SocketBuf pointed to by sbuf onto the current
  socket. Times out if there are contiguous errors
  for the period specified in the write-timeout
  sections of the constructor.

  Returns MSG_NO_ERROR if successful.
  
  Errors:
  ERR_CURRENT_SOCKET_NOT_OPEN
  ERR_UDP_PACKET_TOO_LARGE
  ERR_UDP_SEND_FAILED
  ERR_TIMEOUT_WRITING_PACKET
  ERR_SIGPIPE_WRITING_PACKET
  *************************************************/

int
SocketMan::writeWithTimeout(SocketBuf *sbuf)
{
  int retval;

  global_sm = (void *) this;
  global_sm_or_sc = SM_GLOBAL_SWITCH;
  global_sigpipe_raised = 0;

  retval = _writeWithTimeout(current_socket, sbuf, write_timeout_float, use_udp);

  global_sm = NULL;

  return retval;
}

/**************************************************
  SocketMan::getLastMessage

  Returns the last message set *on the current
  socket*. This is different from the SocketClient
  version, in which there is only one socket.
  *************************************************/

int
SocketMan::getLastMessage()
{
  if (current_socket)
    return current_socket->last_message;
  else
    {
      if (!quiet)
	{
	  cerr << "SocketMan::getLastMessage: no current socket" << endl;
	}
      global_message = ERR_CURRENT_SOCKET_NOT_OPEN;
      return global_message;
    }
}

/**************************************************
  SocketMan::getNumberOfConnections

  Returns the number of clients to which the
  Socket Manager is connected.
  *************************************************/

int
SocketMan::getNumberOfConnections()
{
  return number_of_connections;
}

/**************************************************
  SocketMan::cleanupUDPSockets

  Attempts to find all UDP sockets for which the
  client has died by calling select() on each and
  looking for an error. Does not currently work. 
  *************************************************/

void
SocketMan::cleanupUDPSockets()
{
  struct timeval timeout = {0, 0};

  rewindSockets();

  while (nextSocket())
    {
      if (pollOnFDWithTimeout(current_socket->fd, timeout) == ERR_DURING_SELECT)
	{
	  if (!quiet)
	    {
	      cerr << "SocketMan::cleanupUDPSockets: Socket ";
	      cerr << current_socket->fd << " appears to have closed" << endl;
	      cerr << "Deleting it from the list" << endl;
	    }
	  deleteSocket();
	}
    }
}

/**************************************************
  SocketMan::getUsingUDP

  Returns 1 if the Socket Manager is using UDP for
  connected sockets, 0 otherwise.
  *************************************************/

int
SocketMan::getUsingUDP()
{
  return (use_udp ? 1 : 0);
}

/**************************************************
  SocketMan::setUsingUDP

  Switches between using TCP and UDP modes for
  connected sockets. Returns MSG_NO_ERROR if the
  switch was successful (or a no-op).

  Errors:
  ERR_SHUTTING_DOWN_CONNECTED_SOCKET
  ERR_CLOSING_CONNECTED_SOCKET
  ERR_DELETING_CONNECTED_SOCKET
  *************************************************/

int
SocketMan::setUsingUDP(int on_or_off)
{
  int ret;

  if (on_or_off)
    {
      if (!use_udp)
	{
	  // first need to close all connected sockets
	  if (number_of_connections > 0)
	    if ((ret = closeConnectedSockets()) != MSG_NO_ERROR)
	      return ret;
	  
	  use_udp = 1;
	}
    }
  else
    {
      if (use_udp)
	{
	  // need to close all connected sockets
	  if ((ret = closeConnectedSockets()) != MSG_NO_ERROR)
	    return ret;

	  use_udp = 0;
	}
    }
  return MSG_NO_ERROR;
}
  
