#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include "common.h"
#include "slib.h"
#include <math.h>
#include <stropts.h>

#include <gl.h>
#include <device.h>
#include <vl/vl.h>
#include <vl/dev_sirius.h>
#include <math.h>

#include "contour.h"
#include "packet.h"

#include "find_features.proto"
#include "track_features.proto"

#define BS_DEBUG 0
#define BS_DEBUG_THAD 0
#define SANITY_DEBUG 0
#define SLEEP 0

#define VALID_EXPIRATION_TIME 2   /* in frames --
				   how soon does the last valid position expire? */
#define VALID_TRACKING_TIME 5   /* in frames --
				 how long before tracking is turned on? */

#define LEFT_FOOT_EXTREMITY_ID 1
#define LEFT_HAND_EXTREMITY_ID 2
#define HEAD_EXTREMITY_ID 3
#define RIGHT_HAND_EXTREMITY_ID 4
#define RIGHT_FOOT_EXTREMITY_ID 5
#define FloatAbs(number) ((number) >= 0.)?(number):(-1. * (number))

#define HEAD_THRESHOLD 5.  /* In failure cases, don't allow the head to
			      stray more than 5 inches away from the center */
#define HAND_THRESHOLD 6.  /* Also, don't let the hand get closer than 6
			      inches to the center */


int
  InTopOfBoundingBox(float tlx, float tly, 
		     float brx, float bry,
		     float yValue)
{
  if ((yValue <= tly) && 
      (yValue >= (bry + (1. / 3.) * (tly - bry))))
    return TRUE;
  else
    return FALSE;
}

int
  InBottomOfBoundingBox(float tlx, float tly, 
			float brx, float bry,
			float yValue)
{
  if ((yValue >= bry) && 
      (yValue <= (bry + (.4 * (tly - bry)))))
    return TRUE;
  else
    return FALSE;
}

int
  Quadrant(float centroidx,
	   float centroidy,
	   float testPointx,
	   float testPointy)
{
  if ((testPointx <= centroidx) && (testPointy < centroidy))
    return 1;
  else
    if ((testPointx < centroidx) && (testPointy >= centroidy))
      return 2;
    else
      if ((testPointx >= centroidx) && (testPointy > centroidy))
	return 3;
      else
	return 4;
}

float
  DistanceSquared(float ax,
		  float ay,
		  float bx,
		  float by)
{
  return (float) ((bx - ax) * (bx - ax) + (by - ay) * (by - ay));
}

int
  ThresholdScan(float *oldPos,
		float extremities[NUMBER_OF_EXTREMA * 3],
		int startCounter,
		int endCounter,
		float *newPos,
		int *newIndex,
		float Threshold)
{
  float tempPos[2];
  int flag = 0;
  float calcedDistance = 0.;

  tempPos[0] = newPos[0];
  tempPos[1] = newPos[1];

  while (startCounter < endCounter)
    {
      calcedDistance = DistanceSquared(oldPos[0], 
				       oldPos[1], 
				       extremities[startCounter], 
				       extremities[startCounter + 1]);

#if BS_DEBUG
      fprintf(stderr, "TScan: calcedDistance = %f\n", calcedDistance);
#endif     

      if ((!flag) || 
	  (calcedDistance < Threshold && calcedDistance < DistanceSquared(oldPos[0],
									  oldPos[1],
									  newPos[0],
									  newPos[1])))
	{
	  newPos[0] = extremities[startCounter];
	  newPos[1] = extremities[startCounter + 1];
	  *newIndex = startCounter;
	  flag = 1;
	}
      startCounter += 3;
    }

/* KBR 7/1/94: inserted this to fix MAJOR bug that allowed a tracked point
   to leap major distances when going from tracking -> nontracking. */

  if ((flag) && (DistanceSquared(oldPos[0],
				 oldPos[1],
				 newPos[0],
				 newPos[1]) < Threshold))
    flag = 1;
  else
    {
      flag = 0;
      newPos[0] = tempPos[0];
      newPos[1] = tempPos[1];
    }
  return(flag);
}

void
  ScanContourByExtremityID(int whichExtremity,
			   float centroidx,
			   float centroidy,
			   float *contour,
			   int numOfPoints,
			   float tlx, float tly,
			   float brx, float bry,
			   float *newPos)
{
  int counter = 0;
  int pointQuad;
  int flag = 0;

  newPos[0] = centroidx;
  newPos[1] = centroidy;
  
  switch (whichExtremity)
    {

    case LEFT_FOOT_EXTREMITY_ID:
      {
	while (Quadrant(centroidx,
			centroidy,
			contour[counter],
			contour[counter + 1]) == 1)
	  {
	    if ((contour[counter] <= newPos[0]) &&
		(contour[counter + 1] <= newPos[1]))   /* Pick the point that's farthest down
							  and to the left on the contour */
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	  counter += 3;
	  }
	
	if (flag == 0)   /* In case of failure, put the left foot on the bottom of the bounding
			    box, 1/3 of the way over */
	  {
	    newPos[0] = tlx + (1. / 3.) * (brx - tlx);
	    newPos[1] = bry;
	  }
      }
      break;

    case LEFT_HAND_EXTREMITY_ID:
      {
	counter = 3;

	while (counter < numOfPoints - 3)
	  {
	    if ((contour[counter] < newPos[0]) &&
		((contour[counter - 2] < (.5 * (tly + bry))) &&
		 (contour[counter + 1] >= (.5 * (tly + bry)))))  /* Pick the outermost point
								    on the boundary from the
								    bottom half of the bounding box
								    to the upper half */
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	    counter += 3;
	  }
	
	if (flag == 0)  /* In failure case, put it on the left edge 
			   of the bounding box, halfway up */
	  {
	    newPos[0] = tlx;
	    newPos[1] = .5 * (tly + bry);
	  }

#if BS_DEBUG
      fprintf(stderr, "*******In left hand of ScanContourByExtremityID %d %d\n",
	      newPos[0], newPos[1]);
#endif     

      }
      break;

    case HEAD_EXTREMITY_ID:
      {
	while (counter < numOfPoints - 3)
	  {
	    if ((contour[counter + 1] >= newPos[1]) &&
		(fabs(contour[counter] - centroidx) < HEAD_THRESHOLD))  /* Pick the highest point
									   on the contour within
									   6 inches from the center
									   of the bounding box */
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	    counter += 3;	    
	  }

	if (flag == 0)
	  {
	    newPos[0] = centroidx;
	    newPos[1] = tly;
	  }
      }
      break;

    case RIGHT_HAND_EXTREMITY_ID:
      {
#if BS_DEBUG
      fprintf(stderr, "In right hand of ScanContourByExtremityID %d %d\n",
	      newPos[0], newPos[1]);
#endif     

	counter = numOfPoints - 6;

	while (counter >= 0)
	  {
	    if ((contour[counter] > newPos[0]) &&
		((contour[counter + 4] < (.5 * (tly + bry))) &&
		 (contour[counter + 1] >= (.5 * (tly + bry)))))  /* Pick the farthest point to the right
								    of the contour that's on the top half/
								    bottom half boundary of the bounding box */
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	    counter -= 3;
	  }

	if (flag == 0)  /* Failure case: put the hand on the right edge of the bbox halfway up */
	  {
	    newPos[0] = brx;
	    newPos[1] = bry + .5 * (tly - bry);
	  }
	     
#if BS_DEBUG
      fprintf(stderr, "In right hand of ScanContourByExtremityID %f %f counter: %d\n",
	      newPos[0], newPos[1], counter);
#endif     

      }
      break;
    case RIGHT_FOOT_EXTREMITY_ID:
      {
	counter = numOfPoints - 3;
       
	while (Quadrant(centroidx,
			centroidy,
			contour[counter],
			contour[counter + 1]) == 4)  /* Pick the point that's farthest down and to the right */
	  {
	    if ((contour[counter] >= newPos[0]) &&
		(contour[counter + 1] <= newPos[1]))
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	    counter -= 3;
	  }

	if (flag == 0)  /* Stick the foot on the bottom of the bounding box 2/3 of the way over */
	  {
	    newPos[0] = tlx + (COORD_TYPE) ((2. / 3.) * (float) (brx - tlx));
	    newPos[1] = bry;
	  }
      }
      break;
    }
}

int
  PickPointByExtremityID(int whichExtremity,
			 float *extremities,
			 int startCounter,
			 int endCounter,
			 float *newPos,
			 int *newIndex,
			 float centroidx, 
			 float centroidy,
			 float tlx, float tly,
			 float brx, float bry)
{
  int flag = 0;

  newPos[0] = centroidx;
  newPos[1] = centroidy;
  
  switch (whichExtremity)
    {
    case LEFT_FOOT_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((Quadrant(centroidx,
			centroidy,
			extremities[startCounter],
			extremities[startCounter + 1]) == 1) &&
	      (!(InTopOfBoundingBox(tlx, tly,
				    brx, bry,
				    extremities[startCounter + 1]))) &&
	      (extremities[startCounter] < newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case LEFT_HAND_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((InTopOfBoundingBox(tlx, tly,
				  brx, bry,
				  extremities[startCounter + 1])) &&
	      ((centroidx - extremities[startCounter]) > HAND_THRESHOLD) &&
	      (extremities[startCounter] < newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case HEAD_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((extremities[startCounter + 1] > newPos[1]) &&
	      (fabs(extremities[startCounter] - centroidx) < HEAD_THRESHOLD))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case RIGHT_HAND_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((InTopOfBoundingBox(tlx, tly,
				  brx, bry,
				  extremities[startCounter + 1])) &&
	      ((extremities[startCounter] - centroidx > HAND_THRESHOLD)) &&
	      (extremities[startCounter] > newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case RIGHT_FOOT_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	if ((Quadrant(centroidx,
		      centroidy,
		      extremities[startCounter],
		      extremities[startCounter + 1]) == 4) &&
	    (!(InTopOfBoundingBox(tlx, tly,
				  brx, bry,
				  extremities[startCounter + 1]))) &&
	    (extremities[startCounter] > newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    default:
      fprintf(stderr, "Fatal: PickPointByExtremityID\n");
    }
  return(flag);
}

void
  LookForNewPoint(int whichExtremity,
		  float centroidx,
		  float centroidy,
		  float tlx, float tly,
		  float brx, float bry,
		  float *contour,
		  int contourLength,
		  unsigned long int frameNumber,
		  float *extremityCentroidList,
		  int startCounter,
		  int endCounter,
		  int *centIndex,
		  float *oldPos,
		  unsigned char oldValidBit,
		  float *lastValidPos,
		  unsigned long int lastValidFrame,
		  unsigned char *validBit,
		  float *newPos)
{
  float latestPos[2] = {0., 0.};
  int latestIndex = 0;
  static float Threshold = 144.;  /* this is 12 inches squared */
  static int validCounter[5] = {0,0,0,0,0};
  int thad_counter;
  int ThreshScannedFlag = 0;
  int ThreshScanSucceededFlag = 0;
  int LastValidFrameStillValid = 0;

  *validBit = 0;  /* Default is to be invalid */

  newPos[0] = centroidx;
  newPos[1] = centroidy;

  if (oldValidBit == 2)
    {
      ThreshScannedFlag = 1;
      ThreshScanSucceededFlag = ThresholdScan(oldPos,
					      extremityCentroidList,
					      startCounter,
					      endCounter,
					      latestPos,
					      &latestIndex,
					      Threshold);
    }
  else
    {
      if (lastValidFrame > 0 && 
	  (frameNumber - lastValidFrame) < VALID_EXPIRATION_TIME) /* Last valid value still okay
								     to use as a reference */
	{
	  LastValidFrameStillValid = 1;
	  ThreshScannedFlag = 1;
	  ThreshScanSucceededFlag = ThresholdScan(lastValidPos,
						  extremityCentroidList,
						  startCounter,
						  endCounter,
						  latestPos,
						  &latestIndex,
						  Threshold);  /* new point within threshold of old */
	}
    }

  if (ThreshScanSucceededFlag)
    {
      *centIndex = (3 + latestIndex);  
      newPos[0] = latestPos[0];
      newPos[1] = latestPos[1];
      *validBit = 2;  /* flag this value as valid too */
    }
  else     /* This means that if (a)
	      we didn't do a threshold
	      scan, or (b) that scan
	      failed, do the failure cases.
	      If you don't get it, set up a
	      truth table! */
    {
      if (PickPointByExtremityID(whichExtremity,
				 extremityCentroidList,
				 startCounter,
				 endCounter,
				 latestPos,
				 &latestIndex,
				 centroidx,
				 centroidy,
				 tlx, tly,
				 brx, bry))   /* We found an extremity in the
						 default section of the bounding box */
	{
	  *validBit = 1;
	  newPos[0] = latestPos[0];
	  newPos[1] = latestPos[1];
	  if (!LastValidFrameStillValid)  /* Try to regain tracking */
	    {
	      if (DistanceSquared(latestPos[0],
				  latestPos[1],
				  oldPos[0],
				  oldPos[1]) < Threshold)
		{
		  validCounter[whichExtremity - 1]++;
		  if (validCounter[whichExtremity - 1] > VALID_TRACKING_TIME)
                                         /* Extremity was found over several frames */
		    {
		      *validBit = 2;  /* IMPORTANT!
					 This is how we get back on track
					 after a series of invalid tracks. */
		      validCounter[whichExtremity - 1] = 0;
		      *centIndex += (3 + latestIndex);
		    }
		}
	      else
		{
		  validCounter[whichExtremity - 1] = 0;
		}
	    }
	}
      else
	{
	  if (LastValidFrameStillValid)  /* Prevents an extremity from glitching if
					    tracking loses for just a short time */
	    {
	      newPos[0] = lastValidPos[0];
	      newPos[1] = lastValidPos[1];
	    }
	  else
	    {
	      ScanContourByExtremityID(whichExtremity,
				       centroidx,
				       centroidy,
				       contour,
				       contourLength,
				       tlx, tly,
				       brx, bry,
				       newPos);  /* Last failure case */
	    }
	}
    }
}  /* end of procedure */

void
  LookForLeftFoot(float centroidx, float centroidy,
		  float tlx, float tly,
		  float brx, float bry,
		  float *contour,
		  int contourLength,
		  unsigned long int frameNumber,
		  float *extremityCentroidList,
		  int extListLength,
		  int *centIndex,
		  float *oldPos,
		  bitMaskType oldBitMask,
		  float *lastValidPos,
		  unsigned long int lastValidFrame,
		  bitMaskType *newBitMask,
		  float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = *centIndex;

  if (frameNumber - lastValidFrame >= VALID_EXPIRATION_TIME)
    {
      while ((endCounter < extListLength) && 
	     (Quadrant(centroidx, 
		       centroidy, 
		       extremityCentroidList[endCounter], 
		       extremityCentroidList[endCounter + 1]) == 1) &&
	     (InBottomOfBoundingBox(tlx, tly, brx, bry, extremityCentroidList[endCounter + 1])))
	endCounter += 3;
    }
  else
    endCounter = extListLength;

  LookForNewPoint(LEFT_FOOT_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.left_foot_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->left_foot_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForLeftHand(float centroidx, float centroidy,
		  float tlx, float tly,
		  float brx, float bry,
		  float *contour,
		  int contourLength,
		  unsigned long int frameNumber,
		  float *extremityCentroidList,
		  int extListLength,
		  int *centIndex,
		  float *oldPos,
		  bitMaskType oldBitMask,
		  float *lastValidPos,
		  unsigned long int lastValidFrame,
		  bitMaskType *newBitMask,
		  float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = *centIndex;
  int pointQuad;

  if (frameNumber - lastValidFrame >= VALID_EXPIRATION_TIME)
    {

      /* thad-do we need to flush out bad extremities here ?*/
      pointQuad = Quadrant(centroidx, 
			   centroidy, 
			   extremityCentroidList[endCounter], 
			   extremityCentroidList[endCounter + 1]);
      
      while((endCounter < extListLength) && 
	    (pointQuad == 1 || pointQuad == 2))
	{
	  endCounter += 3;
	  if (endCounter < extListLength)
	    pointQuad = Quadrant(centroidx,
				 centroidy, 
				 extremityCentroidList[endCounter], 
				 extremityCentroidList[endCounter + 1]);
	}
    }
  else
    {
      startCounter = 0;
      endCounter = extListLength;
    }

  LookForNewPoint(LEFT_HAND_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.left_hand_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->left_hand_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForHead(float centroidx, float centroidy,
	      float tlx, float tly,
	      float brx, float bry,
	      float *contour,
	      int contourLength,
	      unsigned long int frameNumber,
	      float *extremityCentroidList,
	      int extListLength,
	      int *centIndex,
	      float *oldPos,
	      bitMaskType oldBitMask,
	      float *lastValidPos,
	      unsigned long int lastValidFrame,
	      bitMaskType *newBitMask,
	      float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = *centIndex;
  int pointQuad = 0;
  static float HeadThreshold = 100.;  /* this is 10 inches squared */

  if (frameNumber - lastValidFrame >= VALID_EXPIRATION_TIME)
    {
      while ((endCounter < extListLength) &&
	     (Quadrant(centroidx,
		       centroidy,
		       extremityCentroidList[endCounter],
		       extremityCentroidList[endCounter + 1]) != 4))
	endCounter += 3;

      if ((endCounter == startCounter) &&
	  (endCounter != extListLength))
	endCounter += 3;
    }
  else
    {
      startCounter = 0;
      endCounter = extListLength;
    }

  LookForNewPoint(HEAD_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.head_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->head_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForRightHand(float centroidx, float centroidy,
		   float tlx, float tly,
		   float brx, float bry,
		   float *contour,
		   int contourLength,
		   unsigned long int frameNumber,
		   float *extremityCentroidList,
		   int extListLength,
		   int *centIndex,
		   float *oldPos,
		   bitMaskType oldBitMask,
		   float *lastValidPos,
		   unsigned long int lastValidFrame,
		   bitMaskType *newBitMask,
		   float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = extListLength;
  int pointQuad;

#if BS_DEBUG_THAD
  fprintf(stderr, "extremity list length: %d\n", extListLength);
#endif

  
  if (frameNumber - lastValidFrame < VALID_EXPIRATION_TIME)
    startCounter = 0;

  LookForNewPoint(RIGHT_HAND_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.right_hand_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->right_hand_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForRightFoot(float centroidx, float centroidy,
		   float tlx, float tly,
		   float brx, float bry,
		   float *contour,
		   int contourLength,
		   unsigned long int frameNumber,
		   float *extremityCentroidList,
		   int extListLength,
		   int *centIndex,
		   float *oldPos,
		   bitMaskType oldBitMask,
		   float *lastValidPos,
		   unsigned long int lastValidFrame,
		   bitMaskType *newBitMask,
		   float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = extListLength;

  /* Flush out bad extremities in case previous ones weren't valid and
     didn't take care of this for us */

  if (frameNumber - lastValidFrame < VALID_EXPIRATION_TIME)
    startCounter = 0;

  LookForNewPoint(RIGHT_FOOT_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.right_foot_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->right_foot_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  DoConfidenceValueReassignment(float centroidx,
				float centroidy,
				float *contour,
				int numOfPoints,
				float tlx, float tly,
				float brx, float bry,
				float *firstValue,
				unsigned char firstValidBit,
				int firstExtremityID,
				float *secondValue,
				unsigned char secondValidBit,
				int secondExtremityID)
{
  if (firstValidBit)
    {
      ScanContourByExtremityID(secondExtremityID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       secondValue);
    }
  else
    {
      ScanContourByExtremityID(firstExtremityID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       firstValue);
    }
}

void
  DoFinalSanityCheck(float centroidx,
		     float centroidy,
		     float *contour,
		     int numOfPoints,
		     float tlx, float tly,
		     float brx, float bry,
		     float *newLeftFoot,
		     float *newLeftHand,
		     float *newHead,
		     float *newRightHand,
		     float *newRightFoot,
		     bitMaskType *newBitMask,
		     float Threshold)
{
#if SANITY_DEBUG
  fprintf(stderr, "dofinalsanitycheck values:\n");
  fprintf(stderr, "left foot: %f %f\nleft hand: %f %f\n",
	  newLeftFoot[0],
	  newLeftFoot[1],
	  newLeftHand[0],
	  newLeftHand[1]);
  fprintf(stderr, "head: %f %f\nright hand: %f %f\nright foot: %f %f\n",
	  newHead[0],
	  newHead[1],
	  newRightHand[0],
	  newRightHand[1],
	  newRightFoot[0],
	  newRightFoot[1]);
#endif

  if (DistanceSquared(newLeftFoot[0], 
		      newLeftFoot[1], 
		      newLeftHand[0], 
		      newLeftHand[1]) < Threshold)
    {
      newBitMask->left_foot_status = 0;
      newBitMask->left_hand_status = 0;
#if SANITY_DEBUG
      fprintf(stderr, "bashed left foot and left hand\n");
#endif
      ScanContourByExtremityID(LEFT_FOOT_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newLeftFoot);
      ScanContourByExtremityID(LEFT_HAND_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newLeftHand);
    }
  if (DistanceSquared(newHead[0], 
		      newHead[1], 
		      newLeftHand[0], 
		      newLeftHand[1]) < Threshold)
    {
      newBitMask->head_status = 0;
      newBitMask->left_hand_status = 0;
#if SANITY_DEBUG
      fprintf(stderr, "bashed left hand and head\n");
#endif
      ScanContourByExtremityID(LEFT_HAND_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newLeftHand);
      ScanContourByExtremityID(HEAD_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newHead);
    }
  if (DistanceSquared(newHead[0], 
		      newHead[1], 
		      newRightHand[0], 
		      newRightHand[1]) < Threshold)
    {
      newBitMask->head_status = 0;
      newBitMask->right_hand_status = 0;
#if SANITY_DEBUG
      fprintf(stderr, "bashed head and right hand\n");
#endif
      ScanContourByExtremityID(HEAD_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newHead);
      ScanContourByExtremityID(RIGHT_HAND_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newRightHand);
    }
  if (DistanceSquared(newRightHand[0], 
		      newRightHand[1], 
		      newRightFoot[0], 
		      newRightFoot[1]) < Threshold)
    {
      newBitMask->right_hand_status = 0;
      newBitMask->right_foot_status = 0;
#if SANITY_DEBUG
      fprintf(stderr, "bashed right hand and right foot\n");
#endif
      ScanContourByExtremityID(RIGHT_HAND_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newRightHand);
      ScanContourByExtremityID(RIGHT_FOOT_EXTREMITY_ID,
			       centroidx,
			       centroidy,
			       contour,
			       numOfPoints,
			       tlx, tly,
			       brx, bry,
			       newRightFoot);
    }
#if SANITY_DEBUG
  printf("head %f %f\n", newHead[0], newHead[1]);
  fprintf(stderr, "left hand %f %f\n", newLeftHand[0], newLeftHand[1]);
  fprintf(stderr, "right hand %f %f\n", newRightHand[0], newRightHand[1]);
  fprintf(stderr, "left foot %f %f\n", newLeftFoot[0], newLeftFoot[1]);
  fprintf(stderr, "right foot %f %f\n", newRightFoot[0],
	  newRightFoot[1]);
  fflush(stdout);
  fflush(stderr);
#endif
#if SLEEP
  sleep(1);
#endif
}      

void
  trackBodyFeatures(float *contour, /* 3 coords */
		    int numPoints,
		    float *extrema, /* 3 coords */
		    int numExtrema,
		    float cx, float cy,   /* centroid x, centroid y */
		    float tlx, float tly, /* bounding box: top left x, top left y */
		    float brx, float bry, /* bounding box: bottom right x, bottom right y */
		    float *head,          /* index 0 = x coord, index 1 = y coord */
		    float *left_hand,
		    float *right_hand,
		    float *left_foot,
		    float *right_foot,
		    boolean *head_valid,
		    boolean *left_hand_valid,
		    boolean *right_hand_valid,
		    boolean *left_foot_valid,
		    boolean *right_foot_valid)
{
  static unsigned long int frameNumber = 0;
  static unsigned long int lastValidLeftFootFrame = 0;
  static unsigned long int lastValidRightFootFrame = 0;
  static unsigned long int lastValidLeftHandFrame = 0;
  static unsigned long int lastValidRightHandFrame = 0;
  static unsigned long int lastValidHeadFrame = 0;
  static float sanityThreshold = 40.64;  /* this is 6.375 squared */  /* old value was 8 */

  static bitMaskType oldBitMask = {0,0,0,0,0}, bitMask = {0,0,0,0,0};
  static float oldHead[2] = {0., 0.};
  static float oldLeftHand[2] = {0., 0.};
  static float oldRightHand[2] = {0., 0.};
  static float oldLeftFoot[2] = {0., 0.};
  static float oldRightFoot[2] = {0., 0.};
  static float lastValidHead[2] = {0., 0.};
  static float lastValidLeftHand[2] = {0., 0.};
  static float lastValidRightHand[2] = {0., 0.};
  static float lastValidLeftFoot[2] = {0., 0.};
  static float lastValidRightFoot[2] = {0., 0.};
  int centIndex = 0;
  
  sanityCheck(extrema,
	      numExtrema,
	      tlx,
	      tly,
	      brx,
	      bry);

#if BS_DEBUG_THAD
  printf("test1\n");
#endif

  LookForLeftFoot(cx, cy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  (numPoints * 3),
		  frameNumber,
		  extrema,
		  (numExtrema * 3),
		  &centIndex,
		  oldLeftFoot,
		  oldBitMask,
		  lastValidLeftFoot,
		  lastValidLeftFootFrame,
		  &bitMask,
		  left_foot);
  LookForLeftHand(cx, cy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  (numPoints * 3),
		  frameNumber,
		  extrema,
		  (numExtrema * 3),
		  &centIndex,
		  oldLeftHand,
		  oldBitMask,
		  lastValidLeftHand,
		  lastValidLeftHandFrame,
		  &bitMask,
		  left_hand);
  LookForHead(cx, cy,
	      tlx, tly,
	      brx, bry,
	      contour,
	      (numPoints * 3),
	      frameNumber,
	      extrema,
	      (numExtrema * 3),
	      &centIndex,
	      oldHead,
	      oldBitMask,
	      lastValidHead,
	      lastValidHeadFrame,
	      &bitMask,
	      head);
  LookForRightHand(cx, cy,
		   tlx, tly,
		   brx, bry,
		   contour,
		   (numPoints * 3),
		   frameNumber,
		   extrema,
		   (numExtrema * 3),
		   &centIndex,
		   oldRightHand,
		   oldBitMask,
		   lastValidRightHand,
		   lastValidRightHandFrame,
		   &bitMask,
		   right_hand);
  LookForRightFoot(cx, cy,
		   tlx, tly,
		   brx, bry,
		   contour,
		   (numPoints * 3),
		   frameNumber,
		   extrema,
		   (numExtrema * 3),			   
		   &centIndex,
		   oldRightFoot,
		   oldBitMask,
		   lastValidRightFoot,
		   lastValidRightFootFrame,
		   &bitMask,
		   right_foot);
  DoFinalSanityCheck(cx, cy,
		     contour,
		     (numPoints * 3),
		     tlx, tly,
		     brx, bry,
		     left_foot,
		     left_hand,
		     head,
		     right_hand,
		     right_foot,
		     &bitMask,
		     sanityThreshold);
  

#if BS_DEBUG_THAD
  printf("bb upper left %f %f lower right %f %f\n", 
	  tlx,
	  tly,
	  brx,
	  bry);
#endif
	
  if (bitMask.head_status == 2)
    {
      head[0] = (.1 * oldHead[0]) + (.9 * head[0]);
      head[1] = (.1 * oldHead[1]) + (.9 * head[1]);
      lastValidHead[0] = head[0];
      lastValidHead[1] = head[1];
      lastValidHeadFrame = frameNumber;
    }
  else
    {
      head[0] = (.3 * oldHead[0]) + (.7 * head[0]);
      head[1] = (.3 * oldHead[1]) + (.7 * head[1]);
    }

  if (bitMask.left_foot_status == 2)
    {
      left_foot[0] = (.1 * oldLeftFoot[0]) + (.9 * left_foot[0]);
      left_foot[1] = (.1 * oldLeftFoot[1]) + (.9 * left_foot[1]);
      lastValidLeftFoot[0] = left_foot[0];
      lastValidLeftFoot[1] = left_foot[1];
      lastValidLeftFootFrame = frameNumber;
    }
  else
    {
      left_foot[0] = (.3 * oldLeftFoot[0]) + (.7 * left_foot[0]);
      left_foot[1] = (.3 * oldLeftFoot[1]) + (.7 * left_foot[1]);
    }

  if (bitMask.left_hand_status == 2)
    {
      left_hand[0] = (.1 * oldLeftHand[0]) + (.9 * left_hand[0]);
      left_hand[1] = (.1 * oldLeftHand[1]) + (.9 * left_hand[1]);
      lastValidLeftHand[0] = left_hand[0];
      lastValidLeftHand[1] = left_hand[1];
      lastValidLeftHandFrame = frameNumber;
    }
  else
    {
      left_hand[0] = (.3 * oldLeftHand[0]) + (.7 * left_hand[0]);
      left_hand[1] = (.3 * oldLeftHand[1]) + (.7 * left_hand[1]);
    }      

  if (bitMask.right_hand_status == 2)
    {
      right_hand[0] = (.1 * oldRightHand[0]) + (.9 * right_hand[0]);
      right_hand[1] = (.1 * oldRightHand[1]) + (.9 * right_hand[1]);
      lastValidRightHand[0] = right_hand[0];
      lastValidRightHand[1] = right_hand[1];
      lastValidRightHandFrame = frameNumber;
    }
  else
    {
      right_hand[0] = (.3 * oldRightHand[0]) + (.7 * right_hand[0]);
      right_hand[1] = (.3 * oldRightHand[1]) + (.7 * right_hand[1]);
    }

  if (bitMask.right_foot_status == 2)
    {
      right_foot[0] = (.1 * oldRightFoot[0]) + (.9 * right_foot[0]);
      right_foot[1] = (.1 * oldRightFoot[1]) + (.9 * right_foot[1]);
      lastValidRightFoot[0] = right_foot[0];
      lastValidRightFoot[1] = right_foot[1];
      lastValidRightFootFrame = frameNumber;
    }
  else
    {
      right_foot[0] = (.3 * oldRightFoot[0]) + (.7 * right_foot[0]);
      right_foot[1] = (.3 * oldRightFoot[1]) + (.7 * right_foot[1]);
    }
  
  frameNumber += 1;

  oldBitMask = bitMask;
  oldLeftFoot[0] = left_foot[0];
  oldLeftFoot[1] = left_foot[1];
  oldRightFoot[0] = right_foot[0];
  oldRightFoot[1] = right_foot[1];
  oldLeftHand[0] = left_hand[0];
  oldLeftHand[1] = left_hand[1];
  oldRightHand[0] = right_hand[0];
  oldRightHand[1] = right_hand[1];
  oldHead[0] = head[0];
  oldHead[1] = head[1];

  *head_valid = bitMask.head_status;
  *left_hand_valid = bitMask.left_hand_status;
  *left_foot_valid = bitMask.left_foot_status;
  *right_hand_valid = bitMask.right_hand_status;
  *right_foot_valid = bitMask.right_foot_status;
  

}



