#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>
#include <bstring.h>

#include "common.h"
#include "ring_buffer.h"
#include "user_model.h"

/* protos */
#include "common.proto"
#include "ring_buffer.proto"
#include "user_model.proto"

/* GESTUREPARAMS */

GestureParams *
makeGestureParams(void)
{
  GestureParams *gp = (GestureParams *) getMemory(sizeof(GestureParams));
  
  return gp;
}

void
freeGestureParams(GestureParams *gp)
{
  free(gp);
}


void
loadGestureParams(GestureParams *gp, char *filename)
{
  FILE *fp;
  char buf[128];
  
  if ((fp = fopen(filename, "r")) == NULL)
    {
      fprintf(stderr, "loadGestureParams: can't open %s\n", filename);
      return;
    }
  
  fscanf(fp, "%s %f\n", buf, &(gp->hysteresis));
  fscanf(fp, "%s %f\n", buf, &(gp->pointing_width));
  fscanf(fp, "%s %f\n", buf, &(gp->sitting_height));
  fscanf(fp, "%s %f\n", buf, &(gp->standing_height));
  fscanf(fp, "%s %f\n", buf, &(gp->spread_eagle_width));
  fscanf(fp, "%s %f\n", buf, &(gp->pointing_still_time));
  fscanf(fp, "%s %f\n", buf, &(gp->max_dist_for_hand_not_moving));
  fscanf(fp, "%s %f\n", buf, &(gp->throwing_thresh));
  fscanf(fp, "%s %f\n", buf, &(gp->throwing_time));
  fscanf(fp, "%s %d\n", buf, &(gp->oscillations_for_wave));
  fscanf(fp, "%s %f\n", buf, &(gp->waving_thresh));
  fscanf(fp, "%s %f\n", buf, &(gp->min_hand_dist_for_wave));
  fscanf(fp, "%s %f\n", buf, &(gp->min_foot_height_for_kick));
  fscanf(fp, "%s %f\n", buf, &(gp->kicking_thresh));
  fscanf(fp, "%s %f\n", buf, &(gp->hand_over_head_height));
  fscanf(fp, "%s %f\n", buf, &(gp->min_hand_dist_for_extend));
  fscanf(fp, "%s %f\n", buf, &(gp->min_hand_move_dist_for_wave));
  fscanf(fp, "%s %f\n", buf, &(gp->bending_over_height));
  
  fclose(fp);
  
}



/* CameraToWorld */

CameraToWorld *
makeCameraToWorld(void)
{
  CameraToWorld *ctw = (CameraToWorld *) getMemory(sizeof(CameraToWorld));

  ctw->implemented = FALSE;
  return ctw;
}

void
freeCameraToWorld(CameraToWorld *ctw)
{
  free(ctw);
}

/********/
/* Time */
/********/

Timer
getTime(void)
{
  struct timeval time;
  struct timezone tzp;

  /* punt time zone.  Only used for relative */
  gettimeofday(&time, &tzp);

  return time;  
}

/* in seconds */
float
durationTime(Timer t1, Timer t2)
{
  float secs;
  float usecs;
  
  secs = (float) (t2.tv_sec - t1.tv_sec);
  usecs = (float) (t2.tv_usec - t1.tv_usec);
  usecs = usecs * (float) (1.0e-6); /* turns into seconds */
  
  return (secs + usecs);
  
}



/*************/
/* UserModel */
/*************/

UserModel *
makeUserModel(GestureParams *gp, CameraToWorld *ctw)
{
  UserModel *um = (UserModel *) getMemory(sizeof(UserModel));
  int i;
  
  um->gp = gp;
  um->camera = ctw;
  
  um->buffer = (RingBuffer **) getMemory(sizeof(RingBuffer *) * 
					  NUMBER_OF_BUFFERS);
  for(i=0; i<NUMBER_OF_BUFFERS; i++)
    um->buffer[i] = makeRingBuffer(BUFFER_LENGTH);
  
  um->normal_width = NORMAL_WIDTH;
  um->normal_height=0.0;
  
  um->user_exists = FALSE;
  um->acquired_model = FALSE;
  um->ticks = 0;

  return um;

}

/* I do not free CameraToWorld or GestureParams */
void
freeUserModel(UserModel *um)
{
  int i;
  
  for(i=0; i< NUMBER_OF_BUFFERS; i++)
    freeRingBuffer(um->buffer[i]);
  
  free(um);
  
}

void
resetUserModel(UserModel *um)
{
  int i;
  
  if (um->ticks == 0)
    {
      return;
    }
  
  um->ticks = 0;
  um->user_exists = FALSE;
  um->acquired_model = FALSE;
  
  for(i=0; i<NUMBER_OF_BUFFERS; i++)
    rbReset(um->buffer[i]);

  fprintf(stderr, "Lost user.  Resetting\n");
  
}


/* Expects WORLD coordinates */
void
updateUserModel(UserModel *um,
		float cx,
		float cy,
		float left,
		float right,
		float top,
		float bottom,
		float hlx,
		float hly,
		float hrx,
		float hry,
		float flx,
		float fly,
		float frx,
		float fry,
		float hlv, /* hand valid confidence */
		float hrv)
{
  Timer now = getTime();
  float dur = durationTime(um->start_time, now);
  boolean error;
  float p1;
  float dt;
  
#ifdef DEBUG
  printf("C(%f %f) BB(%f,%f,%f,%f) HL(%f %f) HR(%f %f) FL(%f %f) FR(%f %f) T(%f)\n",
	 cx,cy,left,right,top,bottom,hlx,hly,hrx,hry,flx,fly,frx,fry,dur);

#endif

  if (!(um->user_exists))
    {
      fprintf(stderr, "Got new user\n");
      um->ticks = 0;
      um->start_time = getTime();
      um->user_exists = TRUE;     
      um->normal_height = 65.0;
      um->normal_width = NORMAL_WIDTH;
    }
  
  um->ticks++;
  
  /* average width and height */
  if (um->ticks < FRAMES_FOR_AVERAGE + FRAMES_BEFORE_AVERAGE && um->ticks >= FRAMES_BEFORE_AVERAGE)
    {
      float ticks = um->ticks;
      /* make width smarter */
      /* um->normal_width = (um->normal_width * (float) (ticks-1) +
	 (right - left)) / (float) ticks; */

      um->normal_height = (um->normal_height * (float) (ticks-1) +
			   (top - bottom)) / (float) ticks;
      um->normal_width = um->normal_height * (WIDTH_FRACTION_OF_HEIGHT);
      printf("Normal width/height: %f %f\n", um->normal_width,
	     um->normal_height);
      
    }
  else
    {
      um->acquired_model = TRUE;
    }
  
  /* get time diff between frames */
  dt = dur - rbGetElement(um->buffer[TIME], 0, &error);
  
  rbAddElement(um->buffer[TIME], dur);
  
  /* update derivatives */
  p1 = rbGetElement(um->buffer[LEFT], 0, &error);
  if (error)
    rbAddElement(um->buffer[DLEFT], 0.0);
  else
    rbAddElement(um->buffer[DLEFT], (left - p1)/dt);
  
  /* update derivatives */
  p1 = rbGetElement(um->buffer[RIGHT], 0, &error);
  if (error)
    rbAddElement(um->buffer[DRIGHT], 0.0);
  else
    rbAddElement(um->buffer[DRIGHT], (right - p1)/dt);
  

  /* update derivatives */
  p1 = rbGetElement(um->buffer[TOP], 0, &error);
  if (error)
    rbAddElement(um->buffer[DTOP], 0.0);
  else
    rbAddElement(um->buffer[DTOP], (top - p1)/dt);
  

  /* update derivatives */
  p1 = rbGetElement(um->buffer[BOTTOM], 0, &error);
  if (error)
    rbAddElement(um->buffer[DBOTTOM], 0.0);
  else
    rbAddElement(um->buffer[DBOTTOM], (bottom - p1)/dt);

  /* update derivatives */
  p1 = rbGetElement(um->buffer[HAND_LEFT_X], 0, &error);
  if (error)
    rbAddElement(um->buffer[DHAND_LEFT_X], 0.0);
  else
    rbAddElement(um->buffer[DHAND_LEFT_X], (hlx - p1)/dt);

  /* update derivatives */
  p1 = rbGetElement(um->buffer[HAND_LEFT_Y], 0, &error);
  if (error)
    rbAddElement(um->buffer[DHAND_LEFT_Y], 0.0);
  else
    rbAddElement(um->buffer[DHAND_LEFT_Y], (hly - p1)/dt);

  /* update derivatives */
  p1 = rbGetElement(um->buffer[HAND_RIGHT_X], 0, &error);
  if (error)
    rbAddElement(um->buffer[DHAND_RIGHT_X], 0.0);
  else
    rbAddElement(um->buffer[DHAND_RIGHT_X], (hrx - p1)/dt);

  /* update derivatives */
  p1 = rbGetElement(um->buffer[HAND_RIGHT_Y], 0, &error);
  if (error)
    rbAddElement(um->buffer[DHAND_RIGHT_Y], 0.0);
  else
    rbAddElement(um->buffer[DHAND_RIGHT_Y], (hry - p1)/dt);
  

  /* foot derivs */
  /* update derivatives */
  p1 = rbGetElement(um->buffer[FOOT_LEFT_X], 0, &error);
  if (error)
    rbAddElement(um->buffer[DFOOT_LEFT_X], 0.0);
  else
    rbAddElement(um->buffer[DFOOT_LEFT_X], (flx - p1)/dt);

  /* update derivatives */
  p1 = rbGetElement(um->buffer[FOOT_LEFT_Y], 0, &error);
  if (error)
    rbAddElement(um->buffer[DFOOT_LEFT_Y], 0.0);
  else
    rbAddElement(um->buffer[DFOOT_LEFT_Y], (fly - p1)/dt);

  /* update derivatives */
  p1 = rbGetElement(um->buffer[FOOT_RIGHT_X], 0, &error);
  if (error)
    rbAddElement(um->buffer[DFOOT_RIGHT_X], 0.0);
  else
    rbAddElement(um->buffer[DFOOT_RIGHT_X], (frx - p1)/dt);

  /* update derivatives */
  p1 = rbGetElement(um->buffer[FOOT_RIGHT_Y], 0, &error);
  if (error)
    rbAddElement(um->buffer[DFOOT_RIGHT_Y], 0.0);
  else
    rbAddElement(um->buffer[DFOOT_RIGHT_Y], (fry - p1)/dt);
  
  
  rbAddElement(um->buffer[CENTROID_X], cx);
  rbAddElement(um->buffer[CENTROID_Y], cy);
  rbAddElement(um->buffer[LEFT], left);
  rbAddElement(um->buffer[RIGHT], right);
  rbAddElement(um->buffer[TOP], top);
  rbAddElement(um->buffer[BOTTOM], bottom);
  rbAddElement(um->buffer[HAND_LEFT_X], hlx);
  rbAddElement(um->buffer[HAND_LEFT_Y], hly);
  rbAddElement(um->buffer[HAND_RIGHT_X], hrx);
  rbAddElement(um->buffer[HAND_RIGHT_Y], hry);
  rbAddElement(um->buffer[FOOT_LEFT_X], flx);
  rbAddElement(um->buffer[FOOT_LEFT_Y], fly);
  rbAddElement(um->buffer[FOOT_RIGHT_X], frx);
  rbAddElement(um->buffer[FOOT_RIGHT_Y], fry);
  rbAddElement(um->buffer[HAND_LEFT_VALID], hlv);
  rbAddElement(um->buffer[HAND_RIGHT_VALID], hrv);

  
}


float
normalizedHeight(UserModel *um, int tick, boolean *error)
{
  float height;
  
  height = rbGetElement(um->buffer[TOP], tick, error);
  if (*error) return 0.0;
  
  height -= rbGetElement(um->buffer[BOTTOM], tick, error);
  if (*error) return 0.0;
  
  height = (float) fabsf((float)height);
  height /= um->normal_height; /* normalize to normal height */
  return height;

}

float
normalizedWidth(UserModel *um, int tick, boolean *error)
{
  float width;
  
  width = rbGetElement(um->buffer[RIGHT], tick, error);
  if (*error) return 0.0;
  
  width -= rbGetElement(um->buffer[LEFT], tick, error);
  if (*error) return 0.0;
  
  width = (float) fabsf((float) width);
  width  /= um->normal_width; /* normalize to normal width */
  return width;

}


boolean
personSitting(UserModel *um)
{
  static boolean sitting = FALSE; /* memory.  Were we sitting last call */
  float thresh;
  GestureParams *gp = um->gp;
  float height;
  boolean error;
  
  if (sitting) 
    thresh = gp->sitting_height + gp->hysteresis;
  else
    thresh = gp->sitting_height - gp->hysteresis;

  height = normalizedHeight(um, 0, &error);
  if (error) return FALSE;
  
  sitting = (boolean) (height <= thresh);
  return sitting;    
      
}



boolean
personBending(UserModel *um)
{
  static boolean bending = FALSE; /* memory.  Were we sitting last call */
  float thresh;
  GestureParams *gp = um->gp;
  float height;
  boolean error;
  
  if (bending) 
    thresh = gp->bending_over_height + gp->hysteresis;
  else
    thresh = gp->bending_over_height - gp->hysteresis;

  height = normalizedHeight(um, 0, &error);
  if (error) return FALSE;
  
  bending = (boolean) (height <= thresh);
  return bending;    
      
}


boolean
personHasHandOverHead(UserModel *um)
{
  static boolean tall = FALSE; /* memory.  Were we tall last call */
  float thresh;
  GestureParams *gp = um->gp;
  float height;
  boolean error;
  
  if (tall) 
    thresh = gp->hand_over_head_height - gp->hysteresis;
  else
    thresh = gp->hand_over_head_height + gp->hysteresis;

  
  
  height = normalizedHeight(um, 0, &error);
  if (error) return FALSE;
  
  tall  = (boolean) (height >= thresh);
  return tall;    
}


boolean
personPointing(UserModel *um, Hand *hand)
{
  static boolean left_pointing = FALSE;
  static boolean right_pointing = FALSE;
  float pthresh, sthresh;
  float width; /* width and normalized width */
  float bb_left, bb_right;
  boolean error;
  GestureParams *gp = um->gp;
  
  bb_left = rbGetElement(um->buffer[LEFT], 0, &error);
  if (error) return FALSE;

  bb_right = rbGetElement(um->buffer[RIGHT], 0, &error);
    
  width = (float) fabsf((float) (bb_right - bb_left))/um->normal_width;

  if (left_pointing || right_pointing)
    pthresh = gp->pointing_width - gp->hysteresis;
  else
    pthresh = gp->pointing_width + gp->hysteresis;
  
  if (left_pointing && right_pointing)
    sthresh = gp->spread_eagle_width - gp->hysteresis;
  else
    sthresh = gp->spread_eagle_width + gp->hysteresis;
  
  if (width > sthresh  &&
      handNotMoving(um, RIGHT_HAND, gp->pointing_still_time) &&
      handNotMoving(um, LEFT_HAND, gp->pointing_still_time))
    /* spread eagle */
    {
      left_pointing = TRUE;
      right_pointing = TRUE;
      *hand = BOTH_HANDS;
      return TRUE;
    }

  else if (width > pthresh)
    {
        float cx, cy;
	float middle;
	float dx;
	float hly, hry;

	hly = rbGetElement(um->buffer[HAND_LEFT_Y], 0, &error);
	hry = rbGetElement(um->buffer[HAND_RIGHT_Y], 0, &error);
	
	middle = (bb_right - bb_left)/2.0 + bb_left;
	
	cx = rbGetElement(um->buffer[CENTROID_X], 0, &error);
	cy = rbGetElement(um->buffer[CENTROID_Y], 0, &error);
	
	dx = cx - middle;
	if (dx < 0.0 && handNotMoving(um, RIGHT_HAND, gp->pointing_still_time)
	    && hry > cy)
	  {
	    *hand = RIGHT_HAND;
	    right_pointing = TRUE;
	    left_pointing = FALSE;
	    return TRUE;
	  }
	else if ( dx > 0.0 &&
		 handNotMoving(um, LEFT_HAND, gp->pointing_still_time) &&
		 hly > cy)
	  {
	    *hand = LEFT_HAND;
	    right_pointing = FALSE;
	    left_pointing = TRUE;
	    return TRUE;
	  }
	else if (dx == 0.0)
#if 1
	  fprintf(stderr, "Centroid at middle for pointing!\n");
#endif
      }

  else
    {
      left_pointing = FALSE;
      right_pointing = FALSE;
      return FALSE;
    }
  return FALSE;
}


boolean
handExtended(UserModel *um, Hand which, int start)
{
  rbIndex hxi, hyi, valid;
  boolean error;
  float v;
  float dist;
  
  if (which == LEFT_HAND)
    {
      hxi = HAND_LEFT_X;
      hyi  = HAND_LEFT_Y;
      valid = HAND_LEFT_VALID;
    }
  else
    {
      hxi = HAND_RIGHT_X;
      hyi  = HAND_RIGHT_Y;
      valid = HAND_RIGHT_VALID;
    }
  
  v = rbGetElement(um->buffer[valid], start, &error);
  if (error || v < 0.5) /* hand invisible */
    return FALSE;

  dist = rbGetElement(um->buffer[CENTROID_X], 0, &error) -
    rbGetElement(um->buffer[hxi], 0, &error);
  if (error) return FALSE;
  
  return (ABS(dist) > um->gp->min_hand_dist_for_extend);
  
}


/* should be modified to take a start tick!!! */
boolean 
handNotMoving(UserModel *um, Hand which, float time)
{
  float hx, hy; /* initial position */
  float dur = 0.0;
  float x= 0.0, y=0.0;
  float points = 1.0;
  boolean error;
  rbIndex hxi, hyi;
  int tick = 0;
  GestureParams *gp = um->gp;
  
  if (which == LEFT_HAND)
    {
      hxi = HAND_LEFT_X;
      hyi = HAND_LEFT_Y;
    }
  else 
    {
      hxi = HAND_RIGHT_X;
      hyi = HAND_RIGHT_Y;
    }

  hx = rbGetElement(um->buffer[hxi], 0, &error);
  if (error) return FALSE;
  hy = rbGetElement(um->buffer[hyi], 0, &error);
  
  x = hx;
  y = hy;
  error = FALSE;
  tick ++;
  while (dur < time && !error)
    {
      float dx, dy;

      dur += rbGetElement(um->buffer[TIME], tick-1, &error) - 
	rbGetElement(um->buffer[TIME], tick, &error);
      if (error) continue;
      
      x += rbGetElement(um->buffer[hxi], tick, &error);
      y += rbGetElement(um->buffer[hyi], tick, &error);
      points += 1.0;
      
      /* if distance between average point and initial changes by 
	 more than thresh, nbot moving */
      dx = x/points - hx;
      dy = y/points - hy;
      if (sqrtf(dx*dx + dy*dy) >= gp->max_dist_for_hand_not_moving)
	return FALSE;
      tick++;
    }
  if (error)
    return FALSE;
  else return TRUE;
}


boolean
personStanding(UserModel *um)
{
  static boolean standing = FALSE; /* memory.  Were we standing last call */
  float dist;
  GestureParams *gp = um->gp;
  float height;
  boolean error;
  
  height = rbGetElement(um->buffer[TOP], 0, &error);
  if (error) return FALSE;
  
  height -= rbGetElement(um->buffer[BOTTOM], 0, &error);
  if (error) return FALSE;
  
  height /= um->normal_height; /* normalize to normal height */
  
  dist = ABS(height - 1.0);
  standing = (boolean) (dist <= gp->standing_height);
  return standing;    
  

}


/* 
  is there an oscillation within the last _secs_, starting
  at _start_ticks_?  Return index where it ends (further in past
  than _start_ticks_, ie > _start_ticks_ or 0 for none.
  oscillation is defined as derivative exceeding threshold
  in in one direction adn then the other */
int
oscillation(UserModel *um, rbIndex which, int start_ticks, float secs,
	    float deriv_thresh)
{
  /* if I only had a shotgun */
  boolean error;
  float start_time;
  float dur=0.0;
  float direction = 0.0;
  int ticks = start_ticks;
  
  start_time = rbGetElement(um->buffer[TIME], start_ticks, &error);
  if (error) return 0;
  error = FALSE;
  
  while(dur <= secs && !error)
    {
      float dx;
      
      ticks++;
      dur = start_time - rbGetElement(um->buffer[TIME], ticks, &error) ;
      if (error) continue; /* we outta here */
      dx = rbGetElement(um->buffer[which], ticks, &error);
      if (direction == 0.0)
	{
	  if (dx >= deriv_thresh)
	    direction = 1.0;
	  else if (dx <= -deriv_thresh)
	    direction = -1.0;
	}
      else if (direction == 1.0 && dx <= -deriv_thresh)
	return ticks;
      else if (direction == -1.0 && dx >= deriv_thresh)
	return ticks;      
    }
  return 0;

}


/* lookm for a jerk starting within _secs_ of  _start_ticks_ . 
   Set *direction to direction
   of jerk (-1.0, 1.0) and return the tick for the end of the jerk, or
   the last one before the end of the buffer.  If not jerk, return 0.
   Jerk must occur within _secs_ seconds and go above a thresh of _thresh_.
   */
int
jerk(UserModel *um, rbIndex which, int start_ticks,
     float secs, float thresh, float *direction)
{

  /* if I only had a shotgun */
  boolean error;
  float start_time;
  float dur = 0.0;
  int ticks = start_ticks;
  
  *direction = 0.0;
  start_time = rbGetElement(um->buffer[TIME], start_ticks, &error);
  if (error) return 0;
  error = FALSE;
  
  while(dur <= secs && !error && *direction == 0.0)
    {
      float dx;
      
      dur = start_time - rbGetElement(um->buffer[TIME], ticks, &error) ;
      if (error) continue; /* we outta here */
      dx = rbGetElement(um->buffer[which], ticks, &error);
      
      if (dx >= thresh)
	*direction = 1.0;
      else if (dx <= -thresh)
	*direction = -1.0;

      ticks++;

    }

  if (error || dur > secs) return FALSE;
  
  /* else we have a jerk end.  Look for start */
  while (!error)
    {
      float dx;
      
      dx = rbGetElement(um->buffer[which], ticks, &error);
      if (error) continue;

      if (*direction == -1.0 && dx > -thresh)
	return ticks;
      else if (*direction == 1.0 && dx < thresh)
	return ticks;
      
      ticks++;

    }
  
  return ticks-1; /* return as much as in the buffer */
  
}

/* person making a throwing gesture with hand specified in _which_.
 Does not do BOTH_HANDS.  */
boolean
personThrowing(UserModel *um, Hand which)
{
  GestureParams *gp = um->gp;
  float direction;
  int tick;
  boolean error;
  
  switch (which)
    {
     case LEFT_HAND:
      /* was there a jerk ending in the last gp->throwing_time?*/
      tick =  jerk(um, 
		   DHAND_LEFT_X, 
		   0, 
		   gp->throwing_time, 
		   gp->throwing_thresh,
		   &direction);
      break;
      
     case RIGHT_HAND:
      tick = jerk(um, 
		  DHAND_RIGHT_X, 
		  0, 
		  gp->throwing_time, 
		  gp->throwing_thresh,
		  &direction);
      break;
      
     default:
      fprintf(stderr, "No such hand\n");
      break;
    }

  if (!tick) return FALSE;
  error = FALSE;
  /* if so, check the current width */
  if ((normalizedWidth(um, 0, &error) >= gp->pointing_width) && !error)
    return TRUE;
  else return FALSE;
  
      
}



boolean
personKicking(UserModel *um, Foot which)
{
  GestureParams *gp = um->gp;
  int tick;
  float direction=0.0;
  boolean error;
  float dist;
  
  switch (which)
    {
     case LEFT_FOOT:
      tick = jerk(um, 
		  DFOOT_LEFT_X, 
		  0, 
		  gp->throwing_time, 
		  gp->kicking_thresh,
		  &direction);
      dist = rbGetElement(um->buffer[FOOT_LEFT_Y], tick, &error) - 
	rbGetElement(um->buffer[BOTTOM], tick, &error);
      if (error) return FALSE;
      if (ABS(dist) < gp->min_foot_height_for_kick) /* if not hight enough, no kick */
	tick = 0;
      
      break;
      
     case RIGHT_FOOT:
      tick = jerk(um, 
		  DFOOT_RIGHT_X, 
		  0, 
		  gp->throwing_time, 
		  gp->kicking_thresh, 
		  &direction);
      dist = rbGetElement(um->buffer[FOOT_RIGHT_Y], tick, &error) - 
	rbGetElement(um->buffer[BOTTOM], tick, &error);
      if (error) return FALSE;
      if (ABS(dist) < gp->min_foot_height_for_kick) /* if not hight enough, no kick */
	tick = 0;
      
      break;
      
     default:
      fprintf(stderr, "No such foot\n");
      break;
    } 

  if (!tick) return FALSE;
  else return TRUE;
  
}


boolean
personWaving(UserModel *um, Hand which)
{
  /* harmonic oscillation or hormonic osculation? */
  int kisses = 0;
  rbIndex dhxi, dhyi;
  rbIndex hxi, hyi;
  rbIndex hvalid;
  GestureParams *gp = um->gp;
  int last_frame = 0;
  int first_frame = -1;
  float dist;
  boolean error = FALSE;
  
  if (which == LEFT_HAND)
    {
      dhxi = DHAND_LEFT_X;
      dhyi = DHAND_LEFT_Y;
      hxi = HAND_LEFT_X;
      hyi = HAND_LEFT_Y;
      hvalid = HAND_LEFT_VALID;
      
    }
  else
    {
      dhxi = DHAND_RIGHT_X;
      dhyi = DHAND_RIGHT_Y;
      hxi = HAND_RIGHT_X;
      hyi = HAND_RIGHT_Y;
      hvalid = HAND_RIGHT_VALID;
    }
  
  
  while(kisses < gp->oscillations_for_wave)
    {
      int x_frame;
      int y_frame;

      x_frame = oscillation(um, dhxi, last_frame, 
			    gp->throwing_time, gp->waving_thresh);
      y_frame = oscillation(um, dhyi, last_frame, 
			    gp->throwing_time, gp->waving_thresh);
      
      if (x_frame > 0)
	{
	  if (distanceBetweenPoints(rbGetElement(um->buffer[hxi], 
						 last_frame, 
						 &error),
				    rbGetElement(um->buffer[hyi], 
						 last_frame, 
						 &error),
				    rbGetElement(um->buffer[hxi], 
						 x_frame, 
						 &error),
				    rbGetElement(um->buffer[hyi], 
						 x_frame, 
						 &error))
	      < gp->min_hand_move_dist_for_wave)
	    return FALSE;						 
	  
	   if (rbGetElement(um->buffer[hvalid], x_frame, &error)
	      < 0.5)
	    return FALSE; 
	  
	  last_frame = x_frame;
	  kisses++;
	  if (first_frame < 0) first_frame = x_frame;
	}
      else if (y_frame > 0)
	{
	  if (distanceBetweenPoints(rbGetElement(um->buffer[hxi], 
						 last_frame, 
						 &error),
				    rbGetElement(um->buffer[hyi], 
						 last_frame, 
						 &error),
				    rbGetElement(um->buffer[hxi], 
						 y_frame, 
						 &error),
				    rbGetElement(um->buffer[hyi], 
						 y_frame, 
						 &error))
	      < gp->min_hand_move_dist_for_wave)
	    return FALSE;						 
	  
	  if (rbGetElement(um->buffer[hvalid], y_frame, &error)
	      < 0.5)
	    return FALSE; 
	  
	  last_frame = y_frame;
	  kisses++;
	}
      else 
	{
	  return FALSE;
	}
    }
  
  dist = rbGetElement(um->buffer[hxi], last_frame, &error) - 
    rbGetElement(um->buffer[CENTROID_X], last_frame, &error);  
  if (error || ABS(dist) < gp->min_hand_dist_for_wave)
    return FALSE;
  else
    return TRUE;
  
}

float
distanceBetweenPoints(float x1, float y1, float x2, float y2)
{
  float dx, dy;

  dx = x1 - x2;
  dy = y1 - y2;
  return (sqrtf(dx*dx + dy*dy));
}

boolean
personHasAShotgunAimedAtHead(UserModel *um)
{
  UserModel *crap;

  if (um == crap) return TRUE;
  else return FALSE;
}


void
findGesturesUserModel(UserModel *um, GestureFlags *desired, GestureFlags *found)
{
  bzero((unsigned char *) found, sizeof(GestureFlags));
  

  if (um->user_exists)
    found -> user_exists = TRUE;
  else
    return;

  if (desired->standing)
    found -> standing = personStanding(um);
  
  if (desired->sitting)
    found -> sitting = personSitting(um);
  
  if (desired->hand_over_head)
    found -> hand_over_head = personHasHandOverHead(um);
  
  if (desired->bending_over)
    found -> bending_over = personBending(um);

  if (desired->point_left || desired->point_right)
    {
      Hand which;
      boolean got_one;

      got_one = personPointing(um, &which);
      
      if (got_one)
	{
	  switch (which)
	    {
	     case LEFT_HAND:
	      found -> point_left = TRUE;
	      break;
	      
	     case RIGHT_HAND:
	      found -> point_right = TRUE;
	      break;
	      
	     case BOTH_HANDS:
	      found -> point_left = TRUE;
	      found -> point_right = TRUE;
	      break;
	    }
	}
    }

  if (desired->hand_moving_left)
    found -> hand_moving_left = 
      ! handNotMoving(um, LEFT_HAND, um->gp->pointing_still_time);

  if (desired->hand_moving_right)
    found -> hand_moving_right = 
      ! handNotMoving(um, RIGHT_HAND, um->gp->pointing_still_time);
  
  if (desired -> extend_left)
    found -> extend_left = handExtended(um, LEFT_HAND, 0);

  if (desired -> extend_right)
    found -> extend_right = handExtended(um, RIGHT_HAND, 0);
  
  if (desired->throw_left)
    found -> throw_left = personThrowing(um, LEFT_HAND);

  if (desired->throw_right)
    found -> throw_right = personThrowing(um, RIGHT_HAND);

  if (desired->wave_left)
    found -> wave_left = personWaving(um, LEFT_HAND);

  if (desired->wave_right)
    found -> wave_right = personWaving(um, RIGHT_HAND);
  
  
  if (desired->kick_left)
    found ->kick_left = personKicking(um, LEFT_FOOT);

  if (desired->kick_right)
    found ->kick_right = personKicking(um, RIGHT_FOOT);
  
  
}


void
printGestureFlags(FILE *fp, GestureFlags *gf)
{
  if (gf->user_exists)
    fprintf(fp, "EXISTS ");
  if (gf->standing)
    fprintf(fp, "STANDING ");
  if (gf->sitting)
    fprintf(fp, "SITTING ");
    if (gf->bending_over)
    fprintf(fp, "BENDING ");
  if (gf->hand_over_head)
    fprintf(fp, "HAND_OVER_HEAD ");
  if (gf->extend_left)
    fprintf(fp, "EXTEND_LEFT ");
  if (gf->extend_right)
    fprintf(fp, "EXTEND_RIGHT ");
  if (!gf->hand_moving_left)
    fprintf(fp, "LEFT_STILL ");
  if (!gf->hand_moving_right)
    fprintf(fp, "RIGHT_STILL ");
  
  if(gf->point_left)
    fprintf(fp, "POINT_LEFT ");
  if(gf->point_right)
    fprintf(fp, "POINT_RIGHT ");
  if(gf->kick_left)
    fprintf(fp, "KICK_LEFT ");
  if(gf->kick_right)
    fprintf(fp, "KICK_RIGHT ");
  if(gf->wave_left)
    fprintf(fp, "WAVE_LEFT ");
  if(gf->wave_right)
    fprintf(fp, "WAVE_RIGHT ");
  if(gf->throw_left)
    fprintf(fp, "THROW_LEFT ");
  if(gf->throw_right)
    fprintf(fp, "THROW_RIGHT ");

}


void
flagAllGestures(GestureFlags *gf)
{
  gf->standing = TRUE;
  gf->sitting = TRUE;
  gf->hand_over_head = TRUE;
  gf->bending_over = TRUE;
  gf->point_left = TRUE;
  gf->point_right = TRUE;
  gf->wave_left = TRUE;
  gf->wave_right = TRUE;
  gf->kick_left = TRUE;
  gf->kick_right = TRUE;
  gf->throw_left = TRUE;
  gf->throw_right = TRUE;
  gf->extend_left = TRUE;
  gf->extend_right = TRUE;
  gf->hand_moving_left = TRUE;
  gf->hand_moving_right = TRUE;
  gf->user_exists = TRUE;
}




