#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>
#include <bstring.h>

#include "common.h"
#include "ring_buffer.h"
#include "user_model.h"

/* protos */
#include "common.proto"
#include "ring_buffer.proto"
#include "user_model.proto"
#include "find_features.proto"

#define HEAD_DIST_FROM_TOP 6.0 /* inches */
#define MAX_HAND_SPEED 12.0 /* inches per second */
#define MIN_DT .2

void
findBodyFeatures(const float *contour, /* 3 coords */
		 int numPoints,
		 const float *extrema, /* 3 coords */
		 int numExtrema,
		 float cx, float cy,
		 float tlx, float tly,
		 float brx, float bry,
		 float *head,
		 float *left_hand,
		 float *right_hand,
		 float *left_foot,
		 float *right_foot,
		 boolean *head_valid,
		 boolean *left_hand_valid,
		 boolean *right_hand_valid,
		 boolean *left_foot_valid,
		 boolean *right_foot_valid)
{

  static Timer then;
  Timer now;
  static float dt = -1.0;
  static float last_left[2];
  static float last_right[2];
  static boolean got_last_left = FALSE;
  static boolean got_last_right = FALSE;
  float height, width;
  
  *head_valid = FALSE;
  *left_foot_valid = FALSE;
  *right_foot_valid = FALSE;

  if (dt < 0.0) then = getTime();
  
  now = getTime();
  dt = durationTime(then, now);
  then = now;
  
  height = tly - bry;
  width = brx - tlx;
  
#ifdef DEBUG
  printf("dt: %f\n", dt);
  sanityCheck(extrema, numExtrema, tlx, tly, brx, bry);
#endif


#if 0
  if(got_last_left)
    got_last_left = findHandNearestNeighbor(extrema, numExtrema, 
					    cx, cy, tlx, tly, 
					    brx, bry, last_left,
					    dt, left_hand);
#endif
  got_last_left = FALSE;
  
  if (!got_last_left)
    got_last_left = findLeftHandStanding(contour,
					 numPoints,
					 extrema, numExtrema, 
					 cx, cy, tlx, tly, 
					 brx, bry, left_hand);
  if (!got_last_left)
    {
      left_hand[0] = tlx;
      left_hand[1] = cy;
    }
  
  *left_hand_valid = got_last_left;
  last_left[0] = left_hand[0];
  last_left[1] = left_hand[1];
			  

  /* fid right */
#if 0
  if (got_last_right)
    got_last_right = findHandNearestNeighbor(extrema, numExtrema, 
					     cx, cy, tlx, tly, 
					     brx, bry, last_right,
					     dt, right_hand);
#endif
  got_last_right = FALSE;
  
  if (!got_last_right)
    got_last_right = findRightHandStanding(contour,
					   numPoints,
					   extrema, numExtrema, 
					   cx, cy, tlx, tly, 
					   brx, bry,
					   right_hand);
  if (!got_last_right)
    {
      right_hand[0] = brx; 
      right_hand[1] = cy;
    }
  
  *right_hand_valid = got_last_right;
  last_right[0] = right_hand[0];
  last_right[1] = right_hand[1];

  head[0] = width/2.0 + tlx;
  head[1] = tly;
  
  left_foot[0] = width*.4 + tlx;
  left_foot[1] = bry;

  right_foot[0] = 2.0*width/3.0 + tlx;
  right_foot[1] = bry;
    
  
}



void
sanityCheck(const float *extrema,
	    int num,
	    float tlx,
	    float tly,
	    float brx,
	    float bry)
{
  int i;
    
  for(i=0; i<num; i++)
    {
      float x,y,z;

      x = extrema[3*i];
      y = extrema[3*i+1];
      z = extrema[3*i+2];
      
      if (x < tlx || x > brx ||
	  y < bry || y > tly)
	fprintf(stderr, "BUG!!!!!! Got extrema: %f %f  %d of %d outside bb.  TL %f %f  br %f %f\n", x, y, i, num, tlx, tly, brx, bry);
    }

}



/* heuristic search */
boolean
findLeftHandStanding(const float *contour, /* 3 coords */
		     int numPoints,
		     const float *extrema, /* 3 coords */
		     int numExtrema,
		     float cx, float cy,
		     float tlx, float tly,
		     float brx, float bry,
		     float *hand)
{
  int i;
  float height = tly - bry;
  float width = brx - tlx;

  float leftmost[2];
  boolean got_one = FALSE;
  int ptr;
  
  
  for(i=0,ptr=0; i<numExtrema; i++)
    {
      float x = extrema[ptr++];
      float y = extrema[ptr++];
      float z = extrema[ptr++];
      
      if (y < height/4.0 + bry)
	continue;
      
      /* likely head */
      if (fabsf(y - tly) <= HEAD_DIST_FROM_TOP)
	continue;

      if (x > width*.334 + tlx)
	continue;

      if (!got_one)
	{
	  leftmost[0] = x;
	  leftmost[1] = y;
	  got_one = TRUE;
	}
      else if (x < leftmost[0])
	{
	  leftmost[0] = x;
	  leftmost[1] = y;
	}
    }
  
  hand[0] = leftmost[0];
  hand[1] = leftmost[1];

  return got_one;
}



/* heuristic search */
boolean
findRightHandStanding(const float *contour, /* 3 coords */
		      int numPoints,
		      const float *extrema, /* 3 coords */
		      int numExtrema,
		      float cx, float cy,
		      float tlx, float tly,
		      float brx, float bry,
		      float *hand)
{
  int i;
  float height = tly - bry;
  float width = brx - tlx;

  float rightmost[2];
  boolean got_one = FALSE;
  int ptr;
  
  for(i=0,ptr=0; i<numExtrema; i++)
    {
      float x = extrema[ptr++];
      float y = extrema[ptr++];
      float z = extrema[ptr++];
      
      if (y < height/4.0 + bry)
	continue;

      /* likely head */
      if (fabsf(y - tly) <= HEAD_DIST_FROM_TOP)
	continue;
      
      if (x < width*.66 + tlx)
	continue;

      if (!got_one)
	{
	  rightmost[0] = x;
	  rightmost[1] = y;
	  got_one = TRUE;
	}
      else if (x > rightmost[0])
	{
	  rightmost[0] = x;
	  rightmost[1] = y;
	}
    }

  hand[0] = rightmost[0];
  hand[1] = rightmost[1];
  
  return got_one;

}
	
     
/* returns success */
boolean
  findHandNearestNeighbor(const float *extrema, /* 3 coords */
			  int numExtrema,
			  float cx, float cy,
			  float tlx, float tly,
			  float brx, float bry,
			  float *last,
			  float dt,
			  float *hand)
			      
			      
{
  
  int i;
  float lastx, lasty;
  float max_dist;
  int ptr;
  float dx, dy;

  lastx = last[0];
  lasty = last[1];

  max_dist = MAX_HAND_SPEED*dt;
  
  for (i=0,ptr=0; i<numExtrema; i++)
    {
      float x = extrema[ptr++];
      float y = extrema[ptr++];
      float z = extrema[ptr++];
      float dist;
      
      dx = x - lastx;
      dy = y - lasty;
      
      dist = sqrtf(dx*dx + dy*dy);
      if (dist <= max_dist)
	{
	  hand[0] = x;
	  hand[1] = y;
	  return TRUE;
	}      
    }

  return FALSE;

}

