// QuickMatrix.h

#ifndef _QuickMatrix_H_
#define _QuickMatrix_H_

#include <stddef.h>
#include <stream.h>
#include <assert.h>
#include "QuickMath/real.h"
#include "QuickMath/byte.h"

// QuickMatrix is the basic class that all operations are defined for:
//   ShadowMatrix and AllocMatrix differ only in how they allocate
//    and deallocate themselves
// Because QM does not know how to allocate/destroy itself, you cannot
//  resize it.  QMs can only be used "as is".  Can change values, but
//  cannot change size except through the constructor.

class QuickMatrix
{
protected:
  int elementCount;
  int width;
  real *array;

public:
  virtual ~QuickMatrix() {}

  QuickMatrix( )
    : elementCount(0), width(0), array(0) {}
  QuickMatrix( int h, int w, real* a )
    : elementCount(h*w), width(w), array(a) {}
  // QuickMatrix( const QuickMatrix& ); // default, mem-wise, i.e. SHADOW

  inline QuickMatrix& Copy( const QuickMatrix& qm );
  QuickMatrix& operator=( const QuickMatrix& qm ) { return Copy(qm); }

  int ElementCount() const { return elementCount; }
  int RowCount() const { return Height(); }
  int ColCount() const { return Width(); }
  int Height() const { return( (width) ? elementCount/width : 0); }
  int Width() const { return width; }
  int RowStep() const { return width; }
  int ColStep() const { return 1; }
  real* Array( int r = 0, int c = 0 ) const { return( array + r*width + c ); }
  operator real*() { return array; }

  QuickMatrix& Insert( real * );
  real* Extract( real * );
  QuickMatrix& operator<<( real *p ) { return Insert(p); }
  real* operator>>( real *p ) { return Extract(p); }

#if 0
  QuickMatrix SubMatrix();
  QuickMatrix SubMatrix( int ulr, int ulc, int dh, int dw );
#endif

  QuickMatrix& Add( const QuickMatrix&, const QuickMatrix& );
  QuickMatrix& Subtract(const QuickMatrix&, const QuickMatrix& );
  QuickMatrix& Multiply(const QuickMatrix&, const QuickMatrix& );
  QuickMatrix& MultiplyTranspose( const QuickMatrix&, const QuickMatrix& );
  QuickMatrix& TransposeMultiply( const QuickMatrix&, const QuickMatrix& );
  // QuickMatrix& Divide( const QuickMatrix&, const QuickMatrix& );
  // QuickMatrix& Negate( const QuickMatrix& );
  QuickMatrix& Invert( const QuickMatrix& );
  // QuickMatrix& Transpose( const QuickMatrix& );
  QuickMatrix& Shift( const QuickMatrix&, real val );
  QuickMatrix& Scale( const QuickMatrix&, real val  );
  QuickMatrix& AddEq( const QuickMatrix& op );
  // QuickMatrix& SubtractEq( const QuickMatrix& op );
  // QuickMatrix& MultiplyEq( const QuickMatrix& op );
  // QuickMatrix& DivideEq( const QuickMatrix& op );
  // QuickMatrix& NegateEq();
  QuickMatrix& InvertEq();
  // QuickMatrix& ShiftEq( real val );
  QuickMatrix& ScaleEq( real val );
  QuickMatrix& ConstEq( real val );
  QuickMatrix& Zero() { return ConstEq( 0. ); }
  QuickMatrix& DiagEq( real val );
  QuickMatrix& Identity() { return DiagEq(1.); }
  QuickMatrix& operator+=( const QuickMatrix& op ) { return AddEq(op); }
  // QuickMatrix& operator-=( const QuickMatrix& op ) { return SubtractEq(op); }
  // QuickMatrix& operator*=( const QuickMatrix& op ) { return MultiplyEq(op); }
  // QuickMatrix& operator/=( const QuickMatrix& op ) { return DivideEq(op); }
  // QuickMatrix& operator+=( const real val ) { return ShiftEq(val); }
  // QuickMatrix& operator-=( const real val ) { return ShiftEq(-val); }
  // QuickMatrix& operator*=( const real val ) { return ScaleEq(val); }
  QuickMatrix& operator/=( const real val ) { return ScaleEq(real(1)/val); }
  // QuickMatrix& operator=( const real val ) { return ConstEq(val); }
  real Determinant3();
  real Determinant2();
  real Determinant();

  ostream& Print( ostream&, char* = "\n" );
  istream& Scan( istream& );
  friend ostream& operator<<( ostream&, QuickMatrix& );
  ostream& Write( ostream& );
  istream& Read( istream& );

  real& operator()( int r, int c ) { return array[r*width + c]; }

private:
  int FindMaxPivot( int, int * );
  int LUDecomp( int * );
  void ForwardSubstitute( real *, real *, int * );
  void BackwardSubstitute( real *, real *, int * );

  friend class AllocMatrix;
};

class ShadowMatrix : public QuickMatrix
{
public:
  ShadowMatrix( int h, int w, real *a )
    : QuickMatrix( h, w, a ) {}
};

class AllocMatrix : public QuickMatrix
{
  real *allocArray;
public:
  AllocMatrix( )
    : QuickMatrix( ), allocArray( 0 ) {}
  AllocMatrix( int h, int w );
  AllocMatrix( int h, int w, real *data );
  AllocMatrix( const AllocMatrix& m );
  AllocMatrix( const QuickMatrix& m );

  AllocMatrix& Init( int h, int w );
  AllocMatrix& Init( int h, int w, real *data );
  AllocMatrix& Init( const QuickMatrix& m );

  AllocMatrix& operator=( const QuickMatrix& qm ) { return Init( qm ); }
  
  virtual ~AllocMatrix() { if( allocArray) delete allocArray; }
};

inline
QuickMatrix&
QuickMatrix::Copy( const QuickMatrix& qm )
{
  assert( elementCount == qm.elementCount );
  width = qm.width;
  memcpy( array, qm.array, elementCount * sizeof(real) );
  return *this;
}

inline
ostream&
QuickMatrix::Write( ostream& os )
{
  return( os.write( (char *)array, elementCount * sizeof(real) ) );
}

inline
istream&
QuickMatrix::Read( istream& is )
{
  return( is.read( (char *)array, elementCount * sizeof(real) ) );
}

#endif /* _QuickMatrix_H_ */
