// QuickVector.h

#ifndef _QuickVector_H_
#define _QuickVector_H_

#include <stddef.h>
#include <stream.h>
#include <assert.h>
#include "QuickMath/real.h"
#include "QuickMath/byte.h"

// QuickVector is the basic class that all operations are defined for:
//   ShadowVector and AllocVector differ only in how they allocate
//    and deallocate themselves

class QuickVector
{
  friend class AllocVector;
public:
  virtual ~QuickVector() {}

protected:
  real *array;
  int elementCount;

public:

  QuickVector( )
    : array(0), elementCount(0) {}
  QuickVector( int n, real* a )
    : array(a), elementCount(n) {}
  QuickVector( const QuickVector& );

#if 0
  QuickVector& Init( );
  QuickVector& Init( const QuickVector& );
  QuickVector& Init( int h, int w );
  QuickVector& Init( int h, int w, real* data );
#endif

  inline QuickVector& Copy( const QuickVector& qv );
  QuickVector& operator=( const QuickVector& qv ) { return Copy(qv); }

  int Length() const { return elementCount; }
  int ElementCount() const { return elementCount; }
  real* Array( int i = 0 ) { return( array + i ); }
  operator real*() { return array; }

  QuickVector& Insert( real *p );
#if 0
  real* Extract( real *p );
  QuickVector& operator<<( real *p ) { Insert(p); return *this; }
  real* operator>>( real *p ) { return Extract(p); }
#endif

#if 0
  QuickVector SubVector();
  QuickVector SubVector( int ulr, int ulc, int dh, int dw );
#endif

  QuickVector& Add( const QuickVector&, const QuickVector& );
  QuickVector& Subtract(const QuickVector&, const QuickVector& );
  // QuickVector& Multiply(const QuickVector&, const QuickVector& );
  // QuickVector& Divide( const QuickVector&, const QuickVector& );
  // QuickVector& Negate( const QuickVector& );
  // QuickVector& Invert( const QuickVector& );
  // QuickVector& Transpose( const QuickVector& );
  // QuickVector& Shift( const QuickVector&, real val );
  // QuickVector& Scale( const QuickVector&, real val  );
  QuickVector& AddEq( const QuickVector& op );
  QuickVector& SubtractEq( const QuickVector& op );
  // QuickVector& MultiplyEq( const QuickVector& op );
  // QuickVector& DivideEq( const QuickVector& op );
  QuickVector& NegateEq();
  // QuickVector& InvertEq();
  QuickVector& ShiftEq( real val );
  QuickVector& ScaleEq( real val );
  QuickVector& ConstEq( real val );
  QuickVector& Zero( ) { return ConstEq( 0.); }
  QuickVector& operator+=( const QuickVector& op ) { return AddEq(op); }
  // QuickVector& operator-=( const QuickVector& op ) { return SubtractEq(op); }
  // QuickVector& operator*=( const QuickVector& op ) { return MultiplyEq(op); }
  // QuickVector& operator/=( const QuickVector& op ) { return DivideEq(op); }
  // QuickVector& operator+=( const real val ) { return ShiftEq(val); }
  QuickVector& operator-=( const real val ) { return ShiftEq(-val); }
  // QuickVector& operator*=( const real val ) { return ScaleEq(val); }
  QuickVector& operator/=( const real val ) { return ScaleEq(real(1)/val); }
  // QuickVector& operator=( const real val ) { return ConstEq(val); }
  real InnerProduct( const QuickVector& ) const;
  int operator==( const QuickVector& ) const;
  QuickVector& Abs();

  ostream& Print( ostream& );
  istream& Scan( istream& );
  ostream& Write( ostream& );
  istream& Read( istream& );

  real& operator()( int i ) { return array[i]; }
};

class ShadowVector : public QuickVector
{
public:
  ShadowVector( int n, real *a )
    : QuickVector( n, a ) {}
};

class AllocVector : public QuickVector
{
    real *allocArray;
public:
    AllocVector( )
      : QuickVector( ), allocArray(0) {}
    AllocVector( int n );
    AllocVector( int n, real *data );
    AllocVector( const AllocVector& );
    AllocVector( const QuickVector& );
    
    AllocVector& Init( int n );
    AllocVector& Init( int n, real *data );
    AllocVector& Init( const QuickVector& );
    
    AllocVector& operator=( const QuickVector& qv ) { return Init(qv); }
    
    virtual ~AllocVector() { if( allocArray ) delete allocArray; }
};

inline
QuickVector&
QuickVector::Copy( const QuickVector& qv )
{
  assert( elementCount == qv.elementCount );
  memcpy( array, qv.array, elementCount * sizeof(real) );
  return( *this );
}

inline
ostream&
QuickVector::Write( ostream& os )
{
  return os.write( (char *)array, elementCount * sizeof(real) );
}

inline
istream&
QuickVector::Read( istream& is )
{
  return is.read( (char *)array, elementCount * sizeof(real) );
}

inline
ostream&
operator<<( ostream& os, QuickVector& qv ) { return qv.Print( os ); }

#endif /* _QuickVector_H_ */
