// SubMatrix.h

#ifndef _SubMatrix_H_
#define _SubMatrix_H_

#include <stddef.h>
#include <stream.h>
#include <assert.h>
#include "QuickMath/real.h"
#include "QuickMath/byte.h"
#include "QuickMath/SubVector.h"

class SubMatrix
{
  friend class SubVector;
protected:
  int elementCount;
  int width;
  int spacing;
  real *array;

public:
  SubMatrix( )
    : elementCount(0), width(0), spacing(0), array(0) {}
  SubMatrix( int h, int w, int line, real* a )
    : elementCount(h*w), width(w), spacing(line-w), array(a) {}
  // SubMatrix( const SubMatrix& ); // default, mem-wise, i.e. SHADOW

  inline SubMatrix& Copy( const SubMatrix& sm );
  SubMatrix& operator=( const SubMatrix& sm ) { return Copy(sm); }

  int ElementCount() const { return elementCount; }
  int RowCount() const { return Height(); }
  int ColCount() const { return Width(); }
  int Height() const { return( (width) ? elementCount/width : 0); }
  int Width() const { return width; }
  int RowStep() const { return width; }
  int ColStep() const { return 1; }
  int Line() const { return( width+spacing ); }
  real* Array( int r = 0, int c = 0 ) const { return( array + r*(width+spacing) + c ); }

  SubMatrix& Insert( real * );
  real* Extract( real * ) const;
  SubMatrix& operator<<( real *p ) { return Insert(p); }
  real* operator>>( real *p ) { return Extract(p); }

#if 0
  SubMatrix SubMatrix();
  SubMatrix SubMatrix( int ulr, int ulc, int dh, int dw );
#endif

  SubMatrix& Add( const SubMatrix&, const SubMatrix& );
  SubMatrix& Subtract(const SubMatrix&, const SubMatrix& );
  SubMatrix& Multiply(const SubMatrix&, const SubMatrix& );
  SubMatrix& MultiplyTranspose(const SubMatrix&, const SubMatrix& );
  // SubMatrix& Divide( const SubMatrix&, const SubMatrix& );
  // SubMatrix& Negate( const SubMatrix& );
  SubMatrix& Invert( const SubMatrix& );
  // SubMatrix& Transpose( const SubMatrix& );
  // SubMatrix& Shift( const SubMatrix&, real val );
  // SubMatrix& Scale( const SubMatrix&, real val  );
  SubMatrix& AddEq( const SubMatrix& op );
  // SubMatrix& SubtractEq( const SubMatrix& op );
  // SubMatrix& MultiplyEq( const SubMatrix& op );
  // SubMatrix& DivideEq( const SubMatrix& op );
  // SubMatrix& NegateEq();
  SubMatrix& InvertEq();
  // SubMatrix& ShiftEq( real val );
  SubMatrix& ScaleEq( real val );
  SubMatrix& ConstEq( real val );
  SubMatrix& DiagEq( real val );
  SubMatrix& Identity() { return DiagEq( 1. ); }
  SubMatrix& Zero() { return ConstEq( 0. ); }
  SubMatrix& operator+=( const SubMatrix& op ) { return AddEq(op); }
  // SubMatrix& operator-=( const SubMatrix& op ) { return SubtractEq(op); }
  // SubMatrix& operator*=( const SubMatrix& op ) { return MultiplyEq(op); }
  // SubMatrix& operator/=( const SubMatrix& op ) { return DivideEq(op); }
  // SubMatrix& operator+=( const real val ) { return ShiftEq(val); }
  // SubMatrix& operator-=( const real val ) { return ShiftEq(-val); }
  // SubMatrix& operator*=( const real val ) { return ScaleEq(val); }
  SubMatrix& operator/=( const real val ) { return ScaleEq(real(1)/val); }
  // SubMatrix& operator=( const real val ) { return ConstEq(val); }

  ostream& Print( ostream&, char* = "\n" );
  istream& Scan( istream& );
  friend ostream& operator<<( ostream&, SubMatrix& );
  ostream& Write( ostream& );
  istream& Read( istream& );

  real& operator()( int r, int c ) { return array[r*width + c]; }

private:
  int FindMaxPivot( int, int * );
  int LUDecomp( int * );
  void ForwardSubstitute( real *, real *, int * );
  void BackwardSubstitute( real *, real *, int * );

  friend class AllocMatrix;
};

inline
SubMatrix&
SubMatrix::Copy( const SubMatrix& sm )
{
  assert( elementCount == sm.elementCount );
  width = sm.width;
  memcpy( array, sm.array, elementCount * sizeof(real) );
  return *this;
}

inline
ostream&
SubMatrix::Write( ostream& os )
{
  return( os.write( (char *)array, elementCount * sizeof(real) ) );
}

inline
istream&
SubMatrix::Read( istream& is )
{
  return( is.read( (char *)array, elementCount * sizeof(real) ) );
}

#endif /* _SubMatrix_H_ */
